//{ **************************************************************************** }
//{ FileName............: CONTROL.CPP                                            }
//{ Project.............: SAA7146A                                               }
//{ Author(s)...........: M.Majoor                                               }
//{ Original contents...: Walter Oney ('Programming the Windows Driver Model')   }
//{ Version.............: 3.00                                                   }
//{ ---------------------------------------------------------------------------- }
//{ IOCTL handlers for SAA7146A driver.                                          }
//{                                                                              }
//{ Version  Date      Comment                                                   }
//{ 1.00     20010818  - Initial release                                         }
//{ 1.01     20020917  - Access to common buffer                                 }
//{ 2.00     20020930  - Access to local buffers                                 }
//{                    - IOCTL_DMA_READ now uses bytes                           }
//{                    - Added 'break' after status has been set definately      }
//{                      (less 'if..else..')                                     }
//{ 2.01     20040514  - No changes                                              }
//{ 2.02     20050408  - Bug correction copy size IRQ_READ/IRQ_WRITE             }
//{ 2.03     20050427  - Minor changes to compile with Windows Server 2003 DDK   }
//{ 3.00     20060624  - Changes for 64-bit driver being called by 32-bit process}
//{                      Basically only other typecasting used                   }
//{                    - Version information also returns 32/64-bit status       }
//{                      (debug information only)                                }
//{ **************************************************************************** }


#include "stddcls.h"
#include "driver.h"
#include "ioctls.h"
#include "version.h"
#include "perf.h"

#include "ExResourceGuard.h"
#include "FastMutexGuard.h"


//{ **************************************************************************** }
//{ Descript : Forwards.                                                         }
//{ **************************************************************************** }
NTSTATUS CacheControlRequest   (PDEVICE_EXTENSION pdx,  PIRP Irp, PIRP* pIrp);
VOID     OnCancelPendingIoctl  (PDEVICE_OBJECT    fdo,  PIRP Irp);
NTSTATUS OnCompletePendingIoctl(PDEVICE_OBJECT    junk, PIRP Irp, PDEVICE_EXTENSION pdx);

// ISR synchronization critical section routines
BOOLEAN IsrSynCritSecDmaRelease( PVOID context );
BOOLEAN IsrSynCritSecFifoRelease( PVOID context );
BOOLEAN IsrSynCritSecFifoRead( PVOID context );
BOOLEAN IsrSynCritSecSetS626Active( PVOID context );
BOOLEAN IsrSynCritSecCopyIrqBuffer( PVOID context );
BOOLEAN IsrSynCritSecExchangePtr( PVOID context );
BOOLEAN IsrSynCritSecReadIrqFifoBuffers( PVOID context );
BOOLEAN IsrSynCritSecUpdateActiveIrqBufList( PVOID context );

//{ **************************************************************************** }
//{ Params   : <fdo>     Functional device object                                }
//{            <Irp>     I/O request packet                                      }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : Dispatcher of IRP_MJ_DEVICE_CONTROL requests.                     }
//{ Notes    : Called through a user-mode DeviceIoControl API with the following }
//{            prototype:                                                        }
//{             result = DeviceIoControl(Handle, Code, InputData, InputLength,   }
//{                        OutputData, OutputLength, &Feedback, &Overlapped);    }
//{            Handle (HANDLE) is an open handle open to the device. You obtain  }
//{            this handle by calling CreateFile in the following manner:        }
//{             Handle = CreateFile("\\\\.\\IOCTL", GENERIC_READ | GENERIC_WRITE,}
//{                                 0, NULL, OPEN_EXISTING, flags, NULL);        }
//{             if (Handle == INVALID_HANDLE_VALUE)                              }
//{               <error>                                                        }
//{             ...                                                              }
//{             CloseHandle(Handle);                                             }
//{ **************************************************************************** }
#pragma PAGEDCODE
NTSTATUS DispatchControl(PDEVICE_OBJECT fdo, PIRP Irp)
{
  // You can be sure of being called at PASSIVE_LEVEL, so there's no particular reason
  // for a simple dispatch function to be anywhere but paged memory.
  PAGED_CODE();

  // Get the device extension where all 'user' data is
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

  // Like other dispatch functions, this one needs to claim the remove lock while
  // it does its work. That prevents the device object from disappearing out from
  // underneath us because of a PnP event.
  NTSTATUS status = IoAcquireRemoveLock(&pdx->RemoveLock, Irp);
  if (!NT_SUCCESS(status))
    return CompleteRequest(Irp, status, 0);
  ULONG info = 0;
  BOOLEAN Escape;

  // On a 64-bit system (with a 64-bit compiled driver obviously) we can be
  // called by a 32-bit process. In such a case we have to convert the
  // native size of the driver (64-bit) to the used 32-bit structures by
  // the 32-bit process.
  // With a 32-bit driver we don't have to deal with that: the structures all
  // have the native size.
  #ifdef _WIN64
    BOOLEAN IsNativeSize = !IoIs32bitProcess(Irp);         // V3.00
  #else
    BOOLEAN IsNativeSize = TRUE;                           // V3.00
  #endif

  // The next few statements extract the function code and buffer sizes from the
  // parameters union in the I/O stack. You often need these values no matter which
  // specific IOCTL you're processing, so I find it easier to always include these
  // statements in the function.
  PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
  ULONG              cbin  = stack->Parameters.DeviceIoControl.InputBufferLength;
  ULONG              cbout = stack->Parameters.DeviceIoControl.OutputBufferLength;
  ULONG              code  = stack->Parameters.DeviceIoControl.IoControlCode;

  // MDL used to handle user-supplied buffer
  PMDL callersBufMdl = NULL;

  switch (code)
  {                                                        // Process request
    case IOCTL_GET_VERSION:                                // Code == 0x800
    {                                                      // GET VERSION
      KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_VERSION)\n"));
      #ifdef _WIN64
        if (IsNativeSize)
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_VERSION) - 64-bit version\n"));
        else
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_VERSION) - 64-bit version, 32-bit process\n"));
      #else
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_VERSION) - 32-bit version\n"));
      #endif
      if (cbout < sizeof(GETVERSION))
      { KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_VERSION) -> invalid output parameter length, expected %d, detected %d\n", sizeof(GETVERSION), cbout));
        status = STATUS_INVALID_PARAMETER;
        break;
      }
      PGETVERSION pversion = (PGETVERSION) Irp->AssociatedIrp.SystemBuffer;
      info = sizeof(GETVERSION);
      pversion->majorversion = VERMAJOR;
      pversion->minorversion = VERMINOR;
      pversion->build = BUILD;
      break;
    }

    case IOCTL_GET_STATUS:                                 // Code == 0x801
    {                                                      // GET STATUS
      
      PULONG pindex = (PULONG) Irp->AssociatedIrp.SystemBuffer;
      ULONG  index = *pindex;
      if (IsNativeSize)
      { 
        // Native size
        if ((cbin < sizeof(ULONG)) || (cbout < sizeof(GETSTATUS_NATIVE)))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_STATUS) -> invalid parameter length, input expected %d, input detected %d, output expected %d, output detected %d\n", sizeof(ULONG), cbin, sizeof(GETSTATUS_NATIVE), cbout));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PGETSTATUS_NATIVE pstatus = (PGETSTATUS_NATIVE) Irp->AssociatedIrp.SystemBuffer;
        info = sizeof(GETSTATUS_NATIVE);
        pstatus->interrupts    = pdx->InterruptCount;
        pstatus->isr           = pdx->ISR;
        pstatus->fifoOverflows = pdx->fifoOverflows;
        pstatus->buffersize = 0;                           // In case we DON'T have a valid buffer
        
        if (index < DMABUFFERS)
        {           
          ExResourceSharedGuard guard( &pdx->dmaBufferLock );
          if (pdx->dmaBuffers[index].bufferId >= 0)
          { 
            pstatus->vaBuffer   = pdx->dmaBuffers[index].vaBuffer;
            pstatus->paBuffer   = pdx->dmaBuffers[index].paBuffer;
            pstatus->buffersize = pdx->dmaBuffers[index].buffersize;
          }
        }
      }
      else
      { 
        // 32-bit process using 64-bit driver
        if ((cbin < sizeof(ULONG)) || (cbout < sizeof(GETSTATUS_32)))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_STATUS) -> invalid parameter length, input expected %d, input detected %d, output expected %d, output detected %d\n", sizeof(ULONG), cbin, sizeof(GETSTATUS_32), cbout));
          status = STATUS_INVALID_PARAMETER;
          break;
        }
        
        PGETSTATUS_32 pstatus32 = (PGETSTATUS_32) Irp->AssociatedIrp.SystemBuffer;
        info = sizeof(GETSTATUS_32);
        pstatus32->interrupts    = pdx->InterruptCount;
        pstatus32->isr           = pdx->ISR;
        pstatus32->fifoOverflows = pdx->fifoOverflows;
        pstatus32->buffersize = 0;                         // In case we DON'T have a valid buffer
        
        if (index < DMABUFFERS)
        { 
          ExResourceSharedGuard guard( &pdx->dmaBufferLock );
          if (pdx->dmaBuffers[index].bufferId >= 0)
          { 
            pstatus32->vaBuffer   = (VOID*POINTER_32)PtrToLong(pdx->dmaBuffers[index].vaBuffer);
            pstatus32->paBuffer   = pdx->dmaBuffers[index].paBuffer;
            pstatus32->buffersize = pdx->dmaBuffers[index].buffersize;
          }
        }
      }
      break;
    }

    case IOCTL_SAA7146A_READ:                              // Code == 0x802
    {                                                      // SAA7146A READ
      // A SAA7146 READ consists of an address to read from.
      // For simplicity we use the same type for source and destination.
      if ((cbin < sizeof(ULONG)) || (cbout < sizeof(ULONG)))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_SAA7146A_READ) -> invalid parameter length, input expected %d, input detected %d, output expected %d, output detected %d\n", sizeof(ULONG), cbin, sizeof(ULONG), cbout));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      PULONG poffset = (PULONG) Irp->AssociatedIrp.SystemBuffer;   // Get pointer to buffer
      
      if (((*poffset%4) != 0) || (*poffset >= pdx->memsize))       // Address MUST be valid (4 byte boundary)
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_SAA7146A_READ) -> invalid address (must be on 4 byte boundary)"));
        status = STATUS_INVALID_PARAMETER;
        break;
      }
      
      info = sizeof(ULONG);
      ULONG offset = *poffset;
      *poffset = READ_REGISTER_ULONG((PULONG) (pdx->membase + *poffset));
      //KdPrint((DRIVERNAME " - DispatchControl (IOCTL_SAA7146A_READ) -> read 0x%x, from 0x%x\n", *poffset, offset));
      break;
    }

    case IOCTL_SAA7146A_WRITE:                             // Code == 0x803
    {                                                      // SAA7146A WRITE
      // A SAA7146 WRITE consists of an address to write to and data.
      // For simplicity we use the same type for address and data.
      if (cbin < sizeof(SAA7146WRITE))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_SAA7146A_WRITE) -> \
           invalid parameter length, input expected %d, input detected\n", 
           sizeof(SAA7146WRITE), cbin));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      // Get pointer to buffer
      PSAA7146WRITE poffset = (PSAA7146WRITE) Irp->AssociatedIrp.SystemBuffer;    

      // Address MUST be valid (4 byte boundary)
      if (((poffset->address%4) != 0) || (poffset->address >= pdx->memsize))      
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_SAA7146A_WRITE) -> \
                            invalid address (must be on 4 byte boundary)"));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      WRITE_REGISTER_ULONG((PULONG) (pdx->membase + poffset->address), poffset->data);
      info = 0;
      //KdPrint((DRIVERNAME " - DispatchControl (IOCTL_SAA7146A_WRITE) -> wrote 0x%x, to 0x%x\n", poffset->data, poffset->address));
      break;
    }

    // Note: only ONE can call this at any time (there is only one 'NotifyIrp')
    case IOCTL_WAIT_NOTIFY:                                // Code == 0x804
    {
      //KdPrint((DRIVERNAME " - DispatchControl (IOCTL_WAIT_NOTIFY) "
      //  "Parameters.DeviceIoControl.OutputBufferLength %d\n", cbout));

      if ( !(cbout == sizeof(ULONG)) & !(cbout == sizeof(pdx->s626IrqStatus)) )
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_WAIT_NOTIFY) -> "
           "invalid output buffer size, expected %d or %d, recv %d\n", 
           sizeof(ULONG), sizeof(pdx->s626IrqStatus), cbout));
        status = STATUS_INVALID_PARAMETER;
      }
      else
      {
        // save OutputBufferLength
        PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation( Irp );
        stack->Parameters.Others.Argument2 = (PVOID) cbout;
        
        // the cache operation will push onto the Irp stack to facilitate
        //  calling a IoCompletion routine
        status = CacheControlRequest(pdx, Irp, &pdx->NotifyIrp);

        // set OutputBufferLength in cached IRP stack location because 
        //  IoCopyCurrentIrpStackLocationToNext doesn't appear to copy this parameter.
        stack = IoGetCurrentIrpStackLocation( pdx->NotifyIrp );
        stack->Parameters.DeviceIoControl.OutputBufferLength = cbout;

        //KdPrint((DRIVERNAME " - DispatchControl (IOCTL_WAIT_NOTIFY) buffer 0x%08X, size %d\n",
        //  pdx->NotifyIrp->AssociatedIrp.SystemBuffer, 
        //  stack->Parameters.DeviceIoControl.OutputBufferLength));
      }
      
      break;
    }

    case IOCTL_GENERATE_EVENT:                             // Code == 0x805
    {
      KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GENERATE_EVENT)\n"));

      if (cbin < sizeof(ULONG))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GENERATE_EVENT) -> invalid parameter length, input expected %d, input detected\n", sizeof(ULONG), cbin));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      PIRP nfyirp = UncacheControlRequest(pdx, &pdx->NotifyIrp);

      if (nfyirp)
      { 
        // If complete notification IRP
        *(PULONG) nfyirp->AssociatedIrp.SystemBuffer = *(PULONG) Irp->AssociatedIrp.SystemBuffer;

        PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(nfyirp);      

        // Complete the cached IRP request        
        CompleteRequest(nfyirp, STATUS_SUCCESS, 
          (sizeof(S626IRQSTATUS) == stack->Parameters.DeviceIoControl.OutputBufferLength ? 
          sizeof(S626IRQSTATUS) : sizeof(ULONG)) );
      }
      else
      {
        status = STATUS_UNSUCCESSFUL;
        info = 0;
      }
      break;

      DumpPerfCounters();
    }

    case IOCTL_DMA_ALLOCATE:                               // Code == 0x806
    {                                                      // DMA ALLOCATE
      ULONG CheckSize;
      if (IsNativeSize)
        CheckSize = sizeof(DMABUFFER_NATIVE);
      else
        CheckSize = sizeof(DMABUFFER_32);
      
      if ((cbin < sizeof(ULONG)) || (cbout < CheckSize))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_ALLOCATE) -> invalid parameter length, input expected %d, input detected %d, output expected %d, output detected %d\n", sizeof(ULONG), cbin, CheckSize, cbout));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      { // lock DMA buffers
        ExResourceExcGuard guard( &pdx->dmaBufferLock );

        // Find a free record were we can store our DMA buffer data
        ULONG freebuf;
        freebuf = 0;
        while ((pdx->dmaBuffers[freebuf].bufferId >= 0) && (freebuf < DMABUFFERS))
          freebuf++;
        
        if (freebuf >= DMABUFFERS)
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_ALLOCATE) -> all DMA buffers already in use\n"));
          status = STATUS_INSUFFICIENT_RESOURCES;
          break;
        }

        PULONG size = (PULONG) Irp->AssociatedIrp.SystemBuffer;
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_ALLOCATE) -> buffer %d, allocating %d bytes\n", freebuf, *size));
        
        // Allocate the buffer
        pdx->dmaBuffers[freebuf].vaBuffer =
          (*pdx->AdapterObject->DmaOperations->AllocateCommonBuffer)
            (pdx->AdapterObject, *size, &pdx->dmaBuffers[freebuf].paBuffer, FALSE);
        
        if (!pdx->dmaBuffers[freebuf].vaBuffer)              // If we can't create DMA buffer
        { 
          KdPrint((DRIVERNAME " - Unable to allocate common DMA buffer\n"));
          status = STATUS_UNSUCCESSFUL;      
          break;
        }

        pdx->dmaBuffers[freebuf].bufferId   = freebuf;
        pdx->dmaBuffers[freebuf].buffersize = *size;

        if (IsNativeSize)
        { 
          // Native size
          PDMABUFFER_NATIVE pdmabuffer = (PDMABUFFER_NATIVE) Irp->AssociatedIrp.SystemBuffer;
          info = sizeof(DMABUFFER_NATIVE);
          pdmabuffer->bufferId   = pdx->dmaBuffers[freebuf].bufferId;
          pdmabuffer->vaBuffer   = pdx->dmaBuffers[freebuf].vaBuffer;
          pdmabuffer->paBuffer   = pdx->dmaBuffers[freebuf].paBuffer;
          pdmabuffer->buffersize = pdx->dmaBuffers[freebuf].buffersize;
        }
        else
        { 
          // 64-bit driver called by 32-bit process
          PDMABUFFER_32 pdmabuffer32 = (PDMABUFFER_32) Irp->AssociatedIrp.SystemBuffer;
          info = sizeof(DMABUFFER_32);
          pdmabuffer32->bufferId   = pdx->dmaBuffers[freebuf].bufferId;
          pdmabuffer32->vaBuffer   = (VOID*POINTER_32)PtrToLong(pdx->dmaBuffers[freebuf].vaBuffer);
          pdmabuffer32->paBuffer   = pdx->dmaBuffers[freebuf].paBuffer;
          pdmabuffer32->buffersize = pdx->dmaBuffers[freebuf].buffersize;
        }

      } // release DMA buffer lock

      break;
    }

    case IOCTL_DMA_RELEASE:                                // Code == 0x807
    {                                                      // DMA RELEASE
      if (cbin < sizeof(ULONG))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_RELEASE) -> "
          "invalid parameter length, input expected %d, input detected %d\n", 
          sizeof(ULONG), cbin));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      PULONG index = (PULONG) Irp->AssociatedIrp.SystemBuffer;

      // Must be valid buffer
      if (*index >= DMABUFFERS)                            // V2.03
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_RELEASE) -> "
          "buffer %d does not exist\n", *index));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      { 
        // lock DMA buffers
        ExResourceExcGuard guard( &pdx->dmaBufferLock );

        // DMA buffer must be in use to be de-allocated
        if (pdx->dmaBuffers[*index].bufferId < 0)
        {
  //        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_RELEASE) -> buffer %d never allocated\n", *index));
        }
        else
        {                                                    // Release it
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_RELEASE) -> releasing buffer %d\n", *index));
          // We have to check if the DMA memory we will be releasing is used by FIFO buffers.
          // If so we need to deactivate/deallocate the FIFO buffers too.
          BOOLEAN addressInRange = FALSE;
          PUCHAR  addressStart = (PUCHAR) pdx->dmaBuffers[*index].vaBuffer;
          PUCHAR  addressEnd   = addressStart + pdx->dmaBuffers[*index].buffersize;
          PUCHAR  addressCheck = NULL;
          ULONG Loop;

          { // lock FIFO buffers (already holding DMA buffer lock)
            ExResourceExcGuard guard( &pdx->fifoBufferLock );
            for (Loop = 0; Loop < FIFOBUFFERS; Loop++)
            {
              if (pdx->fifoBuffers[Loop].bufferId >= 0)
              {
                // Check if FIFO address in range of DMA addresses
                addressInRange = FALSE;
                addressCheck = (PUCHAR)pdx->fifoBuffers[Loop].vaBuffer;

                if ((addressCheck >= addressStart) || (addressCheck <= addressEnd))
                  addressInRange = TRUE;

                addressCheck += pdx->fifoBuffers[Loop].bufferSize;

                if ((addressCheck >= addressStart) || (addressCheck <= addressEnd))
                  addressInRange = TRUE;

                // Deactivate/deallocate FIFO
                if (addressInRange == TRUE)
                {
                  // lock irq buffers 
                  //(already holding, in order of acquisition:
                  //  1. DMA buffer lock, 2. FIFO buffer lock)
                  ExResourceExcGuard guard( &pdx->irqBufferLock );                  

                  KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_RELEASE) -> "
                    "also releasing FIFO buffer %d\n", Loop));
                  
                  {
                    // synchronize with ISR the FIFO and IRQ buffer updates required to 
                    //  remove referenced DMA
                    FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );
                    pdx->isrSyncData = &Loop;
                    KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecDmaRelease, pdx );                                        
                  }

                  // now that the FIFO has been disabled, free the memory
                  ExFreePool( pdx->fifoBuffers[Loop].vaBuffer );
                }
              }
            }
          } // unlock FIFO buffers

          // Release DMA memory
          (*pdx->AdapterObject->DmaOperations->FreeCommonBuffer)
            (pdx->AdapterObject, pdx->dmaBuffers[*index].buffersize, pdx->dmaBuffers[*index].paBuffer,
            pdx->dmaBuffers[*index].vaBuffer, FALSE);
          pdx->dmaBuffers[*index].bufferId = -1;
        }

      } // release DMA buffer lock

      info = 0;
      break;
    }

    case IOCTL_DMA_READ:                                   // Code == 0x808
    {                                                      // DMA READ
      if (IsNativeSize)
      { 
        // Native size
        if (cbin < sizeof(TRANSFERBUFFER_NATIVE))
        { KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> invalid parameter length, input expected %d, input detected %d\n", sizeof(TRANSFERBUFFER_NATIVE), cbin));
          status = STATUS_INVALID_PARAMETER;
          break;
        }
        
        PTRANSFERBUFFER_NATIVE transferBuffer = 
          (PTRANSFERBUFFER_NATIVE) Irp->AssociatedIrp.SystemBuffer;

        // allocate a MDL for the user supplied buffer
        callersBufMdl = IoAllocateMdl( 
          transferBuffer->bufferTransfer + transferBuffer->bufferTargetIndex,
          transferBuffer->bufferTransferLength, FALSE, TRUE, NULL );

        if( NULL == callersBufMdl )
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> failed to allocate MDL for %0X\n", 
            transferBuffer->bufferTransfer));
          status = STATUS_IN_PAGE_ERROR;
          break;
        }

        // lock the buffer in non-paged memory
        __try 
        {
          MmProbeAndLockPages( callersBufMdl, UserMode, IoWriteAccess );
        }
        __except(EXCEPTION_EXECUTE_HANDLER)
        {
          status = STATUS_ACCESS_VIOLATION;
          IoFreeMdl( callersBufMdl );
          callersBufMdl = NULL;
          break;
        }  
        
        // Must be valid buffer
        if ((transferBuffer->bufferId < 0) || (transferBuffer->bufferId >= DMABUFFERS))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> buffer %d does not exist\n", transferBuffer->bufferId));
          status = STATUS_INVALID_PARAMETER;
          break;
        }
        
        {
          // lock dmaBuffers
          ExResourceSharedGuard guard( &pdx->dmaBufferLock );

          // DMA buffer must be in use to be read
          if (pdx->dmaBuffers[transferBuffer->bufferId].bufferId < 0)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> buffer %d to be read is not allocated\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }

          ULONG transferIndexEnd = transferBuffer->bufferSourceIndex + 
            transferBuffer->bufferTransferLength;

          // Check for size
          if (transferIndexEnd > pdx->dmaBuffers[transferBuffer->bufferId].buffersize)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> buffer %d data request exceeds available size\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }

          // Copy data to target buffer
          RtlCopyMemory( 
            (PUCHAR)MmGetSystemAddressForMdl(callersBufMdl) + transferBuffer->bufferTargetIndex,
            (PUCHAR)pdx->dmaBuffers[transferBuffer->bufferId].vaBuffer + transferBuffer->bufferSourceIndex, 
            transferBuffer->bufferTransferLength);
          
        } // unlock dmaBuffers

        info = sizeof(TRANSFERBUFFER_NATIVE);
      }
      else
      { // 64-bit driver called by 32-bit process
        
        if (cbin < sizeof(TRANSFERBUFFER_32))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> invalid parameter length, input expected %d, input detected %d\n", sizeof(TRANSFERBUFFER_32), cbin));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PTRANSFERBUFFER_32 transferBuffer = (PTRANSFERBUFFER_32) Irp->AssociatedIrp.SystemBuffer;
        
        // allocate a MDL for the user supplied buffer
        callersBufMdl = IoAllocateMdl( 
          transferBuffer->bufferTransfer + transferBuffer->bufferTargetIndex,
          transferBuffer->bufferTransferLength, FALSE, TRUE, NULL );

        if( NULL == callersBufMdl )
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> failed to allocate MDL for %0X\n", 
            transferBuffer->bufferTransfer));
          status = STATUS_IN_PAGE_ERROR;
          break;
        }

        // lock the buffer in non-paged memory
        __try 
        {
          MmProbeAndLockPages( callersBufMdl, UserMode, IoWriteAccess );
        }
        __except(EXCEPTION_EXECUTE_HANDLER)
        {
          status = STATUS_ACCESS_VIOLATION;
          IoFreeMdl( callersBufMdl );
          callersBufMdl = NULL;
          break;
        }

        // Must be valid buffer
        if ((transferBuffer->bufferId < 0) || (transferBuffer->bufferId >= DMABUFFERS))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> buffer %d does not exist\n", 
            transferBuffer->bufferId));
          status = STATUS_INVALID_PARAMETER;
          break;
        }
                
        {
          // lock dmaBuffers
          ExResourceSharedGuard guard( &pdx->dmaBufferLock );

          // DMA buffer must be in use to be read
          if (pdx->dmaBuffers[transferBuffer->bufferId].bufferId < 0)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> buffer %d to be read is not allocated\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;            
            break;
          }

          ULONG transferIndexEnd = transferBuffer->bufferSourceIndex + 
            transferBuffer->bufferTransferLength;
          
          // Check for size
          if (transferIndexEnd > pdx->dmaBuffers[transferBuffer->bufferId].buffersize)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_READ) -> buffer %d data request exceeds available size\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }
          
          // Copy data to target buffer
          RtlCopyMemory(
            (PUCHAR)MmGetSystemAddressForMdl(callersBufMdl) + transferBuffer->bufferTargetIndex,
            (PUCHAR)pdx->dmaBuffers[transferBuffer->bufferId].vaBuffer + transferBuffer->bufferSourceIndex,
            transferBuffer->bufferTransferLength);          
        }

        info = sizeof(TRANSFERBUFFER_32);
      } //unlock dmaBuffers

      break;
    }

    case IOCTL_DMA_WRITE:                                  // Code == 0x809
    {                                                      // DMA WRITE
      if (IsNativeSize)
      { 
        // Native size
        if (cbin < sizeof(TRANSFERBUFFER_NATIVE))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> invalid parameter length, input expected %d, input detected %d\n", sizeof(TRANSFERBUFFER_NATIVE), cbin));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PTRANSFERBUFFER_NATIVE transferBuffer = (PTRANSFERBUFFER_NATIVE) Irp->AssociatedIrp.SystemBuffer;

        // allocate a MDL for the user supplied buffer
        callersBufMdl = IoAllocateMdl( 
          transferBuffer->bufferTransfer + transferBuffer->bufferSourceIndex, 
          transferBuffer->bufferTransferLength, FALSE, TRUE, NULL );

        if( NULL == callersBufMdl )
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> failed to allocate MDL for %0X\n", 
            transferBuffer->bufferTransfer));
          status = STATUS_IN_PAGE_ERROR;
          break;
        }

        // lock the buffer in non-paged memory
        __try 
        {
          MmProbeAndLockPages( callersBufMdl, UserMode, IoWriteAccess );
        }
        __except(EXCEPTION_EXECUTE_HANDLER)
        {
          status = STATUS_ACCESS_VIOLATION;
          IoFreeMdl( callersBufMdl );
          callersBufMdl = NULL;
          break;
        }

        // Must be valid buffer
        if ((transferBuffer->bufferId < 0) || (transferBuffer->bufferId >= DMABUFFERS))
        { KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> buffer %d does not exist\n", transferBuffer->bufferId));
          status = STATUS_INVALID_PARAMETER;
          break;
        }
        
        { // lock dmaBuffers
          ExResourceExcGuard guard( &pdx->dmaBufferLock );

          // DMA buffer must be in use to be written
          if (pdx->dmaBuffers[transferBuffer->bufferId].bufferId < 0)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> buffer %d to be written is not allocated\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }

          ULONG transferIndexEnd = transferBuffer->bufferTargetIndex + transferBuffer->bufferTransferLength;

          // Check for size
          if (transferIndexEnd > pdx->dmaBuffers[transferBuffer->bufferId].buffersize)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> buffer %d data request exceeds available size\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }

          // Copy data to target buffer
          RtlCopyMemory(
            (PUCHAR)pdx->dmaBuffers[transferBuffer->bufferId].vaBuffer + transferBuffer->bufferTargetIndex,
            (PUCHAR)MmGetSystemAddressForMdl( callersBufMdl ), transferBuffer->bufferTransferLength);

        } // unlock dmaBuffers
      }
      else
      { 
        // 64-bit driver called by 32-bit process
        if (cbin < sizeof(TRANSFERBUFFER_32))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> invalid parameter length, input expected %d, input detected %d\n", sizeof(TRANSFERBUFFER_32), cbin));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PTRANSFERBUFFER_32 transferBuffer = (PTRANSFERBUFFER_32) Irp->AssociatedIrp.SystemBuffer;

        // allocate a MDL for the user supplied buffer
        callersBufMdl = IoAllocateMdl( 
          transferBuffer->bufferTransfer + transferBuffer->bufferSourceIndex,
          transferBuffer->bufferTransferLength, FALSE, TRUE, NULL );

        if( NULL == callersBufMdl )
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> failed to allocate MDL for %0X\n", 
            transferBuffer->bufferTransfer));
          status = STATUS_IN_PAGE_ERROR;
          break;
        }

        // lock the buffer in non-paged memory
        __try 
        {
          MmProbeAndLockPages( callersBufMdl, UserMode, IoWriteAccess );
        }
        __except(EXCEPTION_EXECUTE_HANDLER)
        {
          status = STATUS_ACCESS_VIOLATION;
          IoFreeMdl( callersBufMdl );
          callersBufMdl = NULL;
          break;
        }       

        // Must be valid buffer
        if ((transferBuffer->bufferId < 0) || (transferBuffer->bufferId >= DMABUFFERS))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> buffer %d does not exist\n", transferBuffer->bufferId));
          status = STATUS_INVALID_PARAMETER;
          break;
        }
                
        { // lock dmaBuffers
          ExResourceSharedGuard( &pdx->dmaBufferLock );

          // DMA buffer must be in use to be read
          if (pdx->dmaBuffers[transferBuffer->bufferId].bufferId < 0)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> buffer %d to be written is not allocated\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }

          ULONG transferIndexEnd = transferBuffer->bufferTargetIndex + transferBuffer->bufferTransferLength;

          // Check for size
          if (transferIndexEnd > pdx->dmaBuffers[transferBuffer->bufferId].buffersize)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_DMA_WRITE) -> buffer %d data request exceeds available size\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;           
            break;
          }

          // Copy data to target buffer
          RtlCopyMemory(
            (PUCHAR)pdx->dmaBuffers[transferBuffer->bufferId].vaBuffer + transferBuffer->bufferTargetIndex,
            (PUCHAR)MmGetSystemAddressForMdl( callersBufMdl ), transferBuffer->bufferTransferLength);

        } // unlock dmaBuffers

      } // 64-bit driver called by 32-bit process

      info = 0;
      break;
    }

    case IOCTL_FIFO_ALLOCATE:                              // Code == 0x80A
    {                                                      // FIFO ALLOCATE
      if (IsNativeSize)
      { 
        // Native size
        if ((cbin < sizeof(FIFOTRANSFERBUFFER_NATIVE)) || 
          (cbout < sizeof(FIFOTRANSFERBUFFER_NATIVE)))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_ALLOCATE) -> "
            "invalid parameter length, input expected %d, input detected %d, "
            "output expected %d, output detected %d\n", 
            sizeof(FIFOTRANSFERBUFFER_NATIVE), cbin, 
            sizeof(FIFOTRANSFERBUFFER_NATIVE), cbout));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PFIFOTRANSFERBUFFER_NATIVE transferBuffer = 
          (PFIFOTRANSFERBUFFER_NATIVE) Irp->AssociatedIrp.SystemBuffer;
        ULONG    freebuf;
        ULONG     Loop;
        BOOLEAN validAddress = FALSE;
        PUCHAR  transferAddressEnd = 
          (PUCHAR)transferBuffer->bufferTransfer + transferBuffer->bufferTransferLength;
        // We have to check if the memory we will be buffering does actually 'exist'.
        // Since the driver only allocates DMA memory besides the FIFO memory, we only
        // have to check if DMA memory has been allocated in the indicated memory range.
        Loop = 0;

        {
          // lock dmaBuffers
          ExResourceSharedGuard( &pdx->dmaBufferLock );

          // check that DMA address for this FIFO is valid
          while ((Loop < DMABUFFERS) && (validAddress == FALSE))
          { 
            if (pdx->dmaBuffers[Loop].bufferId >= 0)         // If in use (just what we need)
            { 
              if ( (pdx->dmaBuffers[Loop].vaBuffer <= transferBuffer->bufferTransfer) &&
                (((PUCHAR)pdx->dmaBuffers[Loop].vaBuffer + pdx->dmaBuffers[Loop].buffersize) >= 
                  transferAddressEnd))
                validAddress = TRUE;
            }
            Loop++;
          }

          if (validAddress == FALSE)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_ALLOCATE) -> "
              "invalid memory range (must be IOCTL_DMA_ALLOCATE allocated memory)\n"));
            status = STATUS_UNSUCCESSFUL;
            break;
          }

          // Find a number of free records for the FIFO buffer data
          freebuf = 0;

          { // lock fifoBuffers (already holding DMA buffer lock)
            ExResourceExcGuard( &pdx->fifoBufferLock );            

            // find an available block of FIFO buffers
            while ((pdx->fifoBuffers[freebuf].bufferId >= 0) && 
              ((freebuf + transferBuffer->buffers) <= FIFOBUFFERS))
              freebuf++;

            if ((freebuf + transferBuffer->buffers) > FIFOBUFFERS)
            { 
              KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_ALLOCATE) -> "
                "requested number of FIFO buffers not available\n"));
              status = STATUS_INSUFFICIENT_RESOURCES;
              break;
            }
          
            // Keep copy of first FIFO's index
            transferBuffer->bufferId = freebuf;  

            // Allocate the buffers' memory
            Escape = FALSE;
            for (Loop = 0; Loop < (ULONG)transferBuffer->buffers; Loop++)
            { 
              pdx->fifoBuffers[freebuf].vaBufferSource = transferBuffer->bufferTransfer;
              pdx->fifoBuffers[freebuf].bufferSize   = transferBuffer->bufferTransferLength;
              pdx->fifoBuffers[freebuf].vaBuffer = 
                ExAllocatePool(NonPagedPoolCacheAligned, transferBuffer->bufferTransferLength);
              pdx->fifoBuffers[freebuf].writeTag = FALSE;
              pdx->fifoBuffers[freebuf].bufferIrqsActiveReceived = 0;
              pdx->fifoBuffers[freebuf].bufferOverflows = 0;

              if (!pdx->fifoBuffers[freebuf].vaBuffer)
              {                                                // If we can't allocate FIFO buffer
                // We have to release all buffers we allocated for this call
                while (freebuf >= (ULONG)transferBuffer->bufferId)
                {
                  if (pdx->fifoBuffers[freebuf].bufferId >= 0)         // If in use
                    ExFreePool(pdx->fifoBuffers[freebuf].vaBuffer);    // Release the allocated memory
                  pdx->fifoBuffers[freebuf].bufferId = -1;
                  freebuf--;
                }

                KdPrint((DRIVERNAME " - Unable to allocate FIFO buffers\n"));
                status = STATUS_INSUFFICIENT_RESOURCES;
                Escape = TRUE;
                break;
              }
              
              pdx->fifoBuffers[freebuf].bufferId = 1;                 // Indicate in use
              freebuf++;
            }
          
            // Since the for..loop might have exit with an error we check this
            if (Escape)
              break;
   
          } // unlock fifoBuffers
        } // unlock dmaBuffers

        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_ALLOCATE) -> "
          "allocated %d buffers of %d bytes (FIFO %d)\n", 
          transferBuffer->buffers, transferBuffer->bufferTransferLength, 
          transferBuffer->bufferId));
        info = sizeof(FIFOTRANSFERBUFFER_NATIVE);
      }
      else
      { 
        // 64-bit driver called by 32-bit process
        if ((cbin < sizeof(FIFOTRANSFERBUFFER_32)) || (cbout < sizeof(FIFOTRANSFERBUFFER_32)))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_ALLOCATE) -> "
            "invalid parameter length, input expected %d, input detected %d, "
            "output expected %d, output detected %d\n", sizeof(FIFOTRANSFERBUFFER_32), 
            cbin, sizeof(FIFOTRANSFERBUFFER_32), cbout));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PFIFOTRANSFERBUFFER_32 transferBuffer = 
                                    (PFIFOTRANSFERBUFFER_32) Irp->AssociatedIrp.SystemBuffer;
        ULONG    freebuf;
        ULONG     Loop;
        BOOLEAN validAddress = FALSE;
        PUCHAR  transferAddressEnd = (PUCHAR)(PUCHAR*POINTER_32)transferBuffer->bufferTransfer + 
          transferBuffer->bufferTransferLength;
        // We have to check if the memory we will be buffering does actually 'exist'.
        // Since the driver only allocates DMA memory besides the FIFO memory, we only
        // have to check if DMA memory has been allocated in the indicates memory range.
// NOTE: THE CHECK IS DONE ON THE LOW PART SINCE THE 32-BIT PROCESS ONLY HAS THIS PART ....
        Loop = 0;

        { // lock dmaBuffers
          ExResourceSharedGuard( &pdx->dmaBufferLock );

          while ((Loop < DMABUFFERS) && (validAddress == FALSE))
          { 
            if (pdx->dmaBuffers[Loop].bufferId >= 0)         // If in use (just what we need)
            { 
              if ((PtrToLong(pdx->dmaBuffers[Loop].vaBuffer) <= 
                (LONG)transferBuffer->bufferTransfer) &&
                ((PtrToLong(pdx->dmaBuffers[Loop].vaBuffer) + pdx->dmaBuffers[Loop].buffersize) >= 
                PtrToUlong(transferAddressEnd)))
                  validAddress = TRUE;
            }
            Loop++;
          }

          if (validAddress == FALSE)
          { 
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_ALLOCATE) -> "
              "invalid memory range (must be IOCTL_DMA_ALLOCATE allocated memory)\n"));
            status = STATUS_UNSUCCESSFUL;
            break;
          }

          // Find a number of free records for the FIFO buffer data
          freebuf = 0;

          { // lock fifoBuffers (already holding DMA buffer lock)
            ExResourceExcGuard( &pdx->fifoBufferLock );

            while ((pdx->fifoBuffers[freebuf].bufferId >= 0) && 
              ((freebuf + transferBuffer->buffers) <= FIFOBUFFERS))
                freebuf++;

            if ((freebuf + transferBuffer->buffers) > FIFOBUFFERS)
            {
              KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_ALLOCATE) -> "
                "requested number of FIFO buffers not available\n"));
              status = STATUS_INSUFFICIENT_RESOURCES;
              break;
            }
        
            transferBuffer->bufferId = freebuf;  // Keep copy of first index
            Escape = FALSE;
            for (Loop = 0; Loop < (ULONG)transferBuffer->buffers; Loop++)
            { 
              pdx->fifoBuffers[freebuf].vaBufferSource = transferBuffer->bufferTransfer;
              pdx->fifoBuffers[freebuf].bufferSize     = transferBuffer->bufferTransferLength;
              pdx->fifoBuffers[freebuf].vaBuffer = 
                ExAllocatePool(NonPagedPoolCacheAligned, transferBuffer->bufferTransferLength);
              pdx->fifoBuffers[freebuf].writeTag = FALSE;
              pdx->fifoBuffers[freebuf].bufferIrqsActiveReceived = 0;
              pdx->fifoBuffers[freebuf].bufferOverflows = 0;
              
              if (!pdx->fifoBuffers[freebuf].vaBuffer)
              { 
                // If we can't allocate FIFO buffer
                // We have to release all buffers we allocated for this call
                while (freebuf >= (ULONG)transferBuffer->bufferId)
                {
                  if (pdx->fifoBuffers[freebuf].bufferId >= 0)         // If in use
                    ExFreePool(pdx->fifoBuffers[freebuf].vaBuffer);    // Release the allocated memory

                  pdx->fifoBuffers[freebuf].bufferId = -1;
                  freebuf--;
                }

                KdPrint((DRIVERNAME " - Unable to allocate FIFO buffers\n"));
                status = STATUS_INSUFFICIENT_RESOURCES;
                Escape = TRUE;
                break;
              }

              pdx->fifoBuffers[freebuf].bufferId = 1;                 // Indicate in use
              freebuf++;
            }

            // Since the for..loop might have exit with an error we check this
            if (Escape)
              break;

          } // unlock FIFO buffers          
        } // unlock DMA buffers

        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_ALLOCATE) -> "
          "allocated %d buffers of %d bytes (FIFO %d)\n", 
          transferBuffer->buffers, transferBuffer->bufferTransferLength, 
          transferBuffer->bufferId));
        info = sizeof(FIFOTRANSFERBUFFER_32);
      }
      break;
    }

    case IOCTL_FIFO_RELEASE:                               // Code == 0x80B
    {                                                      // FIFO RELEASE
      if (IsNativeSize)
      { 
        // Native size
        if ((cbin < sizeof(FIFOTRANSFERBUFFER_NATIVE)) || 
                                            (cbout < sizeof(FIFOTRANSFERBUFFER_NATIVE)))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_RELEASE) -> "
            "invalid parameter length, input expected %d, input detected %d, "
            "output expected %d, output detected %d\n", 
            sizeof(FIFOTRANSFERBUFFER_NATIVE), cbin, 
            sizeof(FIFOTRANSFERBUFFER_NATIVE), cbout));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PFIFOTRANSFERBUFFER_NATIVE transferBuffer = 
                                (PFIFOTRANSFERBUFFER_NATIVE) Irp->AssociatedIrp.SystemBuffer;

        LONG Loop;
        
        // Make sure loop indexes are within range 
        //  (if this data is returned then we also return used range)
        if (transferBuffer->bufferId < 0)
          transferBuffer->bufferId = 0;

        if (transferBuffer->bufferId >= FIFOBUFFERS)
          transferBuffer->bufferId = FIFOBUFFERS-1;

        if (transferBuffer->buffers > FIFOBUFFERS)
          transferBuffer->buffers = FIFOBUFFERS;

        if ((transferBuffer->bufferId + transferBuffer->buffers) > FIFOBUFFERS)
          transferBuffer->buffers = FIFOBUFFERS - transferBuffer->bufferId;
        
        {           
          // Note: maintain this order of lock acquisition to avoid deadlock

          // lock fifoBuffers 
          ExResourceExcGuard( &pdx->fifoBufferLock );
          {
            // lock irqBuffers
            ExResourceExcGuard( &pdx->irqBufferLock );

            {            
              // synchronize the FIFO and IRQ buffer updates required with the ISR
              //  while removing the referenced FIFO
              FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );

              LONG bufRange[] = { transferBuffer->bufferId, 
                transferBuffer->bufferId + transferBuffer->buffers };
              pdx->isrSyncData = bufRange;

              KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecFifoRelease, pdx );
            }            

            // after disabling the FIFO in the irqBuffer, release the FIFO's memory
            for( ULONG bufNum = transferBuffer->bufferId; 
              bufNum < (ULONG)(transferBuffer->bufferId + transferBuffer->buffers); ++bufNum )
            {
              if (pdx->fifoBuffers[bufNum].bufferId >= 0)
                ExFreePool(pdx->fifoBuffers[bufNum].vaBuffer);
            }            

          } // unlock irqBuffers

        } // unlock fifoBuffers

        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_RELEASE) -> releasing buffers %d to %d\n", transferBuffer->bufferId, transferBuffer->bufferId + transferBuffer->buffers - 1));
        info = sizeof(FIFOTRANSFERBUFFER_NATIVE);
      }
      else
      { 
        // 64-bit driver called by 32-bit process
        if ((cbin < sizeof(FIFOTRANSFERBUFFER_32)) || (cbout < sizeof(FIFOTRANSFERBUFFER_32)))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_RELEASE) -> invalid parameter length, input expected %d, input detected %d, output expected %d, output detected %d\n", sizeof(FIFOTRANSFERBUFFER_32), cbin, sizeof(FIFOTRANSFERBUFFER_32), cbout));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PFIFOTRANSFERBUFFER_32 transferBuffer = (PFIFOTRANSFERBUFFER_32) Irp->AssociatedIrp.SystemBuffer;
        LONG Loop;

        // Make sure loop indexes are within range (if this data is returned then we also return used range)
        if (transferBuffer->bufferId < 0)
          transferBuffer->bufferId = 0;

        if (transferBuffer->bufferId >= FIFOBUFFERS)
          transferBuffer->bufferId = FIFOBUFFERS-1;

        if (transferBuffer->buffers > FIFOBUFFERS)
          transferBuffer->buffers = FIFOBUFFERS;

        if ((transferBuffer->bufferId + transferBuffer->buffers) > FIFOBUFFERS)
          transferBuffer->buffers = FIFOBUFFERS - transferBuffer->bufferId;

        {           
          // Note: maintain this order of lock acquisition to avoid deadlock

          // lock fifoBuffers 
          ExResourceExcGuard( &pdx->fifoBufferLock );
          {
            // lock irqBuffers
            ExResourceExcGuard( &pdx->irqBufferLock );

            {            
              // synchronize the FIFO and IRQ buffer updates required with the ISR
              //  while removing the referenced FIFO
              FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );
              
              LONG bufRange[] = { transferBuffer->bufferId, 
                transferBuffer->bufferId + transferBuffer->buffers };
              pdx->isrSyncData = bufRange;

              KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecFifoRelease, pdx );
            }            

            // after disabling the FIFO in the irqBuffer, release the FIFO's memory
            for( ULONG bufNum = transferBuffer->bufferId; 
              bufNum < (ULONG)(transferBuffer->bufferId + transferBuffer->buffers); ++bufNum )
            {
              if (pdx->fifoBuffers[bufNum].bufferId >= 0)
                ExFreePool(pdx->fifoBuffers[bufNum].vaBuffer);
            }

          } // unlock irqBuffers
        } // unlock fifoBuffers

        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_RELEASE) -> releasing buffers %d to %d\n", transferBuffer->bufferId, transferBuffer->bufferId + transferBuffer->buffers - 1));
        info = sizeof(FIFOTRANSFERBUFFER_32);
      }
      break;
    }

    case IOCTL_FIFO_READ:                                  // Code == 0x80C
    {                                                      // FIFO READ      
      if (IsNativeSize)
      { 
        // Native size
        if (cbin < sizeof(FIFOTRANSFERBUFFER_NATIVE) || 
          (cbout < sizeof(FIFOTRANSFERBUFFER_NATIVE)))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) ->"
            " invalid parameter length, input expected %d, input detected %d, "
            "output expected %d, output detected %d\n", 
            sizeof(FIFOTRANSFERBUFFER_NATIVE), cbin, 
            sizeof(FIFOTRANSFERBUFFER_NATIVE), cbout));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PFIFOTRANSFERBUFFER_NATIVE transferBuffer = 
                     (PFIFOTRANSFERBUFFER_NATIVE) Irp->AssociatedIrp.SystemBuffer;
        
        // allocate a MDL for the user supplied buffer
        callersBufMdl = IoAllocateMdl( transferBuffer->bufferTransfer, 
                             transferBuffer->bufferTransferLength, FALSE, TRUE, NULL );

        if( NULL == callersBufMdl )
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
            "failed to allocate MDL for %0X\n", transferBuffer->bufferTransfer));
          status = STATUS_IN_PAGE_ERROR;
          break;
        }

        // lock the buffer in non-paged memory
        __try 
        {
          MmProbeAndLockPages( callersBufMdl, UserMode, IoWriteAccess );
        }
        __except(EXCEPTION_EXECUTE_HANDLER)
        {
          status = STATUS_ACCESS_VIOLATION;
          IoFreeMdl( callersBufMdl );
          callersBufMdl = NULL;
          break;
        }       

        // Must be valid buffer
        if ((transferBuffer->bufferId < 0) || (transferBuffer->bufferId >= FIFOBUFFERS))                        // MUST be valid index
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
            "buffer %d does not exist\n", transferBuffer->bufferId));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        { // lock fifoBuffers 
          // Note: even though a write to the FIFO is occuring when resetting the writeTag,
          //  none of the other IOCTL operations care about that data (only the ISR)
          //  thus a shared (multiple readers) lock is used here
          ExResourceSharedGuard( &pdx->fifoBufferLock );

          // Check for in use
          if (pdx->fifoBuffers[transferBuffer->bufferId].bufferId < 0)
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
              "buffer %d never allocated\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }

          // Check for size
          if (transferBuffer->bufferTransferLength 
                         < pdx->fifoBuffers[transferBuffer->bufferId].bufferSize)
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
              "target buffer to small for data FIFO buffer %d\n", 
              transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }
            
          // Copy data to target buffer          
          PVOID target = MmGetSystemAddressForMdl( callersBufMdl );

          {
            // synchronize with ISR while reading Fifo buffer
            //  (in the event of an overrun, this buffer could be simultaneously being
            //  written/overwritten by the ISR)
            FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );

            PVOID params[] = { target, transferBuffer };
            pdx->isrSyncData = params;

            KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecFifoRead, pdx );
          }

          info = sizeof(FIFOTRANSFERBUFFER_NATIVE);

        } // unlock fifoBuffers        
      }
      else
      { 
        // 64-bit driver called by 32-bit process
        if (cbin < sizeof(FIFOTRANSFERBUFFER_32) || (cbout < sizeof(FIFOTRANSFERBUFFER_32)))
        { 
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
            "invalid parameter length, input expected %d, input detected %d, "
            "output expected %d, output detected %d\n", 
            sizeof(FIFOTRANSFERBUFFER_32), cbin, 
            sizeof(FIFOTRANSFERBUFFER_32), cbout));
          status = STATUS_INVALID_PARAMETER;
          break;
        }

        PFIFOTRANSFERBUFFER_32 transferBuffer = 
                          (PFIFOTRANSFERBUFFER_32) Irp->AssociatedIrp.SystemBuffer;

        // allocate a MDL for the user supplied buffer
        callersBufMdl = IoAllocateMdl( transferBuffer->bufferTransfer, 
          transferBuffer->bufferTransferLength, FALSE, TRUE, NULL );

        if( NULL == callersBufMdl )
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
            "failed to allocate MDL for %0X\n", transferBuffer->bufferTransfer));
          status = STATUS_IN_PAGE_ERROR;
          break;
        }

        // lock the buffer in non-paged memory
        __try 
        {
          MmProbeAndLockPages( callersBufMdl, UserMode, IoWriteAccess );
        }
        __except(EXCEPTION_EXECUTE_HANDLER)
        {
          status = STATUS_ACCESS_VIOLATION;
          IoFreeMdl( callersBufMdl );
          callersBufMdl = NULL;
          break;
        }
        
        // Must be valid buffer
        if ((transferBuffer->bufferId < 0) || (transferBuffer->bufferId >= FIFOBUFFERS))                        // MUST be valid index
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
            "buffer %d does not exist\n", transferBuffer->bufferId));
          status = STATUS_INVALID_PARAMETER;
          break;
        }
        
        { // lock fifoBuffers 
          // Note: even though a write to the FIFO is occuring when resetting the writeTag,
          //  none of the other IOCTL operations care about that data (only the ISR)
          //  thus a shared (multiple readers) lock is used here
          ExResourceSharedGuard( &pdx->fifoBufferLock );

          // Check for in use
          if (pdx->fifoBuffers[transferBuffer->bufferId].bufferId < 0)
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
              "buffer %d never allocated\n", transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }
          
          // Check for size
          if (transferBuffer->bufferTransferLength < 
            pdx->fifoBuffers[transferBuffer->bufferId].bufferSize)
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_FIFO_READ) -> "
              "target buffer to small for data FIFO buffer %d\n", 
              transferBuffer->bufferId));
            status = STATUS_UNSUCCESSFUL;
            break;
          }
        
          // Copy data to target buffer          
          PVOID target = MmGetSystemAddressForMdl( callersBufMdl );

          {
            // synchronize with ISR while reading Fifo buffer
            //  (in the event of an overrun, this buffer could be simultaneously being
            //  written/overwritten by the ISR)
            FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );

            PVOID params[] = { target, transferBuffer };
            pdx->isrSyncData = params;

            KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecFifoRead, pdx );
          }

          info = sizeof(FIFOTRANSFERBUFFER_32);

        } //unlock fifoBuffers
      }
      break;
    }

    case IOCTL_IRQ_WRITE:                                  // Code == 0x80D
    {                                                      // IRQ WRITE HANDLING
      if (cbin < sizeof(IRQTRANSFERBUFFER))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
          "invalid parameter length, input expected %d, input detected %d\n", 
          sizeof(IRQTRANSFERBUFFER), cbin));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      PIRQTRANSFERBUFFER transferBuffer = 
                            (PIRQTRANSFERBUFFER) Irp->AssociatedIrp.SystemBuffer;
      
      if ((transferBuffer->IrqId < 0) || (transferBuffer->IrqId >= IRQBUFFERS))
      {
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
          "invalid IRQ ID parameter\n"));
        status = STATUS_INVALID_PARAMETER;
        break;
      }            

      // Check the parameters passed on. Only the active parameters are checked.
      if (transferBuffer->IrqInformation.active)           // Only check if setting active
      {       
        if (transferBuffer->IrqInformation.signalActive)
        {          
          if ((transferBuffer->IrqInformation.signalRegister % 4) != 0)
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
              "invalid SIGNALREGISTER parameters (not dword aligned)\n"));
            status = STATUS_INVALID_PARAMETER;
            break;
          }

          if (transferBuffer->IrqInformation.signalRegister >= pdx->memsize)
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
              "invalid SIGNALREGISTER parameters (outside range)\n"));
            status = STATUS_INVALID_PARAMETER;
            break;
          }
        }

        // The most extensive check is for the FIFO. 
        //  We must check if the FIFO buffers are available.
        if (transferBuffer->IrqInformation.fifoActive)
        {
          // First the very simple range checks
          if ((transferBuffer->IrqInformation.fifoBufferFirst < 0) || 
            (transferBuffer->IrqInformation.fifoBufferFirst > FIFOBUFFERS))
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
              "invalid FIFOBUFFERFIRST parameter\n"));
            status = STATUS_INVALID_PARAMETER;
            break;
          }

          if ((transferBuffer->IrqInformation.fifoBufferLast < 0) || 
            (transferBuffer->IrqInformation.fifoBufferLast > FIFOBUFFERS))
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
              "invalid FIFOBUFFERLAST parameter\n"));
            status = STATUS_INVALID_PARAMETER;
            break;
          }

          if ((transferBuffer->IrqInformation.fifoBufferPrevious < 0) || 
            (transferBuffer->IrqInformation.fifoBufferPrevious > FIFOBUFFERS))
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
              "invalid FIFOBUFFERPREVIOUS parameter\n"));
            status = STATUS_INVALID_PARAMETER;
            break;
          }

          if (transferBuffer->IrqInformation.fifoBufferFirst > 
                                      transferBuffer->IrqInformation.fifoBufferLast)
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
              "invalid parameters: FIFOBUFFERFIRST > FIFOBUFFERLAST\n"));
            status = STATUS_INVALID_PARAMETER;
            break;
          }

          if ((transferBuffer->IrqInformation.fifoBufferPrevious < 
                                    transferBuffer->IrqInformation.fifoBufferFirst) ||
              (transferBuffer->IrqInformation.fifoBufferPrevious > 
                                         transferBuffer->IrqInformation.fifoBufferLast))
          {
            KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
              "invalid FIFOBUFFERPREVIOUS parameter\n"));
            status = STATUS_INVALID_PARAMETER;
            break;
          }

          Escape = FALSE;
          ULONG Loop;
          
          { 
            // lock fifoBuffers
            ExResourceExcGuard( &pdx->fifoBufferLock );

            // check that FIFO to use is allocated
            for (Loop = transferBuffer->IrqInformation.fifoBufferFirst; 
              Loop <= transferBuffer->IrqInformation.fifoBufferLast; Loop++)
            {
              if (pdx->fifoBuffers[Loop].bufferId < 0)
              {
                KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) -> "
                                    "invalid FIFO parameters (FIFO not allocated)\n"));
                status = STATUS_INVALID_PARAMETER;
                Escape = TRUE;
                break;
              }
            }
          
            // In case the for..loop was exit with an error
            if (Escape)
              break;

            // Set FIFO -> IRQ links
            for (Loop = transferBuffer->IrqInformation.fifoBufferFirst; 
                          Loop <= transferBuffer->IrqInformation.fifoBufferLast; Loop++)
              pdx->fifoBuffers[Loop].irqBuffer = (UCHAR)transferBuffer->IrqId;

          } // unlock fifoBuffers

        } // requesting a FIFO        

      } // requesting this IRQ be active      
              
      // determine if the active status of an S626 irq is changing and update S626
      //  active irq list
      BOOLEAN updateActiveIrqBufList = FALSE;
      {
        // lock irqBuffers
        ExResourceExcGuard( &pdx->irqBufferLock );

        if( S626IRQNUMOFFSET <= transferBuffer->IrqId )
        {
          IRQBUFFER currentInfo = { 0 };

          // synchronize with ISR while reading irqBuffer
          FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );
          PVOID addresses[] = { &currentInfo, &(pdx->irqBuffers[transferBuffer->IrqId]) };
          pdx->isrSyncData = addresses;
          KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecCopyIrqBuffer, pdx );

          updateActiveIrqBufList = 
                                (currentInfo.active != transferBuffer->IrqInformation.active);
        }     
    
        // synchronize with ISR while writing irqBuffer
        FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );
        PVOID addresses[] = { &(pdx->irqBuffers[transferBuffer->IrqId]),
          &transferBuffer->IrqInformation };
        pdx->isrSyncData = addresses;
        KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecCopyIrqBuffer, pdx );

        // synchronize with ISR while updating pdx->s626ActiveIrqBufList
        if( updateActiveIrqBufList )
        {
          PVOID params[] = { (PVOID)transferBuffer->IrqId, 
              (PVOID)transferBuffer->IrqInformation.active };
          pdx->isrSyncData = params;
          KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecUpdateActiveIrqBufList, 
                                    pdx );        
        }
      }

      KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_WRITE) writing IRQ %d information\n", 
        transferBuffer->IrqId));
      IRQBUFFER* irqBuf = &(pdx->irqBuffers[ transferBuffer->IrqId ]);
      KdPrint(("Active=%x, EventActive=%x, FifoAct=%x, AutoDisable=%x\n", irqBuf->active, 
        irqBuf->eventActive, irqBuf->fifoActive, irqBuf->irqAutoDisable));

      info = 0;
      break;
    }

    case IOCTL_IRQ_READ:                                   // Code == 0x80E
    {                                                      // IRQ READ HANDLING
      if ((cbin < sizeof(IRQTRANSFERBUFFER)) || (cbout < sizeof(IRQTRANSFERBUFFER)))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_READ) -> "
          "invalid parameter length, input expected %d, input detected %d, "
          "output expected %d, output detected %d\n", sizeof(IRQTRANSFERBUFFER), 
          cbin, sizeof(IRQTRANSFERBUFFER), cbout));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      PIRQTRANSFERBUFFER transferBuffer = 
                                    (PIRQTRANSFERBUFFER) Irp->AssociatedIrp.SystemBuffer;
      if ((transferBuffer->IrqId < 0) || (transferBuffer->IrqId >= IRQBUFFERS))
      {
        KdPrint((DRIVERNAME 
          " - DispatchControl (IOCTL_IRQ_READ) -> invalid IRQ ID parameter\n"));
        status = STATUS_INVALID_PARAMETER;
        break;
      }
      
      {
        ExResourceSharedGuard( &pdx->irqBufferLock );

        // synchronize with ISR while reading irqBuffer
        FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );
        PVOID addresses[] = { &transferBuffer->IrqInformation, 
          &(pdx->irqBuffers[transferBuffer->IrqId]) };
        pdx->isrSyncData = (PVOID)addresses;
        KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecCopyIrqBuffer, pdx );
      }

//      KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_READ) reading IRQ %d information\n", transferBuffer->IrqId));
      info = sizeof(IRQTRANSFERBUFFER);
      break;
    }

    case IOCTL_GET_BUS_LOC:                               // Code == 0x80F
    {                                                     // PCI bus location (bus and slot number)
      // A SAA7146 READ consists of an address to read from.
      // For simplicity we use the same type for source and destination.
      if ((cbin < sizeof(ULONG)) || (cbout < sizeof(ULONG)))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_BUS_LOC) -> invalid parameter length, input expected %d, input detected %d, output expected %d, output detected %d\n", sizeof(ULONG), cbin, sizeof(ULONG), cbout));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      PULONG poffset = (PULONG) Irp->AssociatedIrp.SystemBuffer;   // Get pointer to buffer

      if (((*poffset%4) != 0) || (*poffset >= pdx->memsize))       // Address MUST be valid (4 byte boundary)
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_GET_BUS_LOC) -> invalid address (must be on 4 byte boundary)"));
        status = STATUS_INVALID_PARAMETER;
        break;
      }

      *poffset = (pdx->busNumber << 16) | (pdx->busSlotNumber & 0xFFFF);     
      info = sizeof(ULONG);
      break;
    }

    case IOCTL_IRQ_FIFOS_READ:                               // Code == 0x810
    {
      //  check buffer size
      //    first do a basic check
      if ((cbin != cbout) || (cbout < sizeof(IRQFIFOSREADBUFFER)))
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_FIFOS_READ) ->"
          " invalid parameter length, input expected %d, input detected %d, "
          "output expected %d, output detected %d\n", 
          sizeof(IRQFIFOSREADBUFFER), cbin, sizeof(IRQFIFOSREADBUFFER), cbout));
        status = STATUS_INVALID_PARAMETER; 
        break;
      }

      ULONG byteCntAssertedByBuf = sizeof(LONG) + sizeof(ULONG);
      PIRQFIFOSREADBUFFER inBuf = (PIRQFIFOSREADBUFFER)Irp->AssociatedIrp.SystemBuffer;
      PFIFOTRANSFERBUFFER_NATIVE fifoBuf = inBuf->fifos;

      KdPrint( (DRIVERNAME " - IOCTL_IRQ_FIFOS_READ inBuf=0x%04X, inBuf->fifos=0x%04X, irqId=%d\n",
        inBuf, inBuf->fifos, inBuf->IrqId ) );
      
      // verify FIFO size of the requested IRQ matches the incoming buffer's 
      //  bufferTransfer size      

      // retrieve FIFO buffer size from IRQ information 
      //  (use first buffer in FIFO range as all buffers in the range are required to
      //    have the same size)
      ULONG requiredFifoSizeForIrq = 0;
      {
        // not necessary to synchronize with the ISR because the ISR doesn't 
        //  change the IRQ's FIFO size parameter

        // synchronize in shared mode with other irqBuffer and fifoBuffer operations
        //  (lock fifo then irq buffers in that acquisition order)
        ExResourceSharedGuard( &pdx->fifoBufferLock );
        {
          ExResourceSharedGuard( &pdx->irqBufferLock );

          requiredFifoSizeForIrq = 
            pdx->fifoBuffers[pdx->irqBuffers[ inBuf->IrqId ].fifoBufferFirst].bufferSize;
        }
      }
    
      KdPrint( (DRIVERNAME " - IOCTL_IRQ_FIFOS_READ reqFifoSize=%u, fifosInCallersBuf=%u\n",
        requiredFifoSizeForIrq, inBuf->cntFifosInBuf ) );

      BOOLEAN incorrectFifoSize = FALSE;
      for( ULONG i = 0; i < inBuf->cntFifosInBuf; ++i, ++fifoBuf )
      {
        // check bufferTransfer size for this FIFO
        if( requiredFifoSizeForIrq != fifoBuf->bufferTransferLength )
        {
          KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_FIFOS_READ) ->"
            " invalid FIFO[%d] length, input expected %d, input detected %d\n", 
            i, requiredFifoSizeForIrq, fifoBuf->bufferTransferLength));
          status = STATUS_INVALID_PARAMETER; 

          incorrectFifoSize = TRUE;
          break;
        }

        // add in size of FIFO xfer. struct
        byteCntAssertedByBuf += sizeof(FIFOTRANSFERBUFFER_NATIVE);

        // add in size of "bufferTransfer" buffer for this FIFO        
        byteCntAssertedByBuf += fifoBuf->bufferTransferLength;
      }

      // abort IoControl operation due to incorrect bufferTransferLength 
      //  in one of the FIFO buffer xfer. structures provided by caller
      if( incorrectFifoSize ) break;

      // check total incoming/outgoing buffer size against sizes embedded in
      //  buffer (and verified by IRQ's actual FIFO buffer sources for this
      //  read operation
      if (cbin != byteCntAssertedByBuf)
      { 
        KdPrint((DRIVERNAME " - DispatchControl (IOCTL_IRQ_FIFOS_READ) ->"
          " invalid total parameter length, input expected %d, input detected %d\n", 
          byteCntAssertedByBuf, cbin));
        status = STATUS_INVALID_PARAMETER; 
        break;
      }
      
      //typedef struct _IRQFIFOSREADBUFFER
      //{
      //LONG             IrqId;
      //ULONG            cntFifosInBuf;
      //FIFOTRANSFERBUFFER_NATIVE fifos[1];
      //} IRQFIFOSREADBUFFER, *PIRQFIFOSREADBUFFER;
      ULONG bufferBytesUsed, unReadBytes = 0;
      {
        ExResourceSharedGuard( &pdx->irqBufferLock );

        // synchronize with ISR while reading fifoBuffers
        FastMutexGuard fmGuard( &pdx->isrSyncSetupLock );
        PVOID params[] = { inBuf, 0 };
        pdx->isrSyncData = (PVOID)params;
        
        KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecReadIrqFifoBuffers, pdx );
        
        bufferBytesUsed = (ULONG)params[0];
        unReadBytes = (ULONG)params[1];
      }
      
      info = bufferBytesUsed;      
      status = ( 0 < unReadBytes) ? STATUS_MORE_ENTRIES : STATUS_SUCCESS;
      break;
    }
 
    default:
      // It's a good idea to return a meaningful status code if you're given an IOCTL
      // operation you don't understand.
      KdPrint((DRIVERNAME " - DispatchControl - IOCTL not recognized $%x\n", code));
      status = STATUS_INVALID_DEVICE_REQUEST;
      break;
  }

   // release user supplied buffer MDL
  if( NULL != callersBufMdl )
  {
    MmUnlockPages( callersBufMdl );
    IoFreeMdl( callersBufMdl );
    callersBufMdl = NULL;
  }

  IoReleaseRemoveLock(&pdx->RemoveLock, Irp);
  return status == STATUS_PENDING ? status : CompleteRequest(Irp, status, info);
}


//{ **************************************************************************** }
//{ Params   : <fdo>     Functional device object                                }
//{            <Irp>     I/O request packet                                      }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : Dispatcher of IRP_MJ_INTERNAL_DEVICE_CONTROL requests.            }
//{            Drivers sometimes need to talk to each other too, and they use the}
//{            related IRP_MJ_INTERNAL_DEVICE_CONTROL to do so.                  }
//{            A typical code sequence which calls this is as follows:           }
//{             ASSERT(KeGetCurrentIrql() == PASSIVE_LEVEL);                     }
//{             KEVENT event;                                                    }
//{             KeInitializeEvent(&event, NotificationEvent, FALSE);             }
//{             IO_STATUS_BLOCK iostatus;                                        }
//{             PIRP Irp = IoBuildDeviceIoControlRequest(IoControlCode,          }
//{               DeviceObject, pInBuffer, cbInBuffer, pOutBuffer, cbOutBuffer,  }
//{               TRUE, &event, &iostatus);                                      }
//{             if (IoCallDriver(DeviceObject, Irp) == STATUS_PENDING)           }
//{               KeWaitForSingleObject(&event, Executive, KernelMode, FALSE,    }
//{                 NULL);                                                       }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
NTSTATUS DispatchInternalControl(PDEVICE_OBJECT fdo, PIRP Irp)
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
  IoSkipCurrentIrpStackLocation(Irp);
  return IoCallDriver(pdx->LowerDeviceObject, Irp);
}


//{ **************************************************************************** }
//{ Params   : <pdx>     Device extension                                        }
//{            <status>  Status to pass through                                  }
//{ Return   : -                                                                 }
//{                                                                              }
//{ Descript : Abort pending IOCTL's.                                            }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma PAGEDCODE
VOID AbortPendingIoctls(PDEVICE_EXTENSION pdx, NTSTATUS status)
{
  PAGED_CODE();
  InterlockedExchange(&pdx->IoctlAbortStatus, status);
  CleanupControlRequests(pdx, status, NULL);
}


//{ **************************************************************************** }
//{ Params   : <pdx>     Device extension                                        }
//{            <Irp>     I/O request packet to cache                             }
//{            <pIrp>    I/O request packet to cache it in                       }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : Cache a control request.                                          }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
NTSTATUS CacheControlRequest(PDEVICE_EXTENSION pdx, PIRP Irp, PIRP* pIrp)
{
  ASSERT(KeGetCurrentIrql() <= DISPATCH_LEVEL);

  KIRQL oldirql;
  // We use a spin lock to guard the list of pending IOCTLs and also to guard all
  // of the pointer cells that are reserved to point to the current instance of each
  // different type of asynchronous IOCTL request.
  KeAcquireSpinLock(&pdx->IoctlListLock, &oldirql);

  NTSTATUS status;
  // This is where we enforce the ruleit's more of a design decision, really that
  // only one IRP of each type can be outstanding at one time.
  if (*pIrp)
    status = STATUS_UNSUCCESSFUL;                          // Something already cached here
  else
       // This if statement accommodates the fact that we may need to start failing
       // incoming IRPs at some point because of PnP or power events.
       if (pdx->IoctlAbortStatus)
         status = pdx->IoctlAbortStatus;                   // Rejecting new IRPs for some reason
       else
       {                                                   // Try to cache IRP
         // Since we'll pend this IRP for what might be a long time, we need to
         // have a cancel routine for it.
         IoSetCancelRoutine(Irp, OnCancelPendingIoctl);
         if (Irp->Cancel && IoSetCancelRoutine(Irp, NULL))
           status = STATUS_CANCELLED;                      // Already cancelled
                                                           // Put this IRP on our list of pending IOCTLs. Install a completion
                                                           // routine to nullify the cache pointer. Note that AddDevice would have
                                                           // failed if there were no PDO below us, so we know there's at least
                                                           // one free stack location here.
         else
         {                                                 // If to cache it
           // Here, we've decided to go ahead and cache this IRP so that we can complete
           // it later. Since we're going to end up returning STATUS_PENDING from our
           // DispatchControl function, we need to call IoMarkIrpPending.
           IoMarkIrpPending(Irp);
           status = STATUS_PENDING;

           // We need to have a way to NULL out the cache pointer cell when we cancel
           // the IRP. It's very difficult to get context parameters into a cancel
           // routine, so we decide to set up an I/O completion routine instead.
           // We use the Parameters.Others.Argument1 slot in the stack to record
           // the cache pointer address.
           PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
           //KdPrint( (DRIVERNAME " - CacheControlRequest stack[0] 0x%08X\n", stack) );
           stack->Parameters.Others.Argument1 = (PVOID) pIrp;
           
           IoCopyCurrentIrpStackLocationToNext( Irp ); // need output buffer available in ISR
           
           IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE) OnCompletePendingIoctl, 
             (PVOID) pdx, TRUE, TRUE, TRUE);
           // In order for the completion routine we've just installed to get called,
           // we must advance the I/O stack pointer by calling IoSetNextIrpStackLocation.
           // In this particular driver, we know there must be at least one more stack
           // location for us to use because our AddDevice function would have failed
           // if there hadn't been a driver object underneath ours. The device and file
           // object pointers that later routines need come from the then-current stack
           // location, so we must initialize them as well.
           IoSetNextIrpStackLocation(Irp);                 // So our completion routine will get called
           PFILE_OBJECT fop = stack->FileObject;
           stack = IoGetCurrentIrpStackLocation(Irp);
           stack->DeviceObject = pdx->DeviceObject;        // So IoCancelIrp can give us right ptr
           stack->FileObject   = fop;                      // For cleanup
           //KdPrint( (DRIVERNAME " - CacheControlRequest stack[1] 0x%08X\n", stack) );

           // make sure it's in the list before allowing the ISR to see it
           // in the event an ISR appears before list insertion and tries
           // to find it in the list 
           //*pIrp = Irp;
           InsertTailList(&pdx->PendingIoctlList, &Irp->Tail.Overlay.ListEntry);

           // synchronize with ISR
           PVOID params[] = { pIrp, Irp, NULL };
           pdx->isrSyncExcPtrData = params;
           KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecExchangePtr, pdx );
         }
       }

  KeReleaseSpinLock(&pdx->IoctlListLock, oldirql);
  return status;
}


//{ **************************************************************************** }
//{ Params   : <pdx>     Device extension                                        }
//{            <status>  Status to pass through                                  }
//{            <fop>     File object                                             }
//{ Return   : -                                                                 }
//{                                                                              }
//{ Descript : Cleanup control request.                                          }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma PAGEDCODE
VOID CleanupControlRequests(PDEVICE_EXTENSION pdx, NTSTATUS status, PFILE_OBJECT fop)
{
  LIST_ENTRY cancellist;
  InitializeListHead(&cancellist);

  // Create a list of IRPs that belong to the same file object
  KIRQL oldirql;
  KeAcquireSpinLock(&pdx->IoctlListLock, &oldirql);

  PLIST_ENTRY first = &pdx->PendingIoctlList;
  PLIST_ENTRY next;
  for (next = first->Flink; next != first; )
  {                                                        // For each queued IRP
    PIRP Irp = CONTAINING_RECORD(next, IRP, Tail.Overlay.ListEntry);
    PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);

    // Follow the chain to the next IRP now (so that the next iteration of
    // the loop is properly setup whether we dequeue this IRP or not)
    PLIST_ENTRY current = next;
    next = next->Flink;

    // Skip this IRP if it's not for the same file object as the
    // current IRP_MJ_CLEANUP.
    if (fop && stack->FileObject != fop)
      continue;                                            // Not for same file object

    // Set the CancelRoutine pointer to NULL and remove the IRP from the
    // queue.
    if (!IoSetCancelRoutine(Irp, NULL))
      continue;                                            // Being cancelled right this instant
    RemoveEntryList(current);
    InsertTailList(&cancellist, current);
  }

  // Release the spin lock. We're about to undertake a potentially time-consuming
  // operation that might conceivably result in a deadlock if we keep the lock.
  KeReleaseSpinLock(&pdx->IoctlListLock, oldirql);

  // Complete the selected requests.
  while (!IsListEmpty(&cancellist))
  {                                                        // Cancel selected requests
    next = RemoveHeadList(&cancellist);
    PIRP Irp = CONTAINING_RECORD(next, IRP, Tail.Overlay.ListEntry);
    Irp->IoStatus.Status = STATUS_CANCELLED;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
  }
}


//{ **************************************************************************** }
//{ Params   : <fdo>  Functional device object                                   }
//{            <Irp>  I/O request packet                                         }
//{ Return   : -                                                                 }
//{                                                                              }
//{ Descript : On cancel callback of pending IOCTL.                              }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
VOID OnCancelPendingIoctl(PDEVICE_OBJECT fdo, PIRP Irp)
{
  KIRQL oldirql = Irp->CancelIrql;
  IoReleaseCancelSpinLock(DISPATCH_LEVEL);
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

  // Remove the IRP from whatever queue it's on
  KeAcquireSpinLockAtDpcLevel(&pdx->IoctlListLock);
  RemoveEntryList(&Irp->Tail.Overlay.ListEntry);
  KeReleaseSpinLock(&pdx->IoctlListLock, oldirql);

  // Complete the IRP
  Irp->IoStatus.Status = STATUS_CANCELLED;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);
}


//{ **************************************************************************** }
//{ Params   : <junk>    Functional device object                                }
//{            <Irp>     I/O request packet                                      }
//{            <pdx>     Device extension                                        }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : On complete callback of pending IOCTL.                            }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
NTSTATUS OnCompletePendingIoctl(PDEVICE_OBJECT junk, PIRP Irp, PDEVICE_EXTENSION pdx)
{
  KIRQL oldirql;
  KeAcquireSpinLock(&pdx->IoctlListLock, &oldirql);
  PIRP* pIrp = (PIRP*) IoGetCurrentIrpStackLocation(Irp)->Parameters.Others.Argument1;
  // This statement is the point of installing a completion routine. If the IRP gets
  // cancelled, we'll eventually gain control to nullify the cache pointer.
  if (*pIrp == Irp)
    *pIrp = NULL;
  KeReleaseSpinLock(&pdx->IoctlListLock, oldirql);          

  // restore OutputBufferLength to this Irp stack location
  PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
  stack->Parameters.DeviceIoControl.OutputBufferLength = 
      (ULONG) stack->Parameters.Others.Argument2;

  //KdPrint( (DRIVERNAME " - OnCompletePendingIoctl stack 0x%08X, "
  //  "outLen %d, buffer: 0x%08X, saved len %d\n", 
  //  stack, stack->Parameters.DeviceIoControl.OutputBufferLength, 
  //  Irp->AssociatedIrp.SystemBuffer, stack->Parameters.Others.Argument2) );

  //if( sizeof(S626IRQSTATUS) == stack->Parameters.DeviceIoControl.OutputBufferLength )
  //{
  //  PS626IRQSTATUS status = (PS626IRQSTATUS) Irp->AssociatedIrp.SystemBuffer;
  //  KdPrint( (DRIVERNAME " - OnCompletePendingIoctl %04X, %04X, %04X, %04X\n", 
  //    status->intrStatusWord[0], status->intrStatusWord[1],
  //    status->intrStatusWord[2], status->intrStatusWord[3]) );
  //}

  return STATUS_SUCCESS;
}


//{ **************************************************************************** }
//{ Params   : <pdx>     Device extension                                        }
//{            <pIrp>    I/O request packet                                      }
//{ Return   : <Result>  I/O request packet                                      }
//{                                                                              }
//{ Descript : Uncache control request.                                          }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
PIRP UncacheControlRequest(PDEVICE_EXTENSION pdx, PIRP* pIrp)
{
  ASSERT(KeGetCurrentIrql() <= DISPATCH_LEVEL);

  KIRQL oldirql;
  KeAcquireSpinLock(&pdx->IoctlListLock, &oldirql);

  // In the normal course of events, this statement uncaches an IRP
  //  PIRP Irp = (PIRP) InterlockedExchangePointer(pIrp, NULL);
  //PIRP Irp = (PIRP) InterlockedExchangePointer((volatile PVOID *)pIrp, NULL);  // V2.03  

  // synchronize with ISR
  PVOID params[] = { pIrp, NULL, NULL };
  pdx->isrSyncExcPtrData = params;
  KeSynchronizeExecution( pdx->InterruptObject, IsrSynCritSecExchangePtr, pdx );

  PIRP Irp = (PIRP)params[2];
  if (Irp)
  {                                                        // If an IRP was cached
    // Now that we've uncached our IRP, we don't want it to be cancelled any more.
    // If IoSetCancelRoutine returns NULL, however, we know that this IRP is currently
    // in the process of being cancelled. We return a NULL IRP pointer in that case.

    // Clear the cancel pointer for this IRP. Since both we and the
    // completion routine use a spin lock, it cannot happen that this
    // IRP pointer is suddenly invalid but the cache pointer cell
    // wasn't already NULL.
    if (IoSetCancelRoutine(Irp, NULL))
    {
      RemoveEntryList(&Irp->Tail.Overlay.ListEntry);       // N.B.: a macro!!
    }
    else
      Irp = NULL;                                          // If currently being cancelled
  }

  KeReleaseSpinLock(&pdx->IoctlListLock, oldirql);
  return Irp;
}

// ISR SyncCriticalSections - critical sections executed at DIRQL to synchronize with ISR

BOOLEAN IsrSynCritSecDmaRelease( PVOID context )
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) context;
  const ULONG bufNum = *((PULONG)pdx->isrSyncData);

  pdx->fifoBuffers[bufNum].bufferId = -1;
    
  // Disable IRQ link -> this means that even other FIFO's which are still valid are disabled!
  pdx->irqBuffers[ pdx->fifoBuffers[bufNum].irqBuffer ].fifoActive = FALSE;
  
  return TRUE;
}

BOOLEAN IsrSynCritSecFifoRelease( PVOID context )
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION)context;
  const PULONG bufRange = (const PULONG)pdx->isrSyncData;

  ASSERT( (bufRange[1] > bufRange[0]) && 
    (bufRange[1] - bufRange[0] <= FIFOBUFFERS) );

  // release block of FIFO buffers
  for (ULONG bufNum = bufRange[0]; bufNum < bufRange[1]; ++bufNum)
  {   
    pdx->fifoBuffers[bufNum].bufferId = -1;

    // Disable IRQ link -> this means that even other FIFO's which are still valid are disabled!
    pdx->irqBuffers[ pdx->fifoBuffers[bufNum].irqBuffer ].fifoActive = FALSE;    
  } 

  return TRUE;
}

BOOLEAN IsrSynCritSecFifoRead( PVOID context )
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) context;
  PVOID* params = (PVOID*) pdx->isrSyncData;
  // TODO: support PFIFOTRANSFERBUFFER_32 as well
  PFIFOTRANSFERBUFFER_NATIVE transferBuffer = (PFIFOTRANSFERBUFFER_NATIVE) params[1];

  RtlCopyMemory( params[0], pdx->fifoBuffers[transferBuffer->bufferId].vaBuffer, 
    pdx->fifoBuffers[transferBuffer->bufferId].bufferSize);
        
  // Reset write tag and update returned data
  transferBuffer->bufferTransferLength = 
    pdx->fifoBuffers[transferBuffer->bufferId].bufferSize;
  transferBuffer->bufferOrder          = 
    pdx->fifoBuffers[transferBuffer->bufferId].bufferIrqsActiveReceived;
  transferBuffer->bufferWritten        = 
    pdx->fifoBuffers[transferBuffer->bufferId].writeTag;
  transferBuffer->bufferOverflows      = 
    pdx->fifoBuffers[transferBuffer->bufferId].bufferOverflows;

  transferBuffer->bufferIrqs           = 
    pdx->irqBuffers[pdx->fifoBuffers[transferBuffer->bufferId].irqBuffer].irqsActiveReceived;
  transferBuffer->bufferIrqOverflows   = 
    pdx->irqBuffers[pdx->fifoBuffers[transferBuffer->bufferId].irqBuffer].fifoOverflows;

  pdx->fifoBuffers[transferBuffer->bufferId].writeTag = FALSE;  

  return TRUE;
}

BOOLEAN IsrSynCritSecReadIrqFifoBuffers( PVOID context )
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) context;
  PIRQFIFOSREADBUFFER inBuf = (PIRQFIFOSREADBUFFER)( ((PVOID*)pdx->isrSyncData)[0] );
  PUCHAR targetTransferBuffer = (PUCHAR)inBuf + (sizeof(IRQFIFOSREADBUFFER) + 
    ((inBuf->cntFifosInBuf - 1) * sizeof(FIFOTRANSFERBUFFER_NATIVE)));

  // save caller's params for restoration at routine end
  PVOID callersParams = pdx->isrSyncData;

  // figure out range of FIFOs to copy (of course only copy up to the quantity 
  //  of transferBuffers provided by the caller if less than the number waiting)
  PIRQBUFFER irq = &(pdx->irqBuffers[inBuf->IrqId]);

  KdPrint( (DRIVERNAME " - ReadFifos: Previous=%d, Last=%d, First=%d\n",
    irq->fifoBufferPrevious, irq->fifoBufferLast, irq->fifoBufferFirst ) );

  // start with most recent buffer written and work backwards until non-written
  //  buffer in the range is found
  ULONG fifoBufId = irq->fifoBufferPrevious;
  ULONG numFifoBufsForIrq = irq->fifoBufferLast - irq->fifoBufferFirst + 1;
  ULONG numFifoBufsReadyToRead = 0;
  for(; (numFifoBufsReadyToRead < numFifoBufsForIrq) 
            && pdx->fifoBuffers[fifoBufId].writeTag; ++numFifoBufsReadyToRead )
  {
    //KdPrint( (DRIVERNAME " - ReadFifos: find waiting start, %d writeTag=%u\n", 
    //  fifoBufId, pdx->fifoBuffers[fifoBufId].writeTag ) );

    if( irq->fifoBufferFirst == fifoBufId )
    {
      fifoBufId = irq->fifoBufferLast;
    }
    else
    {
      --fifoBufId;
    }
  }

  //KdPrint( (DRIVERNAME " - ReadFifos: Last Non-Written=%d\n", fifoBufId) );

  // move forward to the first waiting buffer
  if( !pdx->fifoBuffers[fifoBufId].writeTag )
  {
    fifoBufId = 
      (fifoBufId == irq->fifoBufferLast) ? irq->fifoBufferFirst : ++fifoBufId;
  }

  KdPrint( (DRIVERNAME " - ReadFifos: First Waiting=%d\n", fifoBufId) );

  for( ULONG i = 0; 
    (i < inBuf->cntFifosInBuf) && pdx->fifoBuffers[fifoBufId].writeTag; ++i )
  {
    // set FIFO buffer to read
    inBuf->fifos[i].bufferId = fifoBufId;
    
    PVOID params[] = { targetTransferBuffer, &(inBuf->fifos[i]) };
    pdx->isrSyncData = (PVOID) params;

    KdPrint( (DRIVERNAME " - ReadFifos: Reading Fifo=%d\n", inBuf->fifos[i].bufferId) );

    // read FIFO
    IsrSynCritSecFifoRead( context );
    --numFifoBufsReadyToRead;

    // move to next source FIFO buffer
    fifoBufId = 
      (fifoBufId == irq->fifoBufferLast) ? irq->fifoBufferFirst : ++fifoBufId;

    // move to start of next transferBuffer
    targetTransferBuffer += inBuf->fifos[i].bufferTransferLength;    
  }

  // restore caller's params
  pdx->isrSyncData = callersParams;

  // return number of bytes used in caller's buffer  
  ((PVOID*)pdx->isrSyncData)[0] = (PVOID)(targetTransferBuffer - (PUCHAR)inBuf);
  
  //KdPrint( (DRIVERNAME " - ReadFifos: Total bytes read=%d, 0x%0X, numFifosRemaining=%d\n", 
  //  (targetTransferBuffer - (PUCHAR)inBuf), (targetTransferBuffer - (PUCHAR)inBuf),
  //  numFifoBufsReadyToRead ) );

  // return number of bytes of FIFO data unread
  ((PVOID*)pdx->isrSyncData)[1] = 
    (PVOID)((ULONG)(numFifoBufsReadyToRead * inBuf->fifos[0].bufferTransferLength ));

  return TRUE;
}

BOOLEAN IsrSynCritSecSetS626Active( PVOID context )
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) context;
  const BOOLEAN activeState = (BOOLEAN) pdx->isrSyncData;

  pdx->irqBuffers[SAA7146A_S626_INT_NUM].active = activeState;

  return  TRUE;
}

BOOLEAN IsrSynCritSecCopyIrqBuffer( PVOID context )
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) context;
  PVOID* addresses = (PVOID*) pdx->isrSyncData;
 
  RtlCopyMemory( addresses[0], addresses[1], sizeof(IRQBUFFER) );

  return TRUE;
}

BOOLEAN IsrSynCritSecExchangePtr( PVOID context )
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) context;
  PVOID* params = (PVOID*) pdx->isrSyncExcPtrData;
  params[2] = InterlockedExchangePointer( params[0], params[1] );
  return TRUE;
}

BOOLEAN IsrSynCritSecUpdateActiveIrqBufList( PVOID context )
{ 
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) context;
  PVOID* params = (PVOID*) pdx->isrSyncData;

  const LONG irqId = (const LONG)(params[0]);
  const BOOLEAN active = (const BOOLEAN)(params[1]);

  if( active )
  {
    // append to list
    pdx->s626ActiveIrqBufList[pdx->s626ActiveIrqBufCnt++] = irqId;
  }
  else
  {
    // find and remove
    unsigned int idxToRemove = 0;
    while( idxToRemove < pdx->s626ActiveIrqBufCnt )
    {
      if( irqId == pdx->s626ActiveIrqBufList[idxToRemove] ) break;
      ++idxToRemove;
    }
    if( idxToRemove < pdx->s626ActiveIrqBufCnt )
    {
      // shift remaining entries 1 position towards list start
      while( (++idxToRemove) < pdx->s626ActiveIrqBufCnt )
      {
        pdx->s626ActiveIrqBufList[idxToRemove - 1] = 
          pdx->s626ActiveIrqBufList[idxToRemove];        
      }
      // update list count
      --pdx->s626ActiveIrqBufCnt;
    }
    else
    {
      // error, it should have been located in the list
      KdPrint( (DRIVERNAME " - Failed to find irqId=%d in list\n", irqId ) );
    }
  }

  if( (0 == pdx->s626ActiveIrqBufCnt) || (1 == pdx->s626ActiveIrqBufCnt) )
  {    
    // ensure SAA7146A interrupt for S626 is set up correctly    
    pdx->isrSyncData = (PVOID)pdx->s626ActiveIrqBufCnt;
    IsrSynCritSecSetS626Active( pdx );                
    pdx->irqBuffers[SAA7146A_S626_INT_NUM].irqAutoDisable = FALSE;
    
    if( pdx->s626ActiveIrqBufCnt )
    {
      //KdPrint( (DRIVERNAME " - activated S626 master interrupt\n" ) );
    }
    else
    {
      //KdPrint( (DRIVERNAME " - de-activated S626 master interrupt\n" ) );
    }
  }                         

  //KdPrint( (DRIVERNAME " - updated S626_AIB list, cnt=%d\n", pdx->s626ActiveIrqBufCnt) );
  
  for( long i = pdx->s626ActiveIrqBufCnt; 0 < i; )
  {
    --i;
    //KdPrint( (DRIVERNAME " - Entry [%d], %d\n", i, pdx->s626ActiveIrqBufList[i] ) );
  }
  return TRUE;
}