//{ **************************************************************************** }
//{ FileName............: READWRITE.CPP                                          }
//{ Project.............: SAA7146A                                               }
//{ Author(s)...........: M.Majoor                                               }
//{ Original contents...: Walter Oney ('Programming the Windows Driver Model')   }
//{ Version.............: 3.00                                                   }
//{ ---------------------------------------------------------------------------- }
//{ Read/Write request processors for SAA7146A driver.                           }
//{                                                                              }
//{ Version  Date      Comment                                                   }
//{ 1.00     20010818  - Initial release                                         }
//{ 1.01     20020917  - Added 'dd.Dma64BitAddresses = FALSE;'                   }
//{ 2.00     20020930  - FIFO/IRQ buffers initialization/finalization added      }
//{ 2.01     20040514  - EnableData starting value should have been $FFFFFFFE    }
//{ 2.02     20050408  - No changes                                              }
//{ 3.00     20060624  - No changes                                              }
//{ **************************************************************************** }


#include "stddcls.h"
#include "driver.h"
#include "perf.h"

// DEBI command constants.
#define DEBI_CMD_SIZE16			( 2 << 17 )			// Transfer size is 2 bytes.
#define DEBI_CMD_READ			  0x00010000			// Read operation.
#define DEBI_CMD_WRITE			0x00000000			// Write operation.

#define DEBI_CMD_RDWORD			( DEBI_CMD_READ  | DEBI_CMD_SIZE16 )	// Read immediate 2 bytes.
#define DEBI_CMD_WRWORD			( DEBI_CMD_WRITE | DEBI_CMD_SIZE16 )	// Write immediate 2 bytes.

// Master Control Register 2 (MC2) constants.
#define MC2_UPLD_DEBI			0x0002			// Upload DEBI.

// Primary Status Register (PSR) constants.
#define PSR_DEBI_S				0x00080000      // DEBI status flag.

// PCI BUS (SAA7146) REGISTER ADDRESS OFFSETS ////////////////////////////////////
#define P_DEBICMD         0x0080			// DEBI command.
#define P_DEBIAD          0x0088			// DEBI target address.
#define P_MC2             0x0100			// Master control 2.
#define P_PSR             0x0110			// Primary status.

// LOCAL BUS (GATE ARRAY) REGISTER ADDRESS OFFSETS /////////////////////////////////
#define LP_RDDINA				  0x0040			//  Read digital input.												
#define LP_WRCAPSELA			0x0046			//  Write A capture enable.
#define LP_RDCAPFLGA			0x0048			//  Read edges captured.
#define LP_RDINTSELA			0x004A			//  Read interrupt enable register.
#define LP_RDEDGSELA			0x004C			//  Read edge selection register.
#define LP_RDCAPSELA			0x004E			//  Read capture enable register.
#define LP_RDDINB				  0x0050			//  Read digital input.
#define LP_WRCAPSELB			0x0056			//  Write B capture enable.
#define LP_RDCAPFLGB			0x0058			//  Read edges captured.
#define LP_RDINTSELB			0x005A			//  Read interrupt enable register.
#define LP_RDEDGSELB			0x005C			//  Read edge selection register.
#define LP_RDCAPSELB			0x005E			//  Read capture enable register.
#define LP_RDDINC				  0x0060			//  Read digital input.
#define LP_WRCAPSELC			0x0066			//  Write C capture enable.
#define LP_RDCAPFLGC			0x0068			//  Read edges captured.
#define LP_RDINTSELC			0x006A			//  Read interrupt enable register.
#define LP_RDEDGSELC			0x006C			//  Read edge selection register.
#define LP_RDCAPSELC			0x006E			//  Read capture enable register.

												  // Counter Registers (read/write):
#define LP_CR0A					  0x0000			//  0A setup register.
#define LP_CR0B					  0x0002			//  0B setup register.
#define LP_CR1A					  0x0004			//  1A setup register.
#define LP_CR1B					  0x0006			//  1B setup register.
#define LP_CR2A					  0x0008			//  2A setup register.
#define LP_CR2B					  0x000A			//  2B setup register.

												  // Miscellaneous Registers (read/write):
#define LP_RDMISC2				0x0082			//  Read Misc2.
#define LP_MISC1				  0x0088			//  Read/write Misc1.

// Bit masks for MISC1 register that are the same for reads and writes.
#define MISC1_EDCAP				0x1000			// Enable edge capture on DIO chans specified by LP_WRCAPSELx.
#define MISC1_NOEDCAP			0x0000			// Disable edge capture on specified DIO chans.

										// Bit field positions in CRA:
#define CRABIT_INDXSRC_B		14		//   B index source.
#define CRABIT_CLKSRC_B			12		//   B clock source.
#define CRABIT_INDXPOL_A		11		//   A index polarity.
#define CRABIT_LOADSRC_A		 9		//   A preload trigger.
#define CRABIT_CLKMULT_A		 7		//   A clock multiplier.
#define CRABIT_INTSRC_A			 5		//   A interrupt source.
#define CRABIT_CLKPOL_A			 4		//   A clock polarity.
#define CRABIT_INDXSRC_A		 2		//   A index source.
#define CRABIT_CLKSRC_A			 0		//   A clock source.

										// Bit field positions in CRB:
#define CRBBIT_INTRESETCMD		15		//   Interrupt reset command.
#define CRBBIT_INTRESET_B		14		//   B interrupt reset enable.
#define CRBBIT_INTRESET_A		13		//   A interrupt reset enable.
#define CRBBIT_CLKENAB_A		12		//   A clock enable.
#define CRBBIT_INTSRC_B			10		//   B interrupt source.
#define CRBBIT_LATCHSRC			 8		//   A/B latch source.
#define CRBBIT_LOADSRC_B		 6		//   B preload trigger.
#define CRBBIT_CLKMULT_B		 3		//   B clock multiplier.
#define CRBBIT_CLKENAB_B		 2		//   B clock enable.
#define CRBBIT_INDXPOL_B		 1		//   B index polarity.
#define CRBBIT_CLKPOL_B			 0		//   B clock polarity.

// Write event capture enables.
static const unsigned short WRCapSel[]	= { LP_WRCAPSELA, LP_WRCAPSELB, LP_WRCAPSELC };	

// CRA register address to use.
static const unsigned short AdrsCRA[] = { LP_CR0A, LP_CR1A, LP_CR2A, LP_CR0A, LP_CR1A, LP_CR2A };

// CRB register address to use.
static const unsigned short AdrsCRB[] = { LP_CR0B, LP_CR1B, LP_CR2B, LP_CR0B, LP_CR1B, LP_CR2B };

// Counter overflow/index event flag masks for RDMISC2.
#define INDXMASK(C)		( 1 << ( ( (C) > 2 ) ? ( (C) * 2 - 1 ) : ( (C) * 2 +  4 ) ) )
#define OVERMASK(C)		( 1 << ( ( (C) > 2 ) ? ( (C) * 2 + 5 ) : ( (C) * 2 + 10 ) ) )
#define EVBITS(C)		{ 0, OVERMASK(C), INDXMASK(C), OVERMASK(C) | INDXMASK(C) }

// Translation table to map IntSrc into equivalent RDMISC2 event flag bits.
static const unsigned short EventBits[][4] = { EVBITS(0), EVBITS(1), EVBITS(2), EVBITS(3), EVBITS(4), EVBITS(5) };



//{ **************************************************************************** }
//{ Descript : Forwards.                                                         }
//{ **************************************************************************** }
VOID                 OnCancelReadWrite(PDEVICE_OBJECT fdo, PIRP Irp);
IO_ALLOCATION_ACTION AdapterControl   (PDEVICE_OBJECT fdo, PIRP Irp, PVOID regbase, PDEVICE_EXTENSION pdx);

// DEBI (Device Extension Bus Interface, the ISA side of the SAA7146A)
VOID DEBItransfer(PDEVICE_EXTENSION pdx);
USHORT DEBIread(PDEVICE_EXTENSION pdx, USHORT addr);
VOID DEBIwrite( PDEVICE_EXTENSION pdx, unsigned short addr, unsigned short wdata );
VOID DEBIreplace( PDEVICE_EXTENSION pdx, unsigned short addr, unsigned short mask, 
                 unsigned short wdata );


//{ **************************************************************************** }
//{ Params   : <fdo>      Functional device object                               }
//{            <junk>     I/O request packet                                     }
//{            <regbase>  Registry base key                                      }
//{            <pdx>      Device extension                                       }
//{ Return   : <Result>   I/O allocation                                         }
//{                                                                              }
//{ Descript : Adapter control.                                                  }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
IO_ALLOCATION_ACTION AdapterControl(PDEVICE_OBJECT fdo, PIRP junk, PVOID regbase, PDEVICE_EXTENSION pdx)
{
  // The second argument <junk> to AdapterControl is whatever was in the CurrentIrp
  // field of the device object when you called AllocateAdapterChannel. When you
  // use a DEVQUEUE for IRP queuing, you need to ask the DEVQUEUE object what IRP
  // is current. If you use the standard model, wherein IoStartPacket and
  // IoStartNextPacket manage the queue, junk would be the right IRP. In that case,
  // it would have been named Irp instead.
  PIRP Irp    = GetCurrentIrp(&pdx->dqReadWrite);
  PMDL mdl    = Irp->MdlAddress;
  // There are few differences between code to handle input and output operations
  // using DMA, so it's often convenient to handle both operations in a single
  // subroutine. This line of code examines the major function code for the IRP
  // to decide whether a read or write is occurring.
  BOOLEAN isread = IoGetCurrentIrpStackLocation(Irp)->MajorFunction == IRP_MJ_READ;
  // The regbase argument to this function is an opaque handle that identifies
  // the set of map registers that have been reserved for your use during this
  // operation. You'll need this value later, so you should save it in your device extension.
  pdx->regbase = regbase;
  pdx->busy    = TRUE;
  // KeFlushIoBuffers makes sure that the contents of all processor memory caches
  // for the memory buffer you're using are flushed to memory. The third argument
  // (TRUE) indicates that you're flushing the cache in preparation for a DMA operation.
  // The CPU architecture might require this step because, in general, DMA operations
  // proceed directly to or from memory without necessarily involving the caches.
  KeFlushIoBuffers(mdl, isread, TRUE);
  // The MapTransfer routine programs the DMA hardware for one stage of a transfer
  // and returns the physical address where the transfer should start. Notice that
  // you supply the address of an MDL as the second argument to this function. Since
  // you need an MDL at this point, you would ordinarily have opted for the DO_DIRECT_IO
  // buffering method when you first created your device object, and the I/O Manager
  // would therefore have automatically created the MDL for you. You also pass along
  // the map register base address (regbase). You indicate which portion of the MDL
  // is involved in this stage of the operation by supplying a virtual address
  // (pdx->vaddr) and a byte count (pdx->xfer). MapTransfer will use the virtual
  // address argument to calculate an offset into the buffer area, from which it
  // can determine the physical page numbers containing your data.
  PHYSICAL_ADDRESS address = (*pdx->AdapterObject->DmaOperations->MapTransfer)
                             (pdx->AdapterObject, mdl, regbase, pdx->vaddr, &pdx->xfer, !isread);

  // This is the point at which you program your hardware in the device-specific
  // way that is required. You might, for example, use one of the WRITE_Xxx HAL
  // routines to send the physical address and byte count values to registers on
  // your card, and you might thereafter strobe some command register to begin transferring data.

  // TODO start transfer running on the device using the physical buffer
  // address we just got back from MapTransfer

  // We return the constant DeallocateObjectKeepRegisters to indicate that we're
  // done using the adapter object but are still using the map registers. In this
  // particular case (PCI bus master), there will never be any contention for the
  // adapter object in the first place, so it hardly matters that we've released
  // the adapter object. In other bus-mastering situations, though, we might be
  // sharing a DMA controller with other devices. Releasing the adapter object
  // allows those other devices to begin transfers by using a disjoint set of map
  // registers from the ones we're still using.
  return DeallocateObjectKeepRegisters;
}


//{ **************************************************************************** }
//{ Params   : <fdo>     Functional device object                                }
//{            <Irp>     I/O request packet                                      }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : Dispatch cleanup IRP_MJ_CLEANUP.                                  }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma PAGEDCODE
NTSTATUS DispatchCleanup(PDEVICE_OBJECT fdo, PIRP Irp)
{
  PAGED_CODE();
  KdPrint((DRIVERNAME " - IRP_MJ_CLEANUP\n"));
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
  PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
  // TODO If you use multiple queues, code this instead:
  //	CleanupAllRequests(pdx->queues, arraysize(pdx->queues), stack->FileObject, STATUS_CANCELLED);
  CleanupRequests(&pdx->dqReadWrite, stack->FileObject, STATUS_CANCELLED);
  CleanupControlRequests(pdx, STATUS_CANCELLED, stack->FileObject);
  return CompleteRequest(Irp, STATUS_SUCCESS, 0);
}


//{ **************************************************************************** }
//{ Params   : <fdo>     Functional device object                                }
//{            <Irp>     I/O request packet                                      }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : Dispatch create IRP_MJ_CREATE.                                    }
//{ Notes    : You should also acquire the remove lock when you successfully     }
//{            process an IRP_MJ_CREATE. In contrast to the other situations     }
//{            we've considered, you don't release the lock before returning     }
//{            from the DispatchCreate routine. The balancing call to            }
//{            IoReleaseRemoveLock occurs instead in the dispatch routine for    }
//{            IRP_MJ_CLOSE. In other words, you hold the remove lock for the    }
//{            entire time something has a handle open to your device.           }
//{ **************************************************************************** }
#pragma PAGEDCODE
NTSTATUS DispatchCreate(PDEVICE_OBJECT fdo, PIRP Irp)
{
  PAGED_CODE();
  KdPrint((DRIVERNAME " - IRP_MJ_CREATE\n"));
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

  PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);

  NTSTATUS status = STATUS_SUCCESS;
  if (NT_SUCCESS(status))
  {                                                        // If okay to open
    if (InterlockedIncrement(&pdx->handles) == 1)
    {                                                      // First open handle
//      OpenHandle(pdx, stack->FileObject);
    }
  }
  return CompleteRequest(Irp, status, 0);
}


//{ **************************************************************************** }
//{ Params   : <fdo>     Functional device object                                }
//{            <Irp>     I/O request packet                                      }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : Dispatch close IRP_MJ_CLOSE.                                      }
//{ Notes    : See note on <DispatchCreate>                                      }
//{ **************************************************************************** }
#pragma PAGEDCODE
NTSTATUS DispatchClose(PDEVICE_OBJECT fdo, PIRP Irp)
{
  PAGED_CODE();
  KdPrint((DRIVERNAME " - IRP_MJ_CLOSE\n"));
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
  PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);

//  if (pdx->pevent)                                         // If there is a global event allocated
//    DeregisterEventGlobal(pdx);                            // Remove it
  if (InterlockedDecrement(&pdx->handles) == 0)
  {                                                        // If no more open handles
//    DeregisterEvent(pdx, stack->FileObject);
//    CloseHandle(pdx, stack->FileObject);
  }
  return CompleteRequest(Irp, STATUS_SUCCESS, 0);
}


// TODO If you use separate queues for reads and writes, you need to arrange to
// call StartPacket and CancelRequests with the right queue argument. The easiest
// way to do that is to have separate dispatch and cancel functions for IRP_MJ_READ and
// IRP_MJ_WRITE.


//{ **************************************************************************** }
//{ Params   : <fdo>     Functional device object                                }
//{            <Irp>     I/O request packet                                      }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : Dispatch read/write IRP_MJ_READ/WRITE.                            }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma PAGEDCODE
NTSTATUS DispatchReadWrite(PDEVICE_OBJECT fdo, PIRP Irp)
{
  PAGED_CODE();
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

  IoMarkIrpPending(Irp);
  StartPacket(&pdx->dqReadWrite, fdo, Irp, OnCancelReadWrite);
  return STATUS_PENDING;
}


//{ **************************************************************************** }
//{ Params   : <fdo>     Functional device object                                }
//{            <Irp>     I/O request packet                                      }
//{ Return   : <Result>  NTSTATUS                                                }
//{                                                                              }
//{ Descript : On cancel callback read/write.                                    }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
VOID OnCancelReadWrite(IN PDEVICE_OBJECT fdo, IN PIRP Irp)
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
  CancelRequest(&pdx->dqReadWrite, Irp);
}


// TODO If you use separate queues for reads and writes, you will either need
// separate DPC objects for each, or else you'll need to distinguish in this DPC
// routine between the different kinds of IRP so you can reference the correct
// queue in calls to functions like GetCurrentIrp and StartNextPacket.


//{ **************************************************************************** }
//{ Params   : <Dpc>   Deferred procedure call                                   }
//{            <fdo>   Functional device object                                  }
//{            <junk>  I/O request packet                                        }
//{            <pdx>   Device extension                                          }
//{ Return   : -                                                                 }
//{                                                                              }
//{ Descript : Deferred procedure call for interrupt service routine.            }
//{ Notes    : DMA stuff 'disabled'                                              }
//{            Now handles uncached IRP (notification)                           }
//{ **************************************************************************** }
#pragma LOCKEDCODE
VOID DpcForIsr(PKDPC Dpc, PDEVICE_OBJECT fdo, PIRP junk, PDEVICE_EXTENSION pdx)
{
  PIRP nfyirp = UncacheControlRequest(pdx, &pdx->NotifyIrp);
  
  // If still valid notification
  if (nfyirp)                         
  {
    DumpPerfCounters();

    // Complete the cached IRP request
    PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(nfyirp);
    //KdPrint( (DRIVERNAME " - DpcForIsr stack 0x%08X, outLen: Irp %d, buffer: Irp 0x%08X\n",
    //  stack, stack->Parameters.DeviceIoControl.OutputBufferLength, 
    //  nfyirp->AssociatedIrp.SystemBuffer) );
    
    CompleteRequest(nfyirp, STATUS_SUCCESS, 
      stack->Parameters.DeviceIoControl.OutputBufferLength);            
  }

//  NTSTATUS status;
//  PIRP Irp = GetCurrentIrp(&pdx->dqReadWrite);
//  status = Irp->IoStatus.Status;                           // Get ending status left by ISR
//  BOOLEAN isread = IoGetCurrentIrpStackLocation(Irp)->MajorFunction == IRP_MJ_READ;
//  PMDL mdl = Irp->MdlAddress;

//  (*pdx->AdapterObject->DmaOperations->FlushAdapterBuffers)
//    (pdx->AdapterObject, mdl, pdx->regbase, pdx->vaddr, pdx->xfer, !isread);

//  pdx->nbytes -= pdx->xfer;
//  pdx->numxfer += pdx->xfer;

//  if (pdx->nbytes && NT_SUCCESS(status))
//  {                                                        // If start next stage of transfer
//    pdx->vaddr = (PVOID) ((PUCHAR) pdx->vaddr + pdx->xfer);
//    pdx->xfer = pdx->nbytes;
//    ULONG nregs = ADDRESS_AND_SIZE_TO_SPAN_PAGES(pdx->vaddr, pdx->nbytes);
//    if (nregs > pdx->nMapRegistersAllocated)
//    {
//      nregs = pdx->nMapRegistersAllocated;
//      pdx->xfer = nregs * PAGE_SIZE;
//    }
//    PHYSICAL_ADDRESS address =(*pdx->AdapterObject->DmaOperations->MapTransfer)
//                                (pdx->AdapterObject, mdl, pdx->regbase, pdx->vaddr, &pdx->xfer, !isread);

    // TODO Start the next stage of the operation running on the device
//  }
//  else
//  {                                                        // If request is complete
//    (*pdx->AdapterObject->DmaOperations->FreeMapRegisters)
//      (pdx->AdapterObject, pdx->regbase, pdx->nMapRegistersAllocated);
//    StartNextPacket(&pdx->dqReadWrite, fdo);
//    IoCompleteRequest(Irp, IO_NO_INCREMENT);
//    IoReleaseRemoveLock(&pdx->RemoveLock, Irp);
//  }
}


//{ **************************************************************************** }
//{ Params   : <InterruptObject>  Interrupt object                               }
//{            <pdx>              Device extension                               }
//{ Return   : <Result>           TRUE  if handled                               }
//{                               FALSE if not for us                            }
//{                                                                              }
//{ Descript : On interrupt.                                                     }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
BOOLEAN OnInterrupt (PKINTERRUPT InterruptObject, PDEVICE_EXTENSION pdx)
{
  BOOLEAN UseNotifyIrp = FALSE;
  ULONG   IrqBufferIndex;
  ULONG   FifoBufferIndex;
  ULONG   SigData;
  ULONG   EnableData;
 
  PERF_ISR_ENTRY( PERF_ONINTERRUPT );

  // Our first task is to discover whether our own device is trying to interrupt now.
  // We should check for all pending causes of interrupts. If none, we return immediately.

  // If our device is NOT interrupting, return immediate
  ULONG isrstatus = READ_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_ISR));
  if (isrstatus == 0)
    return FALSE;

  pdx->ISR = isrstatus;

  //KdPrint((DRIVERNAME " - OnInterrupt %08X\n", isrstatus));

  // Check all bits
  // Bugfix V2.01 -> EnableData should have been 0xFFFFFFFE instead of 0xFFFE
  // AND data for enable interrupt: the 0 bit shifts left
  EnableData = 0xFFFFFFFE;                    
  for (IrqBufferIndex = 0; IrqBufferIndex < S626IRQNUMOFFSET; IrqBufferIndex++)
  {
    // Check individual interrupt bit
    if ((isrstatus & 1) != 0)
    {
      pdx->irqBuffers[IrqBufferIndex].irqsReceived++;
      //KdPrint((DRIVERNAME " - OnInterrupt %d recv\n", IrqBufferIndex));

      // disable any non-active interrupt that occurs
      if( !pdx->irqBuffers[IrqBufferIndex].active )
      {
        //KdPrint((DRIVERNAME " - OnInterrupt %d disabled\n", IrqBufferIndex));
        ULONG isrenable = READ_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER));
        isrenable &= EnableData;
        WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER), isrenable);
      }      
      else
      {
        // Interrupt must be active
        //KdPrint((DRIVERNAME " - OnInterrupt %d actv\n", IrqBufferIndex));

        // Auto disable of interrupt
        if (pdx->irqBuffers[IrqBufferIndex].irqAutoDisable)
        {
          //KdPrint((DRIVERNAME " - OnInterrupt %d disabled\n", IrqBufferIndex));
          ULONG isrenable = READ_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER));
          isrenable &= EnableData;
          WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER), isrenable);
        }
        
        pdx->irqBuffers[IrqBufferIndex].irqsActiveReceived++;

        // FIFO handling (we do this before handling signaling)
        if (pdx->irqBuffers[IrqBufferIndex].fifoActive)
        {
          // Advance to the next buffer to use and check for overflows
          pdx->irqBuffers[IrqBufferIndex].fifoBufferPrevious++;

          if (pdx->irqBuffers[IrqBufferIndex].fifoBufferPrevious > 
            pdx->irqBuffers[IrqBufferIndex].fifoBufferLast)
          {
            pdx->irqBuffers[IrqBufferIndex].fifoBufferPrevious = 
              pdx->irqBuffers[IrqBufferIndex].fifoBufferFirst;
            pdx->irqBuffers[IrqBufferIndex].fifoBufferCirculated++;
          }

          FifoBufferIndex = pdx->irqBuffers[IrqBufferIndex].fifoBufferPrevious;

          // check allocation of FIFO buffer before using it
          if (pdx->fifoBuffers[FifoBufferIndex].bufferId >= 0)
          {            
            RtlCopyMemory(pdx->fifoBuffers[FifoBufferIndex].vaBuffer,
                          pdx->fifoBuffers[FifoBufferIndex].vaBufferSource,
                          pdx->fifoBuffers[FifoBufferIndex].bufferSize);
            // Check for overflow on buffer
            // Overflows are marked on different locations:
            // . the FIFO buffer itself
            // . the IRQ which uses the FIFO buffer
            // . the global overflow counter
            if (pdx->fifoBuffers[FifoBufferIndex].writeTag == TRUE)
            {
              pdx->fifoOverflows++;
              pdx->irqBuffers[IrqBufferIndex].fifoOverflows++;
              pdx->fifoBuffers[FifoBufferIndex].bufferOverflows++;
            }

            pdx->fifoBuffers[FifoBufferIndex].writeTag = TRUE;
            pdx->fifoBuffers[FifoBufferIndex].bufferIrqsActiveReceived = 
              pdx->irqBuffers[IrqBufferIndex].irqsActiveReceived;
          } // allocated FIFO buffer

        } // FIFO active        

        // handle S626 IRQ's (master SAA7146A IRQ=6) (irq buffers 32-83)
        if( SAA7146A_S626_INT_NUM == IrqBufferIndex )
        {
          // get the timestamp for this capture
          pdx->s626IrqStatus.timeStamp = KeQueryPerformanceCounter( NULL );

          // TODO: improve performance(?) by reading Interrupt Selects
          //  LP_RDINTSELA,B,C in IOCTL dispatch and caching for ISR's use
          //  as they change infrequently and via the IOCTL interface.

          // Get all DIO IRQs.  These are derived by masking DIO events captures 
          //  against the DIO interrupt enables that are stored in the readable 
          //  interrupt enable registers.
          pdx->s626IrqStatus.intrStatusWord[0] = DEBIread( pdx, LP_RDCAPFLGA ) & 
            DEBIread( pdx, LP_RDINTSELA );
          pdx->s626IrqStatus.intrStatusWord[1] = DEBIread( pdx, LP_RDCAPFLGB ) &
            DEBIread( pdx, LP_RDINTSELB );
          pdx->s626IrqStatus.intrStatusWord[2] = DEBIread( pdx, LP_RDCAPFLGC ) &
            DEBIread( pdx, LP_RDINTSELC );
          
	        // Get all counter IRQs.  These are counter event captures 
          //  (in the RDMISC2 register).	        
          pdx->s626IrqStatus.intrStatusWord[3] = DEBIread( pdx, LP_RDMISC2 );

          //KdPrint((DRIVERNAME " - OnInterrupt S626 [0-3]: %04X, %04X, %04X, %04X\n", 
          //  pdx->s626IrqStatus.intrStatusWord[0], pdx->s626IrqStatus.intrStatusWord[1],
          //  pdx->s626IrqStatus.intrStatusWord[2], pdx->s626IrqStatus.intrStatusWord[3]));

          // Reset all S626 interrupts to avoid missing succeeding interrupts.  
          //  The alternative of disabling all S626 interrupts until the
          //  user's ISR is run and re-enables S626 interrupts is unacceptable 
          //  in ensuring FIFO actions don't miss any interrupt events.

          // disable S626 master interrupt if not already done as part of 
          //  irqAutoDisable function performed above
          if( !pdx->irqBuffers[IrqBufferIndex].irqAutoDisable )
          {
            ULONG isrenable = READ_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER));
            isrenable &= EnableData;
            WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER), isrenable);
            //KdPrint( (DRIVERNAME " - Disabled S626 Master Interrupt\n" ) );
          }

          USHORT cntrIntrConfigWrd = 0;

          // reset all event capture status flags 
          //  and read interrupt enables for counters
          for( unsigned short wordIdx = 0; wordIdx < 3; ++wordIdx )
          {
            // reset all DIO channels (16 at the same time)
            
            //   accomodate the use of edge capture interrupts 
            //    on DIO channels 0-39 only
            unsigned short channelMsk = (wordIdx < 2) ? 0xFFFF : 0x00FF ;

            //   disable the capture.
	          DEBIwrite( pdx, LP_MISC1, MISC1_NOEDCAP );
	          DEBIwrite( pdx, WRCapSel[wordIdx], channelMsk );

	          //   re-enable the capture.
	          DEBIwrite( pdx, LP_MISC1, MISC1_EDCAP );
	          DEBIwrite( pdx, WRCapSel[wordIdx], channelMsk );

            // retrieve Counters' interrupt configurations and 
            //  reset all counters' event capture status
          
            // TODO: improve performance by reading Interrupt Selects,
            //  AdrsCRA and AdrsCRB, in IOCTL dispatch and caching for ISR's use
            //  as they change infrequently and via the IOCTL interface.

            // read CRA interrupt status and insert in mask for interrupt config word
            USHORT cntrStatusWrd = DEBIread( pdx, AdrsCRA[wordIdx] );
            USHORT cntrIntrSrc = ((cntrStatusWrd >> CRABIT_INTSRC_A) & (unsigned short)3);
            cntrIntrConfigWrd |= EventBits[wordIdx][cntrIntrSrc];

            //KdPrint((DRIVERNAME " - cntrIntrConfigWrd CRA 0x%04X\n", cntrIntrConfigWrd));
              
            // read CRB interrupt status and insert in mask for capture status word
            cntrStatusWrd = DEBIread( pdx, AdrsCRB[wordIdx] );
            cntrIntrSrc = ((cntrStatusWrd >> CRBBIT_INTSRC_B) & (unsigned short)3);
            cntrIntrConfigWrd |= EventBits[wordIdx + 1][cntrIntrSrc];

            //KdPrint((DRIVERNAME " - cntrIntrConfigWrd CRB 0x%04X\n", cntrIntrConfigWrd));

            // use CRB status word to reset both counters' interrupt requests
            //  (this operation will do the A & B counters at the same time)
            DEBIwrite( pdx, AdrsCRB[wordIdx], ( cntrStatusWrd & 0x1FFF )| 0xE000 );
            //DEBIreplace( pdx, AdrsCRB[wordIdx], 0x1FFF, 0xE000 );

            //KdPrint((DRIVERNAME " - OnInterrupt Reset S626 %d\n", wordIdx ) );            
          }

          // mask counters' interrupt status with counters' interrupt enable/configuration
          pdx->s626IrqStatus.intrStatusWord[3] &= cntrIntrConfigWrd;
          //KdPrint((DRIVERNAME " - OnInterrupt Cnt Intr Status 0x%04X\n", \
          //  pdx->s626IrqStatus.intrStatusWord[3] ) );            

          // reenable S626 master interrupt if irqAutoDisable not in effect
          if( !pdx->irqBuffers[IrqBufferIndex].irqAutoDisable )
          {
            ULONG intrMsk = READ_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER));
            intrMsk |= (1 << SAA7146A_S626_INT_NUM);
            WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER), intrMsk);
            //KdPrint( (DRIVERNAME " - Re-enabled S626 Master Interrupt\n" ) );
          }
          
          unsigned int numActive = pdx->s626ActiveIrqBufCnt;
          //KdPrint( (DRIVERNAME " - numActive=%d\n", numActive ) );

          // no active S626 interrupt buffers
          if( 0 == numActive )
          {            
            // default to configuration of S626 master interrupt
            if (pdx->irqBuffers[IrqBufferIndex].eventActive)
            {
              KdPrint((DRIVERNAME " - OnInterrupt %d S626 eventActive\n", IrqBufferIndex));
              UseNotifyIrp = TRUE;
            }
          }

          while( 0 < numActive )
          {
            unsigned int s626IrqBufIdx = pdx->s626ActiveIrqBufList[--numActive];
            //KdPrint( (DRIVERNAME " - Itr S626IrqBufs %d\n", s626IrqBufIdx) );

            // only for debugging
            if(  S626IRQNUMOFFSET >= s626IrqBufIdx || IRQBUFFERS < s626IrqBufIdx )
            {
              KdPrint( (DRIVERNAME " - ERROR: s626IrqBufIdx out of range %d\n",
                s626IrqBufIdx ) );
              KdPrint( (DRIVERNAME " - ERROR: numActive=0x%08X\n", numActive) );
              break;
            }

            // 40 DIO channels have IRQ buffer numbers: 32-71, groups 0-2
            // 12 Counter interrupts (6 counters, overflow and index each)
            //  have IRQ buffer numbers: 72-83, group 3
            BOOLEAN intStatus = FALSE;
            if( 72 > s626IrqBufIdx )
            {
              // process DIO channels
              intStatus = 
                pdx->s626IrqStatus.intrStatusWord[(s626IrqBufIdx - S626IRQNUMOFFSET) / 16] >> 
                (s626IrqBufIdx % 16) & 1;
            }
            else
            {
              // process counter overflow and index interrupts
              // bits 0-3 of status word not used, bits 4-15 represent 12 counter bits
              intStatus = 
                (pdx->s626IrqStatus.intrStatusWord[3] >> (s626IrqBufIdx - 72 + 4)) & 1;                
            }

            if( intStatus )
            {              
              if( pdx->irqBuffers[s626IrqBufIdx].eventActive )
              {
                KdPrint((DRIVERNAME " - OnInterrupt S626 Buf=%d  eventActive\n", 
                  s626IrqBufIdx));
                UseNotifyIrp = TRUE;
              }

              if( pdx->irqBuffers[s626IrqBufIdx].fifoActive )
              {                
                // Advance to the next buffer to use and check for overflows
                pdx->irqBuffers[s626IrqBufIdx].fifoBufferPrevious++;

                if (pdx->irqBuffers[s626IrqBufIdx].fifoBufferPrevious > 
                  pdx->irqBuffers[s626IrqBufIdx].fifoBufferLast)
                {
                  pdx->irqBuffers[s626IrqBufIdx].fifoBufferPrevious = 
                    pdx->irqBuffers[s626IrqBufIdx].fifoBufferFirst;
                  pdx->irqBuffers[s626IrqBufIdx].fifoBufferCirculated++;
                }

                // get current FIFO buffer index to use
                const ULONG s626FifoBufIdx = 
                                  (ULONG) pdx->irqBuffers[s626IrqBufIdx].fifoBufferPrevious;

                // check allocation of FIFO buffer before using it
                if (pdx->fifoBuffers[s626FifoBufIdx].bufferId >= 0)
                {
                  // iterate buffer's register list
                  const P626IRQREGLIST regList = 
                            (P626IRQREGLIST) pdx->fifoBuffers[s626FifoBufIdx].vaBufferSource;

                  // beginning address of copy target buffer
                  PULONG fifoBuffer = (PULONG)pdx->fifoBuffers[s626FifoBufIdx].vaBuffer;

                  // insert interrupt timestamp                  
                  RtlCopyMemory( fifoBuffer, &pdx->s626IrqStatus.timeStamp, 
                    sizeof(pdx->s626IrqStatus.timeStamp) );
                  fifoBuffer += (sizeof(pdx->s626IrqStatus.timeStamp) / sizeof(fifoBuffer[0]));
                  
                  // ensure a write past buffer end doesn't happen if the count of registers
                  //  isn't correctly specified in the register list
                  PULONG fifoBufferEnd = (PULONG)pdx->fifoBuffers[s626FifoBufIdx].vaBuffer +
                    (pdx->fifoBuffers[s626FifoBufIdx].bufferSize / sizeof(fifoBuffer[0]));

                  // read register addresses specified in register list of source buffer
                  for( UINT32* regAddr = regList->s626RegAddr; 
                    (regAddr < (regAddr + regList->regCount)) && (fifoBuffer < fifoBufferEnd);
                    ++regAddr, ++fifoBuffer )
                  {                      
                    *fifoBuffer = DEBIread( pdx, (USHORT)(*regAddr) );
                    //KdPrint((DRIVERNAME " - Fifo %d read register 0x%04X, %04X\n", 
                    //  s626FifoBufIdx, *regAddr, (USHORT)*fifoBuffer));
                  }

                  // Check for overflow on buffer
                  // Overflows are marked on different locations:
                  // . the FIFO buffer itself
                  // . the IRQ which uses the FIFO buffer
                  // . the global overflow counter
                  if (pdx->fifoBuffers[s626FifoBufIdx].writeTag == TRUE)
                  {
                    pdx->fifoOverflows++;
                    pdx->irqBuffers[s626IrqBufIdx].fifoOverflows++;
                    pdx->fifoBuffers[s626FifoBufIdx].bufferOverflows++;
                  }

                  pdx->fifoBuffers[s626FifoBufIdx].writeTag = TRUE;
                  pdx->fifoBuffers[s626FifoBufIdx].bufferIrqsActiveReceived = 
                                              pdx->irqBuffers[s626IrqBufIdx].irqsActiveReceived;

                  fifoBuffer = (PULONG)pdx->fifoBuffers[s626FifoBufIdx].vaBuffer;
                  //KdPrint((DRIVERNAME " - Fifo %d 0x%08X, %08X, %08X, %08X\n", \
                  //  s626FifoBufIdx, fifoBuffer[0], fifoBuffer[1], fifoBuffer[2], \
                  //  fifoBuffer[3] ));

                } // FIFO allocated
              } // FIFO active

            } // interrupt status true
          } // for each S626 IRQ                            

        } // handle S626 master interrupt
        else
        {
          // EVENT handling
          if (pdx->irqBuffers[IrqBufferIndex].eventActive)
          {
            KdPrint((DRIVERNAME " - OnInterrupt %d eventActive\n", IrqBufferIndex));
            UseNotifyIrp = TRUE;
          }
        }        

      } // handle all active interrupts

    } // check interrupt bit

    isrstatus >>= 1;                                       // Next bit
    EnableData <<= 1;                                      // Shift bit to the left
    EnableData |= 1;                                       // Make sure shifted bit is 1
  } // for each SAA7146A interrupt bit

  // Clear the interrupt(s) by writing a '1' -> just do them all
  // Note that for a 'persistent' interrupt a new interrupt will be generated again.
  // For instance the RPSx time out interrupts will be generated again if it is
  // not disabled...
  WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_ISR), 0xFFFFFFFF);
  //KdPrint( (DRIVERNAME " - OnInterrupt cleared all interrupts\n") );

  //timeStamp = KeQueryPerformanceCounter(NULL);
  //elapsedTime.QuadPart = 
  //  ((timeStamp.QuadPart - entryTimeStamp.QuadPart)*100000)/perfCntFreq.QuadPart;
  //KdPrint((DRIVERNAME " - Cleared SAA7146_ISR %d time=%u, High=%u, Low=%u\n", 
  //  IrqBufferIndex, elapsedTime.LowPart, timeStamp.HighPart, timeStamp.LowPart));

  // We have to go through all of it again for the signalling. We do this after
  // clearing the interrupt status because a RPS program might just re-enable it
  isrstatus = pdx->ISR;
  for (IrqBufferIndex = 0; IrqBufferIndex < S626IRQNUMOFFSET; IrqBufferIndex++)
  {
    // Check individual bit
    if ((isrstatus & 1) != 0)
    {
      // Interrupt must be active
      if (pdx->irqBuffers[IrqBufferIndex].active)
      {
        // SIG handling
        if (pdx->irqBuffers[IrqBufferIndex].signalActive)
        {
          SigData = READ_REGISTER_ULONG((PULONG) (pdx->membase + pdx->irqBuffers[IrqBufferIndex].signalRegister));
          SigData &= pdx->irqBuffers[IrqBufferIndex].signalAnd;
          SigData |= pdx->irqBuffers[IrqBufferIndex].signalOr;
          SigData ^= pdx->irqBuffers[IrqBufferIndex].signalXor;
          WRITE_REGISTER_ULONG((PULONG) (pdx->membase + pdx->irqBuffers[IrqBufferIndex].signalRegister), SigData);
        }
      }
    }
    isrstatus >>= 1;                                       // Next bit
  }
  
  // HERE WE PUT THE CODE FOR INTERRUPT HANDLING
  //  -> fifoActive
  //     check if the FIFO buffer of <fifoBuffer> is available (BufferId >=0)
  //     ..handle it


  // Check for spurious interrupt.

  // Adjust our interrupt counter
  pdx->InterruptCount++;                                   // Increment our interrupt counter (no need to interlock)
  if (pdx->InterruptCount > 10000000)
    pdx->InterruptCount = 0;


// Note: the global event signalling was originally implemented but only gave a
//       throughput of about 9 times/sec! Not sure why, but we are now using IRP
//       notification instead (meaning that an IRP is waiting for it to be completed)
// SignalEventGlobal(pdx);                                // Signal global event (if any)

//  if (!pdx->busy)
//    return TRUE;

  // When we use a DEVQUEUE, we rely on the queue object to keep track of the current IRP.
  // This interrupt might be one that we don't expect because we're not currently servicing
  // any IRP. In that case, we still have to clear the interrupt but shouldn't do anything else.
//  PIRP Irp = GetCurrentIrp(&pdx->dqReadWrite);
//  ASSERT(Irp);

  // Is there a pending notification (cached) IRP we should call?
  // Must be enabled by IRQ handling
  if (pdx->NotifyIrp && UseNotifyIrp)
  {
    //KdPrint( (DRIVERNAME " - OnInterrupt cached DPC\n") );

    PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(pdx->NotifyIrp);
    //ULONG cbin  = stack->Parameters.DeviceIoControl.InputBufferLength;
    ULONG cbout = stack->Parameters.DeviceIoControl.OutputBufferLength;
    //ULONG code  = stack->Parameters.DeviceIoControl.IoControlCode;    

    if( sizeof(S626IRQSTATUS) == cbout )
    {
      if( NULL != pdx->NotifyIrp->AssociatedIrp.SystemBuffer )
      {
        RtlCopyMemory( pdx->NotifyIrp->AssociatedIrp.SystemBuffer, &pdx->s626IrqStatus, 
          sizeof(pdx->s626IrqStatus) );

        //PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation( pdx->NotifyIrp );
        //KdPrint( (DRIVERNAME " - OnInterrupt stack 0x%08X, buffer 0x%08X, size %d\n",
        //  stack, pdx->NotifyIrp->AssociatedIrp.SystemBuffer, 
        //  stack->Parameters.DeviceIoControl.OutputBufferLength ) );

        //PS626IRQSTATUS status = (PS626IRQSTATUS)pdx->NotifyIrp->AssociatedIrp.SystemBuffer;        
        //KdPrint( (DRIVERNAME " - OnInterrupt %04X, %04X, %04X, %04X\n", 
        //  status->intrStatusWord[0], status->intrStatusWord[1],
        //  status->intrStatusWord[2], status->intrStatusWord[3]) );
      }
      else
      {
        //KdPrint( (DRIVERNAME " - OnInterrupt Notify IRP has NULL buffer\n") )
        ;
      }
    }

    // Let the DPC complete the cached IRP request
    IoRequestDpc(pdx->DeviceObject, NULL, pdx); 
  }

  // It's also possible that a Plug and Play or power event has occurred that will
  // cause any new IRPs to be rejected by the dispatch routine. The DEVQUEUE's
  // AreRequestsBeingAborted function tells us that fact so that we can abort the
  // current request right now. Aborting an active request is a reasonable thing
  // to do with a device such as this that proceeds byte by byte. Similarly it's a
  // good idea to check whether the IRP has been cancelled if it will take a long
  // time to finish the IRP. If your device interrupts only when it's done with a
  // long transfer, you could leave this test out of your ISR.
//  NTSTATUS status;
//  if (Irp->Cancel)
//    status = STATUS_CANCELLED;
//  else
//    status = AreRequestsBeingAborted(&pdx->dqReadWrite);

  // As we progressed through the preceding code, we set the Boolean dpc variable
  // to TRUE if a DPC is now appropriate to complete the current IRP.
//  if (!NT_SUCCESS(status))
//  {                                                        // If to terminate request
//    pdx->busy = FALSE;
//    Irp->IoStatus.Status = status;
//    Irp->IoStatus.Information = 0;
//    IoRequestDpc(pdx->DeviceObject, NULL, pdx);
//    return TRUE;
//  }

  // We are now logically complete, request a DPC
//  Irp->IoStatus.Status = STATUS_SUCCESS;
//  Irp->IoStatus.Information = 0;
//  pdx->busy = FALSE;
//  IoRequestDpc(pdx->DeviceObject, NULL, pdx);

  //timeStamp = KeQueryPerformanceCounter(NULL);
  //elapsedTime.QuadPart = 
  //  ((timeStamp.QuadPart - entryTimeStamp.QuadPart)*100000)/perfCntFreq.QuadPart;
  //KdPrint((DRIVERNAME " - Exit ISR, time=%u, High=%u, Low=%u\n", 
  //  elapsedTime.LowPart, timeStamp.HighPart, timeStamp.LowPart));

  PERF_ISR_EXIT( PERF_ONINTERRUPT );
  return TRUE;
}


//{ **************************************************************************** }
//{ Params   : <fdo>         Functional device object                            }
//{            <raw>         Untranslated partial resource list                  }
//{            <translated>  Translated partial resource list                    }
//{ Return   : <Result>      NTSTATUS                                            }
//{                                                                              }
//{ Descript : Start a single device.                                            }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma PAGEDCODE
NTSTATUS StartDevice(PDEVICE_OBJECT fdo, PCM_PARTIAL_RESOURCE_LIST raw, PCM_PARTIAL_RESOURCE_LIST translated)
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
  NTSTATUS status;

  // Identify the I/O resources we're supposed to use.
  ULONG vector;
  KIRQL irql;
  KINTERRUPT_MODE mode;
  KAFFINITY affinity;
  BOOLEAN irqshare;
  BOOLEAN gotinterrupt = FALSE;

  PHYSICAL_ADDRESS membase;
  BOOLEAN gotmemory = FALSE;

  if (!translated)
    return STATUS_DEVICE_CONFIGURATION_ERROR;              // No resources assigned??

  PCM_PARTIAL_RESOURCE_DESCRIPTOR resource = translated->PartialDescriptors;
  ULONG nres = translated->Count;
  for (ULONG i = 0; i < nres; ++i, ++resource)
  {                                                        // For each resource
    switch (resource->Type)
    {                                                      // Switch on resource type
      case CmResourceTypeInterrupt:
        irql = (KIRQL) resource->u.Interrupt.Level;
        vector = resource->u.Interrupt.Vector;
        affinity = resource->u.Interrupt.Affinity;
        mode = (resource->Flags == CM_RESOURCE_INTERRUPT_LATCHED)
                 ? Latched : LevelSensitive;
        irqshare = resource->ShareDisposition == CmResourceShareShared;
        gotinterrupt = TRUE;
        break;
      case CmResourceTypeMemory:
        membase = resource->u.Memory.Start;
        pdx->memsize = resource->u.Memory.Length;
        gotmemory = TRUE;
        break;
      case CmResourceTypeDevicePrivate:
        // Reserved system resource
//        KdPrint((DRIVERNAME " - Resource type device private (reserved for system use)\n"));
        break;
      default:
        KdPrint((DRIVERNAME " - Unexpected I/O resource type %d\n", resource->Type));
        break;
    }
  }

  // Verify that we got all the resources we were expecting
  if (!(TRUE
      && gotinterrupt
      && gotmemory
     ))
  {
    KdPrint((DRIVERNAME " - Didn't get expected I/O resources\n"));
    return STATUS_DEVICE_CONFIGURATION_ERROR;
  }

  // Determine bus type before proceeding, just so we don't have a bunch of cleanup
  // to do if this call fails
  INTERFACE_TYPE bustype;
  ULONG junk;
  status = IoGetDeviceProperty(pdx->Pdo, DevicePropertyLegacyBusType, sizeof(bustype),
                               &bustype, &junk);
  if (!NT_SUCCESS(status))
  {
    KdPrint((DRIVERNAME " - IoGetDeviceProperty failed - %X\n", status));
    return status;
  }

  // Get bus number and slot number
  // From, http://www.tutorials-blog.com/nt/obtaining-slot/ and Q253232
  status = IoGetDeviceProperty(pdx->Pdo, DevicePropertyBusNumber, sizeof(pdx->busNumber),
    &pdx->busNumber, &junk);

  if (!NT_SUCCESS(status))
  {
    KdPrint((DRIVERNAME " - IoGetDeviceProperty for DevicePropertyBusNumber failed - %X\n", 
      status));
    //return status;
  }

  unsigned long slotFunctionNum;
  // perhaps DevicePropertyUINumber would work as well for this?
  status = IoGetDeviceProperty(pdx->Pdo, DevicePropertyAddress, sizeof(slotFunctionNum),
                                     &slotFunctionNum, &junk);
  if (NT_SUCCESS(status)) 
  {
    //
    // See Article ID: Q253232 it appears that the
    // bus driver does not return the standard 
    // PCI_SLOT_NUMBER structure.
    //
    // For PCI, the DevicePropertyAddress has device, AKA slot, number 
    // in the high word and the function number in the low word. 
    // 
    pdx->busSlotNumber = ((slotFunctionNum >> 16) & 0x00001F);
    //FunctionNumber = (slotFunction & 0x3);
  } 
  else
  {
    KdPrint((DRIVERNAME " - IoGetDeviceProperty for DevicePropertyAddress failed - %X\n", 
      status));
  }

  pdx->membase = (PUCHAR) MmMapIoSpace(membase, pdx->memsize, MmNonCached);
  if (!pdx->membase)
  {
    KdPrint((DRIVERNAME " - Unable to map memory block %I64X, length %X\n", membase, pdx->memsize));
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // Configure a DMA adapter object
  DEVICE_DESCRIPTION dd;
  RtlZeroMemory(&dd, sizeof(dd));
  dd.Version = DEVICE_DESCRIPTION_VERSION;
  dd.InterfaceType = bustype;
  dd.MaximumLength = COMMONBUFFERSIZE;
  dd.Dma32BitAddresses = TRUE;
  dd.Dma64BitAddresses = FALSE;                            // V1.01
  dd.Master = TRUE;

  pdx->AdapterObject = IoGetDmaAdapter(pdx->Pdo, &dd, &pdx->nMapRegisters);
  if (!pdx->AdapterObject)
  {                                                        // If we can't create adapter object
    KdPrint((DRIVERNAME " - Unable to create DMA adapter object\n"));
    if (pdx->membase)
      MmUnmapIoSpace(pdx->membase, pdx->memsize);
    pdx->membase = NULL;
    return STATUS_UNSUCCESSFUL;
  }

  ULONG loop;
  for (loop = 0; loop < DMABUFFERS; loop++)                // Initialize our DMA buffer array
    pdx->dmaBuffers[loop].bufferId = -1;
  for (loop = 0; loop < FIFOBUFFERS; loop++)               // Initialize our FIFO buffer array
    pdx->fifoBuffers[loop].bufferId = -1;
  pdx->fifoOverflows = 0;
  for (loop = 0; loop < IRQBUFFERS; loop++)                // Initialize the IRQ buffers
  {
    pdx->irqBuffers[loop].irqsReceived = 0;
    pdx->irqBuffers[loop].irqsActiveReceived = 0;
    pdx->irqBuffers[loop].active = FALSE;
    pdx->irqBuffers[loop].eventActive = TRUE;              // By default generate notification event if someone waiting for it
    pdx->irqBuffers[loop].signalActive = FALSE;
    pdx->irqBuffers[loop].fifoActive = FALSE;
    pdx->irqBuffers[loop].irqAutoDisable = FALSE;
  }

  // Since we will be working with a common buffer we have to allocate it -> NOW DONE BY APPLICATION!
//  pdx->vaCommonBuffer =
//    (*pdx->AdapterObject->DmaOperations->AllocateCommonBuffer)
//      (pdx->AdapterObject, COMMONBUFFERSIZE, &pdx->paCommonBuffer, FALSE);
//  if (!pdx->vaCommonBuffer)
//  {                                                        // If we can't create common buffer
//    KdPrint((DRIVERNAME " - Unable to allocate common DMA buffer\n"));
//    if (pdx->membase)
//      MmUnmapIoSpace(pdx->membase, pdx->memsize);
//    pdx->membase = NULL;
//    return STATUS_UNSUCCESSFUL;
//  }
//  KdPrint((DRIVERNAME " - Common buffer @ %p, %p%\n", pdx->vaCommonBuffer, pdx->paCommonBuffer));

  // Prevent device from interrupting
  WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER), 0);              // Disable master IRQ
  WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_ISR), 0xFFFFFFFF);     // Clear pending IRQ's

  status = IoConnectInterrupt(&pdx->InterruptObject, (PKSERVICE_ROUTINE) OnInterrupt,
                              (PVOID) pdx, NULL, vector, irql, irql, mode, irqshare, affinity, FALSE);
  if (!NT_SUCCESS(status))
  {
    KdPrint((DRIVERNAME " - IoConnectInterrupt failed - %X\n", status));
    if (pdx->membase)
      MmUnmapIoSpace(pdx->membase, pdx->memsize);
    pdx->membase = NULL;
    if (pdx->AdapterObject)
      (*pdx->AdapterObject->DmaOperations->PutDmaAdapter)
        (pdx->AdapterObject);
    pdx->AdapterObject = NULL;
    return status;
  }

  // Allow all interrupts
// DONE BY APPLICATION ->  WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER), 0xFFFFFFFF);     // Enable all

  return STATUS_SUCCESS;
}


// TODO If you use separate queues for reads and writes, you will need separate
// StartIo routines too.


//{ **************************************************************************** }
//{ Params   : <fdo>  Functional device object                                   }
//{            <Irp>  I/O request packet                                         }
//{ Return   : -                                                                 }
//{                                                                              }
//{ Descript : Start I/O request packet.                                         }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma LOCKEDCODE
VOID StartIo(IN PDEVICE_OBJECT fdo, IN PIRP Irp)
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
  PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
  NTSTATUS status = IoAcquireRemoveLock(&pdx->RemoveLock, Irp);
  if (!NT_SUCCESS(status))
  {
    CompleteRequest(Irp, status, 0);
    return;
  }

  PMDL mdl = Irp->MdlAddress;
  pdx->numxfer = 0;
  pdx->xfer = pdx->nbytes = MmGetMdlByteCount(mdl);
  pdx->vaddr = MmGetMdlVirtualAddress(mdl);

  ULONG nregs = ADDRESS_AND_SIZE_TO_SPAN_PAGES(pdx->vaddr, pdx->nbytes);
  if (nregs > pdx->nMapRegisters)
  {                                                        // If staged transfer needed
    nregs = pdx->nMapRegisters;
    pdx->xfer = nregs * PAGE_SIZE - MmGetMdlByteOffset(mdl);
  }
  pdx->nMapRegistersAllocated = nregs;                     // Save for deallocation later

  status = (*pdx->AdapterObject->DmaOperations->AllocateAdapterChannel)
             (pdx->AdapterObject, fdo, nregs, (PDRIVER_CONTROL) AdapterControl, pdx);
  if (!NT_SUCCESS(status))
  {
    KdPrint((DRIVERNAME " - AllocateAdapterChannel failed - %X\n", status));
    StartNextPacket(&pdx->dqReadWrite, fdo);
    CompleteRequest(Irp, status, 0);
    IoReleaseRemoveLock(&pdx->RemoveLock, Irp);
  }
}


//{ **************************************************************************** }
//{ Params   : <fdo>      Functional device object                               }
//{            <oktouch>  ..                                                     }
//{ Return   : -                                                                 }
//{                                                                              }
//{ Descript : Stop a device.                                                    }
//{ Notes    :                                                                   }
//{ **************************************************************************** }
#pragma PAGEDCODE
VOID StopDevice(IN PDEVICE_OBJECT fdo, BOOLEAN oktouch /* = FALSE */)
{
  PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
  AbortPendingIoctls(pdx, STATUS_CANCELLED);

  if (pdx->InterruptObject)
  {                                                        // If to disconnect interrupt
    // Prevent device from generating more interrupts
    WRITE_REGISTER_ULONG((PULONG) (pdx->membase + SAA7146_IER), 0);  // Disable master IRQ

    IoDisconnectInterrupt(pdx->InterruptObject);
    pdx->InterruptObject = NULL;
  }

  if (pdx->membase)
    MmUnmapIoSpace(pdx->membase, pdx->memsize);
  pdx->membase = NULL;

  ULONG loop;
  // IRQ buffers
  for (loop = 0; loop < IRQBUFFERS; loop++)
  {
    pdx->irqBuffers[loop].active = FALSE;
    pdx->irqBuffers[loop].eventActive = TRUE;
    pdx->irqBuffers[loop].signalActive = FALSE;
    pdx->irqBuffers[loop].fifoActive = FALSE;
  }

  // FIFO buffers
  for (loop = 0; loop < FIFOBUFFERS; loop++)
  {
    if (pdx->fifoBuffers[loop].bufferId >= 0)              // If in use
      ExFreePool(pdx->fifoBuffers[loop].vaBuffer);         // Release the allocated memory
    pdx->fifoBuffers[loop].bufferId = -1;
  }

  // DMA buffers
  for (loop = 0; loop < DMABUFFERS; loop++)
  {
    if (pdx->dmaBuffers[loop].bufferId >= 0)               // If in use
    {                                                      // Release it
      // Release the allocated common buffer
      (*pdx->AdapterObject->DmaOperations->FreeCommonBuffer)
        (pdx->AdapterObject, pdx->dmaBuffers[loop].buffersize, pdx->dmaBuffers[loop].paBuffer,
         pdx->dmaBuffers[loop].vaBuffer, FALSE);
    }
    pdx->dmaBuffers[loop].bufferId = -1;
  }

  if (pdx->AdapterObject)
  {
    // Release common DMA buffer
//    (*pdx->AdapterObject->DmaOperations->FreeCommonBuffer)
//      (pdx->AdapterObject, COMMONBUFFERSIZE, pdx->paCommonBuffer,
//       pdx->vaCommonBuffer, FALSE);
    // Release adapter object
    (*pdx->AdapterObject->DmaOperations->PutDmaAdapter)
      (pdx->AdapterObject);
  }
  pdx->AdapterObject = NULL;
}


// Enable/disable a function or test status bit(s) that are accessed through Main Control Registers 1 or 2.
//#define MC_ENABLE( REGADRS, CTRLWORD )			WR7146( ( REGADRS ), ( (unsigned long)( CTRLWORD ) << 16 ) | (unsigned long)( CTRLWORD ) )
//#define MC_DISABLE( REGADRS, CTRLWORD )			WR7146( ( REGADRS ), (unsigned long)( CTRLWORD ) << 16 )
//#define MC_TEST( REGADRS, CTRLWORD )			( ( RR7146( REGADRS ) & CTRLWORD ) != 0 )

////////////////////////////////////////////////////////////////////////////////////
// Execute a DEBI transfer.  This must be called from within a critical section.
#pragma LOCKEDCODE
VOID DEBItransfer(PDEVICE_EXTENSION pdx)
{
	// Initiate upload of shadow RAM to DEBI control register.
	//MC_ENABLE( P_MC2, MC2_UPLD_DEBI );
  WRITE_REGISTER_ULONG( (PULONG)(pdx->membase + P_MC2), 
    ((unsigned long)( MC2_UPLD_DEBI ) << 16) | (unsigned long)(MC2_UPLD_DEBI) );

	// Wait for completion of upload from shadow RAM to DEBI control register.
	//while ( !MC_TEST( P_MC2, MC2_UPLD_DEBI ) );
  while( !(0 != (READ_REGISTER_ULONG((PULONG)(pdx->membase + P_MC2)) & MC2_UPLD_DEBI)) );

	// Wait until DEBI transfer is done.
	//while ( RR7146(P_PSR) & PSR_DEBI_S );
  while( READ_REGISTER_ULONG((PULONG)(pdx->membase + P_PSR)) & PSR_DEBI_S );
}

/////////////////////////////////////////////////////////////////
// Return a value from a gate array register.
#pragma LOCKEDCODE
USHORT DEBIread(PDEVICE_EXTENSION pdx, USHORT addr)
{
	// Set up DEBI control register value in shadow RAM.
	//WR7146( P_DEBICMD, DEBI_CMD_RDWORD | addr );
  WRITE_REGISTER_ULONG((PULONG) (pdx->membase + P_DEBICMD), DEBI_CMD_RDWORD | addr);

	// Execute the DEBI transfer.
	DEBItransfer(pdx);

	// Fetch target register value.
	//unsigned short retval = (unsigned short)RR7146( P_DEBIAD );
  return (USHORT)READ_REGISTER_ULONG((PULONG)(pdx->membase + P_DEBIAD));

	// Return register value.
	//return retval;
}

/////////////////////////////////////////////////////////////////////////
// Write a value to a gate array register.
#pragma LOCKEDCODE
VOID DEBIwrite( PDEVICE_EXTENSION pdx, unsigned short addr, unsigned short wdata )
{
  // Set up DEBI control register value in shadow RAM.
	//WR7146( P_DEBICMD, DEBI_CMD_WRWORD | addr );
  WRITE_REGISTER_ULONG((PULONG) (pdx->membase + P_DEBICMD), DEBI_CMD_WRWORD | addr);

	//WR7146( P_DEBIAD,  wdata );
  WRITE_REGISTER_ULONG((PULONG) (pdx->membase + P_DEBIAD), wdata );

	// Execute the DEBI transfer.
	DEBItransfer( pdx );
}

////////////////////////////////////////////////////////////////////////////////////////////
// Replace the specified bits in a gate array register.  Imports:  mask specifies bits
// that are to be preserved, wdata is new value to be or'd with the masked original.
#pragma LOCKEDCODE
VOID DEBIreplace( PDEVICE_EXTENSION pdx, unsigned short addr, unsigned short mask, 
                 unsigned short wdata )
{
	// Copy target gate array register into P_DEBIAD register.
	//WR7146( P_DEBICMD, DEBI_CMD_RDWORD | addr );								// Set up DEBI control reg value in shadow RAM.
  WRITE_REGISTER_ULONG((PULONG) (pdx->membase + P_DEBICMD), DEBI_CMD_RDWORD | addr);

	DEBItransfer( pdx );																// Execute the DEBI Read transfer.

	// Write back the modified image.
	//WR7146( P_DEBICMD, DEBI_CMD_WRWORD | addr );								// Set up DEBI control reg value in shadow RAM.

	//WR7146( P_DEBIAD, wdata | ( (unsigned short)RR7146( P_DEBIAD ) & mask ) );	// Modify the register image.

	//DEBItransfer( pdx );																// Execute the DEBI Write transfer.

  DEBIwrite( pdx, addr, wdata |
    (((USHORT)READ_REGISTER_ULONG((PULONG)(pdx->membase + P_DEBIAD))) & mask));
}

