//////////////////////////////////////////////////////////////
// Module    : Cls626.cpp									//
// Version   : 2.31											//
// Function  : Abstract board class for Sensoray Model 626.	//
// Author    : Jim Lamberson								//
// Copyright : (C) 2001 Sensoray							//
//////////////////////////////////////////////////////////////

#include <new>					// Enable dynamic creation/destruction of counter objects.
#include <vector>

#include <ace/OS.h>
#include "Trace.h"

#undef ACE_NTRACE
#define ACE_NTRACE 0

#include "Cls626.h"

// PCI bus interface types.
#define INTEL			1		// Intel bus type.
#define MOTOROLA		2		// Motorola bus type.

/////////////////////////////////////////////////////////////////////////////////
#define PLATFORM		INTEL			// *** SELECT PLATFORM TYPE ***
/////////////////////////////////////////////////////////////////////////////////

#define EOPL					0x80			// End of ADC poll list marker.

// Error codes that must be visible to this base class.
#define ERR_ILLEGAL_PARM		0x00010000		// Illegal function parameter value was specified.
#define ERR_I2C					0x00020000		// I2C error.
#define ERR_COUNTERSETUP		0x00200000		// Illegal setup specified for counter channel.
#define ERR_DEBI_TIMEOUT		0x00400000		// DEBI transfer timed out.

// Physical DMA buffer size in bytes.
#define DMABUF_SIZE				4096			// DMA buffer size.

// Organization (physical order) and size (in DWORDs) of logical DMA buffers contained by ANA_DMABUF.
#define ADC_DMABUF_DWORDS		40				// ADC DMA buffer must hold 16 samples, plus pre/post garbage samples.
#define DAC_WDMABUF_DWORDS		1				// DAC output DMA buffer holds a single sample.
					// All remaining space in 4KB DMA buffer is available for the RPS1 program.

// Address offsets, in DWORDS, from base of DMA buffer.
#define DAC_WDMABUF_OS			ADC_DMABUF_DWORDS

// Interrupt Enable bit in ISR and IER.
#define IRQ_MASK				0x00000040		// IRQ enable for GPIO3.
#define ISR_AFOU				0x00000800		// Audio fifo under/overflow detected.

// RPS command codes.
#define RPS_CLRSIGNAL			0x00000000		// CLEAR SIGNAL
#define RPS_SETSIGNAL			0x10000000		// SET SIGNAL
#define RPS_NOP					0x00000000		// NOP
#define RPS_PAUSE				0x20000000		// PAUSE
#define RPS_UPLOAD				0x40000000		// UPLOAD
#define RPS_JUMP				0x80000000		// JUMP
#define RPS_LDREG				0x90000100		// LDREG (1 unsigned long only)
#define RPS_STREG				0xA0000100		// STREG (1 unsigned long only)
#define RPS_STOP				0x50000000		// STOP

#define RPS_LOGICAL_OR			0x08000000		// Logical OR conditionals.
#define RPS_INVERT				0x04000000		// Test for negated semaphores.

#define RPS_DEBI				0x00000002		// DEBI done
#define RPS_SIG0				0x00200000		// RPS semaphore 0 (used by ADC).
#define RPS_SIG1				0x00400000		// RPS semaphore 1 (used by DAC).
#define RPS_SIG2				0x00800000		// RPS semaphore 2 (not used).
#define RPS_GPIO2				0x00080000		// RPS GPIO2

#define RPS_SIGADC				RPS_SIG0		// Trigger/status for ADC's RPS program.
#define RPS_SIGDAC				RPS_SIG1		// Trigger/status for DAC's RPS program.

// RPS clock parameters.
#define RPSCLK_SCALAR			8						// This is apparent ratio of PCI/RPS clks (undocumented!!).
#define RPSCLK_PER_US			( 33 / RPSCLK_SCALAR )	// Number of RPS clocks in one microsecond.

// Event counter source addresses.
#define SBA_RPS_A0				0x27			// Time of RPS0 busy, in PCI clocks.

// GPIO constants.
#define GPIO_BASE				0x10004000		// GPIO 0,2,3 = inputs, GPIO3 = IRQ; GPIO1 = out.
#define GPIO1_LO				0x00000000		// GPIO1 set to LOW.
#define GPIO1_HI				0x00001000		// GPIO1 set to HIGH.

// Primary Status Register (PSR) constants.
#define PSR_DEBI_E				0x00040000      // DEBI event flag.
#define PSR_DEBI_S				0x00080000      // DEBI status flag.
#define PSR_A2_IN				0x00008000		// Audio output DMA2 protection address reached.
#define PSR_AFOU				0x00000800		// Audio FIFO under/overflow detected.
#define PSR_GPIO2				0x00000020		// GPIO2 input pin: 0=AdcBusy, 1=AdcIdle.
#define PSR_EC0S				0x00000001		// Event counter 0 threshold reached.

// Secondary Status Register (SSR) constants.
#define SSR_AF2_OUT				0x00000200		// Audio 2 output FIFO under/overflow detected.

// Master Control Register 1 (MC1) constants.
#define MC1_SOFT_RESET			0x80000000		// Invoke 7146 soft reset.
#define MC1_SHUTDOWN			0x3FFF0000		// Shut down all MC1-controlled enables.

#define MC1_ERPS1				0x2000			// Enable/disable RPS task 1.
#define MC1_ERPS0				0x1000			// Enable/disable RPS task 0.
#define MC1_DEBI				0x0800			// Enable/disable DEBI pins.
#define MC1_AUDIO				0x0200			// Enable/disable audio port pins.
#define MC1_I2C					0x0100			// Enable/disable I2C interface.
#define MC1_A2OUT				0x0008			// Enable/disable transfer on A2 out.
#define MC1_A2IN				0x0004			// Enable/disable transfer on A2 in.
#define MC1_A1IN				0x0001			// Enable/disable transfer on A1 in.

// Master Control Register 2 (MC2) constants.
#define MC2_UPLD_DEBI			0x0002			// Upload DEBI.
#define MC2_UPLD_IIC			0x0001			// Upload I2C.
#define MC2_RPSSIG2				0x2000			// RPS signal 2 (not used).
#define MC2_RPSSIG1				0x1000			// RPS signal 1 (DAC RPS busy).
#define MC2_RPSSIG0				0x0800			// RPS signal 0 (ADC RPS busy).

#define MC2_ADC_RPS				MC2_RPSSIG0		// ADC RPS busy.
#define MC2_DAC_RPS				MC2_RPSSIG1		// DAC RPS busy.

// PCI BUS (SAA7146) REGISTER ADDRESS OFFSETS ////////////////////////////////////

#define P_PCI_BT_A				0x004C			// Audio DMA burst/threshold control.
#define P_DEBICFG               0x007C			// DEBI configuration.
#define P_DEBICMD               0x0080			// DEBI command.
#define P_DEBIPAGE              0x0084			// DEBI page.
#define P_DEBIAD                0x0088			// DEBI target address.
#define P_I2CCTRL               0x008C			// I2C control.
#define P_I2CSTAT               0x0090			// I2C status.
#define P_BASEA2_IN				0x00AC			// Audio input 2 base physical DMAbuf address.
#define P_PROTA2_IN				0x00B0			// Audio input 2 physical DMAbuf protection address.
#define P_PAGEA2_IN				0x00B4			// Audio input 2 paging attributes.
#define P_BASEA2_OUT			0x00B8			// Audio output 2 base physical DMAbuf address.
#define P_PROTA2_OUT			0x00BC			// Audio output 2 physical DMAbuf protection address.
#define P_PAGEA2_OUT			0x00C0			// Audio output 2 paging attributes.
#define P_RPSPAGE0              0x00C4			// RPS0 page.
#define P_RPSPAGE1              0x00C8			// RPS1 page.
#define P_RPS0_TOUT				0x00D4			// RPS0 time-out.
#define P_RPS1_TOUT				0x00D8			// RPS1 time-out.
#define P_IER                   0x00DC			// Interrupt enable.
#define P_GPIO                  0x00E0			// General-purpose I/O.
#define P_EC1SSR				0x00E4			// Event counter set 1 source select.
#define P_ECT1R					0x00EC			// Event counter threshold set 1.
#define P_ACON1                 0x00F4			// Audio control 1.
#define P_ACON2                 0x00F8			// Audio control 2.
#define P_MC1                   0x00FC			// Master control 1.
#define P_MC2                   0x0100			// Master control 2.
#define P_RPSADDR0              0x0104			// RPS0 instruction pointer.
#define P_RPSADDR1              0x0108			// RPS1 instruction pointer.
#define P_ISR                   0x010C			// Interrupt status.
#define P_PSR                   0x0110			// Primary status.
#define P_SSR                   0x0114			// Secondary status.
#define P_EC1R					0x0118			// Event counter set 1.
#define P_ADP4					0x0138			// Logical audio DMA pointer of audio input FIFO A2_IN.
#define P_FB_BUFFER1            0x0144			// Audio feedback buffer 1.
#define P_FB_BUFFER2            0x0148			// Audio feedback buffer 2.
#define P_TSL1                  0x0180			// Audio time slot list 1.
#define P_TSL2                  0x01C0			// Audio time slot list 2.

// LOCAL BUS (GATE ARRAY) REGISTER ADDRESS OFFSETS /////////////////////////////////

												// Analog I/O registers:
#define LP_DACPOL				0x0082			//  Write DAC polarity.
#define LP_GSEL					0x0084			//  Write ADC gain.
#define LP_ISEL					0x0086			//  Write ADC channel select.
												// Digital I/O (write only):
#define LP_WRINTSELA			0x0042			//  Write A interrupt enable.
#define LP_WREDGSELA			0x0044			//  Write A edge selection.
#define LP_WRCAPSELA			0x0046			//  Write A capture enable.
#define LP_WRDOUTA				0x0048			//  Write A digital output.
#define LP_WRINTSELB			0x0052			//  Write B interrupt enable.
#define LP_WREDGSELB			0x0054			//  Write B edge selection.
#define LP_WRCAPSELB			0x0056			//  Write B capture enable.
#define LP_WRDOUTB				0x0058			//  Write B digital output.
#define LP_WRINTSELC			0x0062			//  Write C interrupt enable.
#define LP_WREDGSELC			0x0064			//  Write C edge selection.
#define LP_WRCAPSELC			0x0066			//  Write C capture enable.
#define LP_WRDOUTC				0x0068			//  Write C digital output.
												// Digital I/O (read only):
#define LP_RDDINA				0x0040			//  Read digital input.
#define LP_RDCAPFLGA			0x0048			//  Read edges captured.
#define LP_RDINTSELA			0x004A			//  Read interrupt enable register.
#define LP_RDEDGSELA			0x004C			//  Read edge selection register.
#define LP_RDCAPSELA			0x004E			//  Read capture enable register.
#define LP_RDDINB				0x0050			//  Read digital input.
#define LP_RDCAPFLGB			0x0058			//  Read edges captured.
#define LP_RDINTSELB			0x005A			//  Read interrupt enable register.
#define LP_RDEDGSELB			0x005C			//  Read edge selection register.
#define LP_RDCAPSELB			0x005E			//  Read capture enable register.
#define LP_RDDINC				0x0060			//  Read digital input.
#define LP_RDCAPFLGC			0x0068			//  Read edges captured.
#define LP_RDINTSELC			0x006A			//  Read interrupt enable register.
#define LP_RDEDGSELC			0x006C			//  Read edge selection register.
#define LP_RDCAPSELC			0x006E			//  Read capture enable register.
												// Counter Registers (read/write):
#define LP_CR0A					0x0000			//  0A setup register.
#define LP_CR0B					0x0002			//  0B setup register.
#define LP_CR1A					0x0004			//  1A setup register.
#define LP_CR1B					0x0006			//  1B setup register.
#define LP_CR2A					0x0008			//  2A setup register.
#define LP_CR2B					0x000A			//  2B setup register.
												// Counter PreLoad (write) and Latch (read) Registers:
#define	LP_CNTR0ALSW			0x000C			//  0A lsw.
#define	LP_CNTR0AMSW			0x000E			//  0A msw.
#define	LP_CNTR0BLSW			0x0010			//  0B lsw.
#define	LP_CNTR0BMSW			0x0012			//  0B msw.
#define	LP_CNTR1ALSW			0x0014			//  1A lsw.
#define	LP_CNTR1AMSW			0x0016			//  1A msw.
#define	LP_CNTR1BLSW			0x0018			//  1B lsw.
#define	LP_CNTR1BMSW			0x001A			//  1B msw.
#define	LP_CNTR2ALSW			0x001C			//  2A lsw.
#define	LP_CNTR2AMSW			0x001E			//  2A msw.
#define	LP_CNTR2BLSW			0x0020			//  2B lsw.
#define	LP_CNTR2BMSW			0x0022			//  2B msw.
												// Miscellaneous Registers (read/write):
#define LP_MISC1				0x0088			//  Read/write Misc1.
#define LP_WRMISC2				0x0090			//  Write Misc2.
#define LP_RDMISC2				0x0082			//  Read Misc2.

// Bit masks for MISC1 register that are the same for reads and writes.
#define MISC1_WENABLE			0x8000			// Enable writes to MISC2 (except Clear Watchdog bit).
#define MISC1_WDISABLE			0x0000			// Disable writes to MISC2.
#define MISC1_EDCAP				0x1000			// Enable edge capture on DIO chans specified by LP_WRCAPSELx.
#define MISC1_NOEDCAP			0x0000			// Disable edge capture on specified DIO chans.

// Bit masks for MISC1 register reads.
#define RDMISC1_WDTIMEOUT		0x4000			// Watchdog timer timed out.

// Bit masks for MISC2 register writes.
#define WRMISC2_WDCLEAR			0x8000			// Reset watchdog timer to zero.
#define WRMISC2_CHARGE_ENABLE	0x4000			// Enable battery trickle charging.

// Bit masks for MISC2 register that are the same for reads and writes.
#define MISC2_BATT_ENABLE		0x0008			// Backup battery enable.
#define MISC2_WDENABLE			0x0004			// Watchdog timer enable.
#define MISC2_WDPERIOD_MASK		0x0003			// Watchdog interval select mask.

// Bit masks for ACON1 register.
#define A2_RUN					0x40000000				// Run A2 based on TSL2.
#define A1_RUN					0x20000000				// Run A1 based on TSL1.
#define A1_SWAP					0x00200000				// Use big-endian for A1.
#define A2_SWAP					0x00100000				// Use big-endian for A2.
#define WS_MODES				0x00019999				// WS0 = TSL1 trigger input, WS1-WS4 = CS* outputs.

#if PLATFORM == INTEL									// Base ACON1 config: always run A1 based on TSL1.
#define ACON1_BASE				( WS_MODES | A1_RUN )
#elif PLATFORM == MOTOROLA
#define ACON1_BASE				( WS_MODES | A1_RUN | A1_SWAP | A2_SWAP )
#endif

#define ACON1_ADCSTART			ACON1_BASE				// Start ADC: run A1 based on TSL1.
#define ACON1_DACSTART			( ACON1_BASE | A2_RUN )	// Start transmit to DAC: run A2 based on TSL2.
#define ACON1_DACSTOP			ACON1_BASE				// Halt A2.

// Bit masks for ACON2 register.
#define A1_CLKSRC_BCLK1			0x00000000				// A1 bit rate = BCLK1 (ADC).
#define A2_CLKSRC_X1			0x00800000				// A2 bit rate = ACLK/1 (DACs).
#define A2_CLKSRC_X2			0x00C00000				// A2 bit rate = ACLK/2 (DACs).
#define A2_CLKSRC_X4			0x01400000				// A2 bit rate = ACLK/4 (DACs).
#define INVERT_BCLK2			0x00100000				// Invert BCLK2 (DACs).
#define BCLK2_OE				0x00040000				// Enable BCLK2 (DACs).
#define ACON2_XORMASK			0x000C0000				// XOR mask for ACON2 active-low bits.

#define ACON2_INIT				( ACON2_XORMASK ^ ( A1_CLKSRC_BCLK1 | A2_CLKSRC_X2 | INVERT_BCLK2 | BCLK2_OE ) )

// Bit masks for timeslot records.
#define WS1		     			0x40000000				// WS output to assert.
#define WS2		     			0x20000000
#define WS3		     			0x10000000
#define WS4		     			0x08000000
#define RSD1					0x01000000				// Shift A1 data in on SD1.
#define SDW_A1					0x00800000				// Store rcv'd unsigned char at next unsigned char of DWORD1 buffer.
#define SIB_A1					0x00400000				// Store rcv'd unsigned char at next unsigned char of FB1 buffer.
#define SF_A1					0x00200000				// Write unsigned long buffer to input FIFO.
														// Select parallel-to-serial converter's data source:
#define XFIFO_0					0x00000000				//   Data fifo byte 0.
#define XFIFO_1					0x00000010				//   Data fifo byte 1.
#define XFIFO_2					0x00000020				//   Data fifo byte 2.
#define XFIFO_3					0x00000030				//   Data fifo byte 3.
#define XFB0					0x00000040				//   FB_BUFFER byte 0.
#define XFB1					0x00000050				//   FB_BUFFER byte 1.
#define XFB2					0x00000060				//   FB_BUFFER byte 2.
#define XFB3					0x00000070				//   FB_BUFFER byte 3.
#define SIB_A2					0x00000200				// Store next dword from A2's input shifter to FB2 buffer.
#define SF_A2					0x00000100				// Store next dword from A2's input shifter to its input fifo.
#define LF_A2					0x00000080				// Load next dword from A2's output fifo into its output dword buffer.
#define XSD2					0x00000008				// Shift data out on SD2.
#define RSD3					0x00001800				// Shift data in on SD3.
#define RSD2					0x00001000				// Shift data in on SD2.
#define LOW_A2					0x00000002				// Drive last SD low for 7 clks, then tri-state.
#define EOS		     			0x00000001				// End of superframe.

// Function macros -----------------------------------------------------------------------

// Write/Read unsigned long to/from SAA7146 device register.
#define WR7146( REGADRS, VAL )			WriteMem( (unsigned long)(MyBase + (REGADRS)), (VAL) )
#define RR7146( REGADRS )						ReadMem ( (unsigned long)(MyBase + (REGADRS))        )

// Enable/disable a function or test status bit(s) that are accessed through Main Control Registers 1 or 2.
#define MC_ENABLE( REGADRS, CTRLWORD )			WR7146( ( REGADRS ), ( (unsigned long)( CTRLWORD ) << 16 ) | (unsigned long)( CTRLWORD ) )
#define MC_DISABLE( REGADRS, CTRLWORD )			WR7146( ( REGADRS ), (unsigned long)( CTRLWORD ) << 16 )
#define MC_TEST( REGADRS, CTRLWORD )			( ( RR7146( REGADRS ) & CTRLWORD ) != 0 )

// Write a time slot control record to TSL2.
#define VECTPORT( VECTNUM )						( P_TSL2 + ( (VECTNUM) << 2 ) )
#define SETVECT( VECTNUM, VECTVAL )				WR7146( VECTPORT( VECTNUM ), (VECTVAL) )

// Validate a function argument by comparing to maximum legal value.
#define ValidateParm( PARM, MAXVAL )			if ( ( PARM ) > ( MAXVAL ) ) { DeclareError( ERR_ILLEGAL_PARM ); return; }
#define ValidateParmRtnval( PARM, MAXVAL )		if ( ( PARM ) > ( MAXVAL ) ) { DeclareError( ERR_ILLEGAL_PARM ); return 0; }

// Function prototypes for non-member functions defined in this module -------------------

void DummyErrHandler( unsigned long ErrMask );

//////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  STARTUP & SHUTDOWN FUNCTIONS  //////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////
// Board class constructor/destructor.

Cls626::Cls626()
{
	// Initialize class storage elements.
	MyAddress		= 0;				          // Reset board physical address.
  MyBase      = 0;
	MyErrFlags		= 0;				        // Clear all error flags.
	MyErrCallback	= DummyErrHandler;	// Init pointer to error handler function.

  MyDacpol = 0;

	// Create counter objects.
	CreateCounters();
}

Cls626::~Cls626()
{
	// Destroy all counter objects.
	DestroyCounters();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// Board class initializer -- THIS MUST BE CALLED BY THE OS INTERFACE CLASS CONSTRUCTOR.
// These functions are not performed by the board class constructor because the derived OS interface
// class, which is based on this board class, must complete its own self-initialization before any
// hardware accesses can be legally performed.
//
// Imports:  IsBoardRevA is a boolean that indicates whether the associated Model 626 board is RevA.
//////////////////////////////////////////////////////////////////////////////////////////////////////

void Cls626::BoardOpen( unsigned char IsBoardRevA )
{
	// Start thread-safe execution.
	CriticalBegin();

	// Disable interrupts.
	InterruptMask();

	// Invoke SAA7146 soft reset.
	WR7146( P_MC1, MC1_SOFT_RESET );

	// Allocate DMA buffers using AllocDMAB(), a pure virtual function provided by the OS Interface class.
	// Each call to AllocDMAB() allocates physically contiguous buffer space at least as large as the
	// requested size.  The buffer must begin on a page boundary, and pages must contain at least 4KB
	// of contiguous memory.  This class (and the AllocDMAB() function) must be modified if the target
	// OS can not provide the minimum 4KB physical page size.  AllocDMAB() sets the "allocation error"
	// flag in MyErrFlags if any buffer allocation fails.
	AllocDMAB( DMABUF_SIZE, &MyANABuf );		// Allocate buffer to receive ADC data.
	AllocDMAB( DMABUF_SIZE, &MyRPSBuf );		// Allocate buffer for RPS1 commands that will control the ADC.

	// Abort startup if DMA buffers couldn't be allocated.
	if ( MyErrFlags )
		return;

	// Enable DEBI and audio pins, enable I2C interface.
	MC_ENABLE( P_MC1, MC1_DEBI | MC1_AUDIO | MC1_I2C );

	// Init DEBI functional configuration used for all DEBI operations.
	DEBIinit();

	// Init GPIO so that ADC Start* is negated.
	WR7146( P_GPIO, GPIO_BASE | GPIO1_HI );

	// VERSION 2.01 CHANGE: REV A & B BOARDS NOW SUPPORTED BY DYNAMIC EEPROM ADDRESS SELECTION.
	// Initialize the I2C interface, which is used to access the onboard serial EEPROM.  The EEPROM's I2C
	// DeviceAddress is hardwired to a value that is dependent on the 626 board revision.  On all board
	// revisions, the EEPROM stores TrimDAC calibration constants for analog I/O.  On RevB and higher
	// boards, the DeviceAddress is hardwired to 0 to enable the EEPROM to also store the PCI SubVendorID
	// and SubDeviceID; this is the address at which the SAA7146 expects a configuration EEPROM to reside.
	// On RevA boards, the EEPROM device address, which is hardwired to 4, prevents the SAA7146 from
	// retrieving PCI sub-IDs, so the SAA7146 uses its built-in default values, instead.
	MyI2CAdrs = IsBoardRevA ? 0xA8 : 0xA0;		// Set I2C EEPROM DeviceType (0xA0) and DeviceAddress<<1.
	I2CInit();									// Initialize I2C interface.

	// Init audio interface functional attributes:  set DAC/ADC serial clock rates, invert DAC
	// serial clock so that DAC data setup times are satisfied, enable DAC serial clock out.
	WR7146( P_ACON2, ACON2_INIT );

	// Init the ADC interface.
	InitADC();

	// Init the DAC interface.
	InitDAC();

	// Init Trim DACs to calibrated values.  Do it twice because the SAA7146 audio channel does
	// not always reset properly and sometimes causes the first few TrimDAC writes to malfunction.
	LoadTrimDACs();
	LoadTrimDACs();		// Insurance.

	/////////////////////////////////////////////////////////////////////////////////////////////
	// Manually init all gate array hardware in case this is a soft reset (we have no way of
	// determining whether this is a warm or cold start).  This is necessary because the gate
	// array will reset only in response to a PCI hard reset; there is no soft reset function.

	// Init all DAC outputs to 0V and init all DAC setpoint and polarity images.
	for ( unsigned short chan = 0; chan < 4; chan++)
		SetDAC( chan, 0 );

	// Init image of WRMISC2 Battery Charger Enabled control bit.  This image is used when the
	// state of the charger control bit, which has no direct hardware readback mechanism, is queried.
	MyChargeEnabled = 0;

	// Init image of watchdog timer interval in WRMISC2.  This image maintains the value of the
	// control bits of MISC2 are continuously reset to zero as long as the WD timer is disabled.
	MyWDInterval = 0;

	// Init Counter Interrupt Enable mask for RDMISC2.  This mask is applied against MISC2 when
	// testing to determine which timer events are requesting interrupt service.
	MyCounterIntEnabs = 0;

	// Init counter objects.
	CountersInit();

	// Without modifying the state of the Battery Backup Enable, disable the watchdog timer, set
	// DIO channels 0-5 to operate in the standard DIO (vs. counter overflow) mode, disable
	// the battery charger, and reset the watchdog interval selector to zero.
	WriteMISC2( DEBIread( LP_RDMISC2 ) & MISC2_BATT_ENABLE );

	// Initialize the digital I/O subsystem.
	DIOInit();

	// End of thread-safe segment.
	CriticalEnd();
}

////////////////////////////////////////////////////////////////////////////
// Board class shutdown -- MUST BE CALLED BY INTERFACE CLASS DESTRUCTOR.

void Cls626::BoardClose()
{
	// Disable interrupts.
	InterruptMask();

	// Disable the watchdog timer and battery charger.
	WriteMISC2( 0 );

	// Close all interfaces on 7146 device.
	WR7146( P_MC1, MC1_SHUTDOWN );
	WR7146( P_ACON1, ACON1_BASE );

	// Release memory used by DMA buffers.
	CloseDMAB( &MyRPSBuf );
	CloseDMAB( &MyANABuf );
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  STATUS AND ERROR FUNCTIONS  ///////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////
// Return board PCI bus/slot numbers.

unsigned long Cls626::GetAddress()
{
	return MyAddress;
}

/////////////////////////////////////////////////////////////
// Enable application's error handler callback function.

void Cls626::SetErrCallback( FPTR_ERR Callback )
{
	MyErrCallback = Callback;
}

/////////////////////////////////////////////////////////////
// Return error flags and clear any resettable error flags.

unsigned long Cls626::GetErrors()
{
	// Start critical segment.
	CriticalBegin();

	// Cache return value.
	unsigned long ErrFlags = MyErrFlags;

	// Clear any resettable error flags.
	MyErrFlags &= 0x0000FFFF;

	// End critical segment.
	CriticalEnd();

	// return original error flags.
	return ErrFlags;
}

/////////////////////////////////////////////////////////////
// Set one or more error flags.

void Cls626::DeclareError( unsigned long ErrMask )
{
	// Start critical segment.
	CriticalBegin();

	// Set error flag bits as specified by ErrMask.
	MyErrFlags |= ErrMask;

	// Call the application's error handler.  If the application has not yet declared an error
	// handler via SetErrCallback(), the DummyErrHandler() function is executed by default.
	MyErrCallback( MyErrFlags );

	// End critical segment.
	CriticalEnd();
}

/////////////////////////////////////////////////////////////////////////////////////
// This dummy error handler is substituted for the application's callback function
// if the application has not yet called SetErrCallback() to declare a callback.
// Note: this function is not a class member because it is expected to be cdecl.

static void DummyErrHandler( unsigned long ErrMask )	{}

//////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  EEPROM ACCESS FUNCTIONS  //////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

// I2C configuration constants.
#define I2C_CLKSEL			0x0400							// I2C bit rate = PCIclk/480 = 68.75 KHz.
#define I2C_BITRATE			68.75							// I2C bus data bit rate (determined by I2C_CLKSEL) in KHz.
#define I2C_WRTIME			15.0							// Worst case time,in msec, for EEPROM internal write op.

// I2C manifest constants.
#define I2C_RETRIES			( I2C_WRTIME * I2C_BITRATE / 9.0 )	// Max retries to wait for EEPROM write.
#define I2C_ERR				0x0002							// I2C control/status flag ERROR.
#define I2C_BUSY			0x0001							// I2C control/status flag BUSY.
#define I2C_ABORT			0x0080							// I2C status flag ABORT.
#define I2C_ATTRSTART		0x3								// I2C unsigned char attribute START.
#define I2C_ATTRCONT		0x2								// I2C unsigned char attribute CONT.
#define I2C_ATTRSTOP		0x1								// I2C unsigned char attribute STOP.
#define I2C_ATTRNOP			0x0								// I2C unsigned char attribute NOP.
#define I2CR				( MyI2CAdrs | 1 )				// I2C read command  | EEPROM address.
#define I2CW				MyI2CAdrs						// I2C write command | EEPROM address.

// Code macros used for constructing I2C command bytes.
#define I2C_B2(ATTR,VAL)	( ( (ATTR) << 6 ) | ( (VAL) << 24 ) )
#define I2C_B1(ATTR,VAL)	( ( (ATTR) << 4 ) | ( (VAL) << 16 ) )
#define I2C_B0(ATTR,VAL)	( ( (ATTR) << 2 ) | ( (VAL) <<  8 ) )

///////////////////////////////////////////
// Read unsigned char from EEPROM.

unsigned char Cls626::I2Cread( unsigned char addr )
{
	// Start critical segment.
	CriticalBegin();

	// Send EEPROM target address.
	if ( I2Chandshake(
		  I2C_B2( I2C_ATTRSTART, I2CW )		// Byte2 = I2C command: write to I2C EEPROM device.
		| I2C_B1( I2C_ATTRSTOP,  addr )		// Byte1 = EEPROM internal target address.
		| I2C_B0( I2C_ATTRNOP,   0    ) ) )	// Byte0 = Not sent.
	{
		// Abort function and declare error if handshake failed.
		CriticalEnd();
		DeclareError( ERR_I2C );
		return 0;
	}

	// Execute EEPROM read.
	if ( I2Chandshake(
		  I2C_B2( I2C_ATTRSTART, I2CR )		// Byte2 = I2C command: read from I2C EEPROM device.
		| I2C_B1( I2C_ATTRSTOP,  0    )		// Byte1 receives unsigned char from EEPROM.
		| I2C_B0( I2C_ATTRNOP,   0    ) ) )	// Byte0 = Not sent.
	{
		// Abort function and declare error if handshake failed.
		CriticalEnd();
		DeclareError( ERR_I2C );
		return 0;
	}

	// Return copy of EEPROM value.
	unsigned char rtnval = (unsigned char)( RR7146(P_I2CCTRL) >> 16 );
	CriticalEnd();
	return rtnval;
}

////////////////////////////////////////////
// Write unsigned char to EEPROM.

void Cls626::I2Cwrite( unsigned char addr, unsigned char val )
{
	// Start critical segment.
	CriticalBegin();

	// Send EEPROM target address and new data value and invoke internal EEPROM write operation.
	if ( I2Chandshake(
		  I2C_B2( I2C_ATTRSTART, I2CW )		// Byte2 = I2C command: write to I2C EEPROM device.
		| I2C_B1( I2C_ATTRCONT,  addr )		// Byte1 = EEPROM internal target address.
		| I2C_B0( I2C_ATTRSTOP,  val  ) ) )	// Byte0 = New EEPROM data value.
	{
		// Abort function and declare error if handshake failed.
		CriticalEnd();
		DeclareError( ERR_I2C );
		return;
	}

	// Wait for the EEPROM to finish its internal write operation.  This is done by polling with
	// dummy write commands until the EEPROM responds by ACKing the I2C commands being sent to it.
	// A retry counter is used so that we won't hang if there is an I2C hardware fault.
	for ( unsigned long retries = 0; retries < (unsigned long)I2C_RETRIES; retries++ )
	{
		// Halt any pending I2C operations and clear any I2C errors.
		I2CInit();

		// Attempt to handshake a command to the EEPROM.  The handshake will fail
		// if the EEPROM is still executing its internal write operation.
		if ( !I2Chandshake(
			  I2C_B2( I2C_ATTRSTART, I2CW )			// Byte2 = I2C command: write to I2C EEPROM device.
			| I2C_B1( I2C_ATTRSTOP,  addr )			// Byte1 = EEPROM internal target address.
			| I2C_B0( I2C_ATTRNOP,   0    ) ) )		// Byte0 = Not sent.
		{
			CriticalEnd();
			return;			// Exit if EEPROM write operation is finished.
		}
	}

	// Abort function and declare error if handshake failed.
	CriticalEnd();
	DeclareError( ERR_I2C );
}

///////////////////////////////////////////////////////////////////////////////////
// Private helper function: handshake value to EEPROM.  Returns 0 if successful.

unsigned long Cls626::I2Chandshake( unsigned long val )
{
	// Write I2C command to I2C Transfer Control shadow register.
	WR7146( P_I2CCTRL, val );

	// Upload I2C shadow registers into working registers and wait for upload confirmation.
	MC_ENABLE( P_MC2, MC2_UPLD_IIC );
	while ( !MC_TEST( P_MC2, MC2_UPLD_IIC ) );

	// Wait until I2C bus transfer is finished or an error occurs.
	while ( ( RR7146(P_I2CCTRL) & ( I2C_BUSY | I2C_ERR ) ) == I2C_BUSY );
	
	// Return non-zero if I2C error occured.
	return RR7146(P_I2CCTRL) & I2C_ERR;
}

//////////////////////////////////////////////////////////
// Private helper function: initialize I2C interface.

void Cls626::I2CInit()
{
	// Issue an I2C ABORT command to halt any I2C operation in progress and reset BUSY flag.
	WR7146( P_I2CSTAT, I2C_CLKSEL | I2C_ABORT );			// Write I2C control: abort any I2C activity.
	MC_ENABLE( P_MC2, MC2_UPLD_IIC );						// Invoke command upload
	while ( ( RR7146(P_MC2) & MC2_UPLD_IIC ) == 0 );		//   and wait for upload to complete.
	
	// Per SAA7146 data sheet, write to STATUS reg twice to reset all I2C error flags.
	for ( register unsigned short i = 0; i < 2; i++ )
	{
		WR7146( P_I2CSTAT, I2C_CLKSEL );					// Write I2C control: reset error flags.
		MC_ENABLE( P_MC2, MC2_UPLD_IIC );					// Invoke command upload
		while ( !MC_TEST( P_MC2, MC2_UPLD_IIC ) );			//   and wait for upload to complete.
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////  DEBI (GATE ARRAY REGISTER ACCESS) FUNCTIONS  //////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////

// DEBI command constants.
#define DEBI_CMD_SIZE16			( 2 << 17 )			// Transfer size is 2 bytes.
#define DEBI_CMD_READ			  0x00010000			// Read operation.
#define DEBI_CMD_WRITE			0x00000000			// Write operation.

#define DEBI_CMD_RDWORD			( DEBI_CMD_READ  | DEBI_CMD_SIZE16 )	// Read immediate 2 bytes.
#define DEBI_CMD_WRWORD			( DEBI_CMD_WRITE | DEBI_CMD_SIZE16 )	// Write immediate 2 bytes.

// DEBI configuration constants.
#define DEBI_CFG_XIRQ_EN		0x80000000			// Enable external interrupt on GPIO3.
#define DEBI_CFG_XRESUME		0x40000000			// Resume block transfer when XIRQ deasserted.
#define DEBI_CFG_FAST			  0x10000000			// Fast mode enable.
													// 4-bit field that specifies DEBI timeout value in PCI clock cycles:
#define DEBI_CFG_TOUT_BIT		22					//   Finish DEBI cycle after this many clocks.
													// 2-bit field that specifies Endian byte lane steering:
#define DEBI_CFG_SWAP_NONE	0x00000000			//   Straight - don't swap any bytes (Intel).
#define DEBI_CFG_SWAP_2			0x00100000			//   2-byte swap (Motorola).
#define DEBI_CFG_SWAP_4			0x00200000			//   4-byte swap.
#define DEBI_CFG_SLAVE16		0x00080000			// Slave is able to serve 16-bit cycles.
#define DEBI_CFG_INC			  0x00040000			// Enable address increment for block transfers.
#define DEBI_CFG_INTEL			0x00020000			// Intel style local bus.
#define DEBI_CFG_TIMEROFF		0x00010000			// Disable timer.

#if PLATFORM == INTEL

#define DEBI_TOUT				7					// Wait 7 PCI clocks (212 ns) before polling RDY.
#define DEBI_SWAP				DEBI_CFG_SWAP_NONE	// Intel byte lane steering (pass through all byte lanes).

#elif PLATFORM == MOTOROLA

#define DEBI_TOUT				15					// Wait 15 PCI clocks (454 ns) maximum before timing out.
#define DEBI_SWAP				DEBI_CFG_SWAP_2		// Motorola byte lane steering.

#endif

// DEBI page table constants.
#define DEBI_PAGE_DISABLE		0x00000000			// Paging disable.

////////////////////////////////////////////////////////////////////////////////////
// Initialize the DEBI interface for all transfers.

void Cls626::DEBIinit()
{
	// Configure DEBI operating mode.
	WR7146( P_DEBICFG,  DEBI_CFG_SLAVE16			// Local bus is 16 bits wide.
		| ( DEBI_TOUT << DEBI_CFG_TOUT_BIT )		// Declare DEBI transfer timeout interval.
		| DEBI_SWAP									// Set up byte lane steering.
		| DEBI_CFG_INTEL );							// Intel-compatible local bus (DEBI never times out).

	// Paging is disabled.
	WR7146( P_DEBIPAGE, DEBI_PAGE_DISABLE );		// Disable MMU paging.
}

////////////////////////////////////////////////////////////////////////////////////
// Execute a DEBI transfer.  This must be called from within a critical section.

void Cls626::DEBItransfer()
{
	// Initiate upload of shadow RAM to DEBI control register.
	MC_ENABLE( P_MC2, MC2_UPLD_DEBI );

	// Wait for completion of upload from shadow RAM to DEBI control register.
	while ( !MC_TEST( P_MC2, MC2_UPLD_DEBI ) );

	// Wait until DEBI transfer is done.
	while ( RR7146(P_PSR) & PSR_DEBI_S );
}

/////////////////////////////////////////////////////////////////
// Return a value from a gate array register.

unsigned short Cls626::DEBIread( unsigned short addr )
{
	// Start critical segment.
	CriticalBegin();

	// Set up DEBI control register value in shadow RAM.
	WR7146( P_DEBICMD, DEBI_CMD_RDWORD | addr );

	// Execute the DEBI transfer.
	DEBItransfer();

	// Fetch target register value.
	unsigned short retval = (unsigned short)RR7146( P_DEBIAD );

	// End critical segment.
	CriticalEnd();

	// Return register value.
	return retval;
}

/////////////////////////////////////////////////////////////////////////
// Write a value to a gate array register.

void Cls626::DEBIwrite( unsigned short addr, unsigned short wdata )
{
	// Start critical segment.
	CriticalBegin();

	// Set up DEBI control register value in shadow RAM.
	WR7146( P_DEBICMD, DEBI_CMD_WRWORD | addr );
	WR7146( P_DEBIAD,  wdata );

	// Execute the DEBI transfer.
	DEBItransfer();

	// End critical segment.
	CriticalEnd();
}

////////////////////////////////////////////////////////////////////////////////////////////
// Replace the specified bits in a gate array register.  Imports:  mask specifies bits
// that are to be preserved, wdata is new value to be or'd with the masked original.

void Cls626::DEBIreplace( unsigned short addr, unsigned short mask, unsigned short wdata )
{
	// Start critical segment.
	CriticalBegin();

	// Copy target gate array register into P_DEBIAD register.
	WR7146( P_DEBICMD, DEBI_CMD_RDWORD | addr );								// Set up DEBI control reg value in shadow RAM.
	DEBItransfer();																// Execute the DEBI Read transfer.

	// Write back the modified image.
	WR7146( P_DEBICMD, DEBI_CMD_WRWORD | addr );								// Set up DEBI control reg value in shadow RAM.
	WR7146( P_DEBIAD, wdata | ( (unsigned short)RR7146( P_DEBIAD ) & mask ) );	// Modify the register image.
	DEBItransfer();																// Execute the DEBI Write transfer.

	// End critical segment.
	CriticalEnd();
}

//////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////  INTERRUPT FUNCTIONS  ///////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////
// IntActions is the interrupt handler action list for the OS interface class.  This array,
// which is not a member of Cls626, returns a list of Write actions that must take place at the
// beginning of the ISR.  It is used by the OS interface class to set up all kernel-mode driver
// accesses that must occur at the beginning of interrupt service.  Whether to use IntActions
// versus InterruptEnable() depends on the the target OS.  InterruptEnable() is used only if the
// target OS supports direct register access during initial interrupt handling.  If the OS does
// not support this, IntActions must be used instead of InterruptEnable().

// Actions used to negate IRQ at start of hardware interrupt service:

#define	ACT_READ	0
#define ACT_WRITE	1

INT_ACTIONS IntActions = {
	IRQ_MASK,									/* AND mask for IRQ bit in status register.	*/
	0,											/* XOR mask for IRQ bit in status register.	*/

	{	1,										/* 1 action used to retrieve IRQ status.	*/
		{										/*   Actions:								*/
			{ ACT_READ,  P_ISR, 0		 }		/*     Fetch IRQ Pending status.  			*/
		}
	},

	{	2,										/* 2 actions used to negate IRQ.			*/
		{										/*   Actions:								*/
			{ ACT_WRITE, P_IER, 0        },		/*     Disable interrupt.					*/
			{ ACT_WRITE, P_ISR, IRQ_MASK }		/*     Clear IRQ.							*/
		}
	}
};

/////////////////////////////////////////////////////////////////
// Enable master hardware interrupt.

void Cls626::InterruptEnable( unsigned short Enable )
{
	if ( Enable )
		WR7146( P_IER, IRQ_MASK );		// Enable master interrupt.
	else
		InterruptMask();				// Disable master interrupt.
}

/////////////////////////////////////////////////////////////////
// Return state of master hardware interrupt's enable.

unsigned short Cls626::InterruptEnableGet()
{
	return ( ( RR7146( P_IER ) & IRQ_MASK ) != 0 );		// Return non-zero if master interrupt is enabled.
}

/////////////////////////////////////////////////////////////////
// Disable master hardware interrupt and clear any pending IRQ.

void Cls626::InterruptMask()
{
	CriticalBegin();					// Start critical segment.
	WR7146( P_IER, 0 );					// Disable master interrupt.
	WR7146( P_ISR, IRQ_MASK );			// Clear board's master IRQ status flag.
	CriticalEnd();						// End critical segment.
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Return non-zero value if the master interrupt service request flag is asserted, independent
// of whether the master IRQ is enabled.

unsigned short Cls626::InterruptPending()
{
	// Return non-zero if GPIO3 (which is the gate-array's active-high Interrupt Service Request
	// signal) is asserted.
	return ( ( RR7146( P_PSR ) & IRQ_MASK ) != 0 );
}

//////////////////////////////////////////////////////////////////////////////////////////
// Return interrupt request status from all IRQ sources.  An array of four words is
// returned.  The first three words indicate IRQs on the three DIO channel groups:  a
// "1" in a bit position indicates that an IRQ is pending on the corresponding channel.
// The fourth unsigned short indicates IRQs on the six counter channels in the bit order defined
// by the RDMISC2 register:  a "1" in a bit position indicated a pending IRQ for the
// associated overflow or index event.  This function provides a high-performance
// mechanism for interrupt handlers to determine IRQ status of DIO/counter channels.

void Cls626::InterruptStatus( unsigned short *IntFlags )
{
	// Start critical segment.
	CriticalBegin();

	// Get all DIO IRQs.  These are derived by masking DIO events captures against the DIO interrupt
	// enables that are stored in the readable interrupt enable registers.
	*IntFlags++ = DEBIread( LP_RDCAPFLGA ) & DEBIread( LP_RDINTSELA );
	*IntFlags++ = DEBIread( LP_RDCAPFLGB ) & DEBIread( LP_RDINTSELB );
	*IntFlags++ = DEBIread( LP_RDCAPFLGC ) & DEBIread( LP_RDINTSELC );

	// Get all counter IRQs.  These are derived by masking counter event captures (in the RDMISC2
	// register) against the interrupt enables stored in the MyCounterIntEnabs image.
	*IntFlags = DEBIread( LP_RDMISC2 ) & MyCounterIntEnabs;

	// End critical segment.
	CriticalEnd();
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////  ADC FUNCTIONS  ////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

// SAA7146 BUG WORKAROUND
// According to Philips Errata Info, Ref 1.1, dated 16 Jan 1998, the RPS STREG command must
// specify a register address offset that is 4 less than the target address offset.

#define BUGFIX_STREG(REGADRS)	( REGADRS - 4 )

/////////////////////////////////////////////////////////////////////////////
// Initialize ADC.

void Cls626::InitADC()
{
	// Set up TSL1 slot list, which is used to control the accumulation of ADC data:
	//  RSD1   = shift data in on SD1.
	//  SIB_A1 = store data unsigned char at next available location in FB BUFFER1 register.
	WR7146( P_TSL1    , RSD1 | SIB_A1 );							// Fetch ADC high data unsigned char.
	WR7146( P_TSL1 + 4, RSD1 | SIB_A1 | EOS );						// Fetch ADC low data unsigned char; end of TSL1.

	// Enable TSL1 slot list so that it executes all the time.
	WR7146( P_ACON1, ACON1_ADCSTART );

	// Initialize RPS registers used for ADC.
	WR7146( P_RPSADDR1, (unsigned long)MyRPSBuf.PhysicalBase );		// Physical start of RPS program.
	WR7146( P_RPSPAGE1, 0 );										// RPS program performs no explicit mem writes.
	WR7146( P_RPS1_TOUT, 0 );										// Disable RPS timeouts.

	// SAA7146 BUG WORKAROUND.
	// Initialize SAA7146 ADC interface to a known state by invoking ADCs until FB BUFFER 1 register
	// shows that it is correctly receiving ADC data.  This is necessary because the SAA7146 ADC
	// interface does not start up in a defined state after a PCI reset.

	unsigned char PollList = EOPL;									// Create a simple polling list for analog input channel 0.
	ResetADC( &PollList );

	unsigned short AdcData;											// Get initial ADC value.
	ReadADC( &AdcData );
	unsigned short StartVal = AdcData;

	// VERSION 2.01 CHANGE: TIMEOUT ADDED TO PREVENT HANGED EXECUTION.
	// Invoke ADCs until the new ADC value differs from the initial value or a timeout occurs.  The
	// timeout protects against the possibility that the driver is restarting and the ADC data is a
	// fixed value resulting from the applied ADC analog input being unusually quiet or at the rail.
	for ( unsigned short i = 0; i < 500; i++ )
	{
		ReadADC( &AdcData );
		if ( AdcData != StartVal )
			break;
	}
}

////////////////////////////////////////////////////////////////////////////////////
// Execute ADC poll list scan and return ADC data in application pdata[] buffer.

void Cls626::StartADC()
{
	// Start ADC critical segment.
	CriticalBeginAdc();

	// Trigger ADC scan loop start by setting RPS Signal 0.
	MC_ENABLE( P_MC2, MC2_ADC_RPS );
}

void Cls626::WaitDoneADC( unsigned short *pdata )
{
	// Wait until ADC scan loop is finished (RPS Signal 0 reset).
	while ( MC_TEST( P_MC2, MC2_ADC_RPS ) );

	// Init ptr to DMA buffer that holds new ADC data.  We skip the first unsigned short in the buffer
	// because it contains junk data from the final ADC of the previous poll list scan.
	//register unsigned long *readaddr = (unsigned long *)MyANABuf.LogicalBase + 1;

	// Convert ADC data to 16-bit integer values and copy to application buffer.
	//for ( unsigned char i = 0; i < MyAdcItems; i++ )
	//	*pdata++ = (unsigned short)( ReadMem( readaddr++ ) >> 16 );

  // DMA read operation temporary buffer; retrieves results as unsigned longs
  std::vector< unsigned long > ulongBuf( MyAdcItems );

  // We skip the first unsigned short (really unsigned long) in the DMA buffer 
  //  because it contains junk data from the final ADC of the previous poll list scan.
  ReadDmaMem( MyANABuf, (PUCHAR)(&ulongBuf[0]), 
    (ULONG)(sizeof(unsigned long) * ulongBuf.size()), 0, 1 * sizeof(unsigned long) );

  // Convert ADC data to 16-bit integer values and copy to application buffer.
	for ( unsigned char i = 0; i < MyAdcItems; i++ )
		*pdata++ = (unsigned short)( ulongBuf[i] >> 16 );

	// End ADC critical segment.
	CriticalEndAdc();
}

void Cls626::ReadADC( unsigned short *pdata )
{
	// Initiate ADC scan.
	StartADC();

	// Wait for ADC scan to finish.
	WaitDoneADC( pdata );
}    

////////////////////////////////////////////////////////////////////////////
// Prepare for execution of the ADC poll list.

#define GSEL_BIPOLAR5V		0x00F0		// LP_GSEL setting for 5V bipolar range.
#define GSEL_BIPOLAR10V		0x00A0		// LP_GSEL setting for 10V bipolar range.

void Cls626::ResetADC( unsigned char *ppl )
{
	// Start critical segment.
	CriticalBegin();

	// Stop RPS program in case it is currently running.
	MC_DISABLE( P_MC1, MC1_ERPS1 );

	// Set starting logical address to write RPS commands.
	//register unsigned long *pRPS = (unsigned long *)MyRPSBuf.LogicalBase;
  std::vector< unsigned long > dmaSetupBuffer( MyRPSBuf.bufByteSize / sizeof(unsigned long) );
  register std::vector< unsigned long >::iterator pRPS = dmaSetupBuffer.begin();

	// Construct RPS program in MyRPSBuf DMA buffer ----------------------------------------

// warning C4311: 'type cast' : 
// pointer truncation from 'void *' (size=64 bit) to 'unsigned long (size=32 bit)'
// 64 bit physical addresses aren't possible as the board can only handle 32 bit addresses
#pragma warning( disable : 4311 )
  const unsigned long anaPhyBaseAddr = (unsigned long)MyANABuf.PhysicalBase;
  const unsigned long rpsPhyBaseAddr = (unsigned long)MyRPSBuf.PhysicalBase;
#pragma warning( default : 4311)
	
	// Wait for Start trigger from ReadADC().
  *pRPS++ = RPS_PAUSE | RPS_SIGADC;
	//WriteMem( pRPS++, RPS_PAUSE | RPS_SIGADC );

	// SAA7146 BUG WORKAROUND
	// Do a dummy DEBI Write.  This is necessary because the first RPS DEBI Write following a
	// non-RPS DEBI write seems to always fail.  If we don't do this dummy write, the ADC gain
	// might not be set to the value required for the first slot in the poll list; the ADC
	// gain would instead remain unchanged from the previously programmed value.
	//WriteMem( pRPS++, RPS_LDREG | (P_DEBICMD >> 2) );			// Write DEBI Write command and address to shadow RAM.
  *pRPS++ = ( RPS_LDREG | (P_DEBICMD >> 2) );
	//WriteMem( pRPS++, DEBI_CMD_WRWORD | LP_GSEL );
  *pRPS++ = ( DEBI_CMD_WRWORD | LP_GSEL );
	//WriteMem( pRPS++, RPS_LDREG | (P_DEBIAD >> 2) );			// Write DEBI immediate data to shadow RAM:
  *pRPS++ = ( RPS_LDREG | (P_DEBIAD >> 2) );
	//WriteMem( pRPS++, GSEL_BIPOLAR5V );							//   arbitrary immediate data value.
  *pRPS++ = ( GSEL_BIPOLAR5V );
	//WriteMem( pRPS++, RPS_CLRSIGNAL | RPS_DEBI );				// Reset "shadow RAM uploaded" flag.
  *pRPS++ = ( RPS_CLRSIGNAL | RPS_DEBI );
	//WriteMem( pRPS++, RPS_UPLOAD    | RPS_DEBI );				// Invoke shadow RAM upload.
  *pRPS++ = ( RPS_UPLOAD | RPS_DEBI );
	//WriteMem( pRPS++, RPS_PAUSE     | RPS_DEBI );				// Wait for shadow upload to finish.
  *pRPS++ = ( RPS_PAUSE | RPS_DEBI );				

	// Digitize all slots in the poll list.  This is implemented as a for loop to limit the slot
	// count to 16 in case the application forgot to set the EOPL flag in the final slot.
	for ( MyAdcItems = 0; MyAdcItems < 16; MyAdcItems++ )
	{
		// Convert application's poll list item to private board class format.  Each app poll list item
		// is an unsigned char with form (EOPL,x,x,RANGE,CHAN<3:0>), where RANGE code indicates 0 = +-10V, 1 = +-5V,
		// and EOPL = End of Poll List marker.
		unsigned long LocalPPL = ( *ppl << 8 ) | ( *ppl & 0x10 ? GSEL_BIPOLAR5V : GSEL_BIPOLAR10V );

		// Switch ADC analog gain.
		//WriteMem( pRPS++, RPS_LDREG | (P_DEBICMD >> 2) );		// Write DEBI command and address to shadow RAM.
    *pRPS++ = ( RPS_LDREG | (P_DEBICMD >> 2) );
		//WriteMem( pRPS++, DEBI_CMD_WRWORD | LP_GSEL );
    *pRPS++ = ( DEBI_CMD_WRWORD | LP_GSEL );
		//WriteMem( pRPS++, RPS_LDREG | (P_DEBIAD >> 2) );		// Write DEBI immediate data to shadow RAM:
    *pRPS++ = ( RPS_LDREG | (P_DEBIAD >> 2) );
		//WriteMem( pRPS++, LocalPPL );
    *pRPS++ = ( LocalPPL );
		//WriteMem( pRPS++, RPS_CLRSIGNAL | RPS_DEBI );			// Reset "shadow RAM uploaded" flag.
    *pRPS++ = ( RPS_CLRSIGNAL | RPS_DEBI );
		//WriteMem( pRPS++, RPS_UPLOAD    | RPS_DEBI );			// Invoke shadow RAM upload.
    *pRPS++ = ( RPS_UPLOAD | RPS_DEBI );
		//WriteMem( pRPS++, RPS_PAUSE     | RPS_DEBI );			// Wait for shadow upload to finish.
    *pRPS++ = ( RPS_PAUSE | RPS_DEBI );

		// Select ADC analog input channel.
		//WriteMem( pRPS++, RPS_LDREG | (P_DEBICMD >> 2) );		// Write DEBI command and address to shadow.
    *pRPS++ = ( RPS_LDREG | (P_DEBICMD >> 2) );
		//WriteMem( pRPS++, DEBI_CMD_WRWORD | LP_ISEL );
    *pRPS++ = ( DEBI_CMD_WRWORD | LP_ISEL );
		//WriteMem( pRPS++, RPS_LDREG | (P_DEBIAD >> 2) );		// Write DEBI immediate data to shadow.
    *pRPS++ = ( RPS_LDREG | (P_DEBIAD >> 2) );		
		//WriteMem( pRPS++, LocalPPL );
    *pRPS++ = ( LocalPPL );
		//WriteMem( pRPS++, RPS_CLRSIGNAL | RPS_DEBI );			// Reset "shadow RAM uploaded" flag.
    *pRPS++ = ( RPS_CLRSIGNAL | RPS_DEBI );
		//WriteMem( pRPS++, RPS_UPLOAD    | RPS_DEBI );			// Invoke shadow RAM upload.
    *pRPS++ = ( RPS_UPLOAD | RPS_DEBI );			
		//WriteMem( pRPS++, RPS_PAUSE     | RPS_DEBI );			// Wait for shadow upload to finish.
    *pRPS++ = ( RPS_PAUSE | RPS_DEBI );

// warning C4244: 'initializing' : 
// conversion from 'unsigned __int64' to 'unsigned long', possible loss of data
#pragma warning( disable : 4244 )

		// Delay at least 10 microseconds for analog input settling.  Instead of padding with NOPs,
		// we use RPS_JUMP instructions here; this allows us to produce a longer delay than is
		// possible with NOPs because each RPS_JUMP flushes the RPS' instruction prefetch pipeline.
		//unsigned long JmpAdrs = (unsigned long long)MyRPSBuf.PhysicalBase + (unsigned long long)pRPS - 
    //  (unsigned long long)MyRPSBuf.LogicalBase;
    unsigned long JmpAdrs = (unsigned long long)MyRPSBuf.PhysicalBase + 
      ((pRPS - dmaSetupBuffer.begin()) * sizeof( unsigned long ));
   
#pragma warning( default: 4244 )

		for ( unsigned short i = 0; i < ( 10 * RPSCLK_PER_US / 2); i++ )
		{
			JmpAdrs += 8;										// Repeat to implement time delay:
			//WriteMem( pRPS++, RPS_JUMP );						//   Jump to next RPS instruction.
      *pRPS++ = ( RPS_JUMP );
			//WriteMem( pRPS++, JmpAdrs );
      *pRPS++ = ( JmpAdrs );
		}
		
		// Start ADC by pulsing GPIO1.
		//WriteMem( pRPS++, RPS_LDREG | (P_GPIO >> 2) );			// Begin ADC Start pulse.
    *pRPS++ = ( RPS_LDREG | (P_GPIO >> 2) );			
		//WriteMem( pRPS++, GPIO_BASE | GPIO1_LO );
    *pRPS++ = ( GPIO_BASE | GPIO1_LO );
		//WriteMem( pRPS++, RPS_NOP );							// VERSION 2.03 CHANGE: STRETCH OUT ADC START PULSE.
    *pRPS++ = ( RPS_NOP );
		//WriteMem( pRPS++, RPS_LDREG | (P_GPIO >> 2) );			// End ADC Start pulse.
    *pRPS++ = ( RPS_LDREG | (P_GPIO >> 2) );			
		//WriteMem( pRPS++, GPIO_BASE | GPIO1_HI );
    *pRPS++ = ( GPIO_BASE | GPIO1_HI );

		// Wait for ADC to complete (GPIO2 is asserted high when ADC not busy) and for data
		// from previous conversion to shift into FB BUFFER 1 register.
		//WriteMem( pRPS++, RPS_PAUSE | RPS_GPIO2 );				// Wait for ADC done.
    *pRPS++ = ( RPS_PAUSE | RPS_GPIO2 );				

		// Transfer ADC data from FB BUFFER 1 register to MyADCdata[] DMA buffer.
		//WriteMem( pRPS++, RPS_STREG | ( BUGFIX_STREG( P_FB_BUFFER1 ) >> 2 ) );
    *pRPS++ = ( RPS_STREG | ( BUGFIX_STREG( P_FB_BUFFER1 ) >> 2 ) );
		//WriteMem( pRPS++, (unsigned long long)MyANABuf.PhysicalBase + ( MyAdcItems << 2 ) );
    *pRPS++ = ( anaPhyBaseAddr + ( MyAdcItems << 2 ) );

		// If this slot's EndOfPollList flag is set, all channels have now been processed.
		if ( *ppl++ & EOPL )
		{
			MyAdcItems++;		// Adjust poll list item count.
			break;				// Exit poll list processing loop.
		}
	}

	// VERSION 2.01 CHANGE: DELAY CHANGED FROM 250NS to 2US.
	// Allow the ADC to stabilize for 2 microseconds before starting the final (dummy) conversion.
	// This delay is necessary to allow sufficient time between last conversion finished and the start
	// of the dummy conversion.  Without this delay, the last conversion's data value is sometimes set
	// to the previous conversion's data value.
	for ( unsigned short n = 0; n < ( 2 * RPSCLK_PER_US ); n++ )
		//WriteMem( pRPS++, RPS_NOP );
    *pRPS++ = ( RPS_NOP );

	// Start a dummy conversion to cause the data from the last conversion of interest to be shifted in.
	//WriteMem( pRPS++, RPS_LDREG | (P_GPIO >> 2) );				// Begin ADC Start pulse.
  *pRPS++ = ( RPS_LDREG | (P_GPIO >> 2) );				
	//WriteMem( pRPS++, GPIO_BASE | GPIO1_LO );
  *pRPS++ = ( GPIO_BASE | GPIO1_LO );
	//WriteMem( pRPS++, RPS_NOP );								// VERSION 2.03 CHANGE: STRETCH OUT ADC START PULSE.
  *pRPS++ = ( RPS_NOP );								
	//WriteMem( pRPS++, RPS_LDREG | (P_GPIO >> 2) );				// End ADC Start pulse.
  *pRPS++ = ( RPS_LDREG | (P_GPIO >> 2) );				
	//WriteMem( pRPS++, GPIO_BASE | GPIO1_HI );
  *pRPS++ = ( GPIO_BASE | GPIO1_HI );

	// Wait for the data from the last conversion of interest to arrive in FB BUFFER 1 register.
	//WriteMem( pRPS++, RPS_PAUSE | RPS_GPIO2 );					// Wait for ADC done.
  *pRPS++ = ( RPS_PAUSE | RPS_GPIO2 );					
		
	// Transfer final ADC data from FB BUFFER 1 register to MyADCdata[] DMA buffer.
	//WriteMem( pRPS++, RPS_STREG | ( BUGFIX_STREG( P_FB_BUFFER1 ) >> 2 ) );
  *pRPS++ = ( RPS_STREG | ( BUGFIX_STREG( P_FB_BUFFER1 ) >> 2 ) );
	//WriteMem( pRPS++, (unsigned long long)MyANABuf.PhysicalBase + ( MyAdcItems << 2 ) );
  *pRPS++ = ( anaPhyBaseAddr + ( MyAdcItems << 2 ) );

	// Indicate ADC scan loop is finished.
	//WriteMem( pRPS++, RPS_CLRSIGNAL | RPS_SIGADC );					// Signal ReadADC() that scan is done.
  *pRPS++ = ( RPS_CLRSIGNAL | RPS_SIGADC );					

	// Restart RPS program at its beginning.
	//WriteMem( pRPS++, RPS_JUMP );									// Branch to start of RPS program.
  *pRPS++ = ( RPS_JUMP );									
	//WriteMem( pRPS++, (unsigned long long)MyRPSBuf.PhysicalBase );
  *pRPS++ = ( rpsPhyBaseAddr );

	// End of RPS program build  ------------------------------------------------------------

  // write local buffer to DMA physical memory
  WriteDmaMem( MyRPSBuf, (PUCHAR)(&dmaSetupBuffer[0]), 
    (ULONG)((pRPS - dmaSetupBuffer.begin())*sizeof(unsigned long)) );

	// Start executing the RPS program.
	MC_ENABLE( P_MC1, MC1_ERPS1 );

	// End critical segment.
	CriticalEnd();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Execute one ADC.  AdcSpec has form (x,x,x,RANGE,CHAN<3:0>), where RANGE code indicates 0 = +-10V, 1 = +-5V.

unsigned short Cls626::SingleADC( unsigned short AdcSpec )
{
	// Convert application's ADC specification into form appropriate for register programming.
	AdcSpec = ( AdcSpec << 8 ) | ( AdcSpec & 0x10 ? GSEL_BIPOLAR5V : GSEL_BIPOLAR10V );

	// Start critical segment.
	CriticalBegin();

	// Switch ADC analog gain.
	DEBIwrite( LP_GSEL, (unsigned short)AdcSpec );		// Set gain.

	// Select ADC analog input channel.
	DEBIwrite( LP_ISEL, (unsigned short)AdcSpec );		// Select channel.

	// Delay 10 microseconds for analog input settling.
	for ( int i = 0; i < ( 10 * 33 ); i++ )
		RR7146( P_PSR );

	// Start ADC by pulsing GPIO1 low.
	unsigned long GpioImage = RR7146( P_GPIO );
	WR7146( P_GPIO, GpioImage & ~GPIO1_HI );			// Assert ADC Start command
	WR7146( P_GPIO, GpioImage & ~GPIO1_HI );			//   and stretch it out.
	WR7146( P_GPIO, GpioImage & ~GPIO1_HI );
	WR7146( P_GPIO, GpioImage | GPIO1_HI );				// Negate ADC Start command.

	// Wait for ADC to complete (GPIO2 is asserted high when ADC not busy) and for data
	// from previous conversion to shift into FB BUFFER 1 register.
	while ( !( RR7146( P_PSR ) & PSR_GPIO2 ) );			// Wait for ADC done.

	// Allow the ADC to stabilize for 4 microseconds before starting the final (dummy) conversion.
	// This delay is necessary to allow sufficient time between last conversion finished and the start
	// of the dummy conversion.  Without this delay, the last conversion's data value is sometimes set
	// to the previous conversion's data value.
	for ( i = 0; i < ( 4 * 33 ); i++ )
		RR7146( P_PSR );

	// Start a dummy conversion to cause the data from the previous conversion to be shifted in.
	GpioImage = RR7146( P_GPIO );
	WR7146( P_GPIO, GpioImage & ~GPIO1_HI );			// Assert ADC Start command
	WR7146( P_GPIO, GpioImage & ~GPIO1_HI );			//   and stretch it out.
	WR7146( P_GPIO, GpioImage & ~GPIO1_HI );
	WR7146( P_GPIO, GpioImage | GPIO1_HI );				// Negate ADC Start command.

	// Wait for the data to arrive in FB BUFFER 1 register.
	while ( !( RR7146( P_PSR ) & PSR_GPIO2 ) );			// Wait for ADC done.

	// Fetch ADC data from audio interface's input shift register.
	unsigned long AdcData = RR7146( P_FB_BUFFER1 );		// Fetch ADC data.

	// End critical segment.
	CriticalEnd();

	// Return the digitized value.
	return (unsigned short)( AdcData >> 16 );
}

//////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////  DAC FUNCTIONS  ////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

#define MAX_TRIMDAC		11						// Maximum valid TrimDAC channel number.
#define MAX_DAC			3						// Maximum valid application DAC channel number.

// Slot 0 base settings.
#define VECT0			( XSD2 | RSD3 | SIB_A2 )		// Slot 0 always shifts in 0xFF and stores it to FB_BUFFER2.

// TrimDac LogicalChan-to-PhysicalChan mapping table.
static unsigned char trimchan[] = { 10, 9, 8, 3, 2, 7, 6, 1, 0, 5, 4 };

// TrimDac LogicalChan-to-EepromAdrs mapping table.
static unsigned char trimadrs[] = { 0x40, 0x41, 0x42, 0x50, 0x51, 0x52, 0x53, 0x60, 0x61, 0x62, 0x63 };

////////////////////////////////////////////////////////////////////
// Initialize DAC interface.

void Cls626::InitDAC()
{
	// Init Audio2's output DMAC attributes: burst length = 1 DWORD, threshold = 1 DWORD.
	WR7146( P_PCI_BT_A, 0 );

	// Init Audio2's output DMA physical addresses.  The protection address is set to 1 DWORD past the base address
	// so that a single DWORD will be transferred each time a DMA transfer is enabled.
	register unsigned long PhysBuf = (unsigned long)MyANABuf.PhysicalBase + 
    (DAC_WDMABUF_OS * sizeof(unsigned long));
	WR7146( P_BASEA2_OUT, PhysBuf );				// Buffer base adrs.
	WR7146( P_PROTA2_OUT, (PhysBuf + (1 * sizeof(unsigned long))) );		// Protection address.

	// Cache Audio2's output DMA buffer logical address.  This is where DAC data is buffered for A2 output DMA transfers.
	//pMyDacWBuf = (unsigned long *)MyANABuf.LogicalBase + DAC_WDMABUF_OS;

	// Audio2's output channels does not use paging.  The protection violation handling bit is set so that the
	// DMAC will automatically halt and its PCI address pointer will be reset when the protection address is reached.
	WR7146( P_PAGEA2_OUT, 8 );

	// Initialize time slot list 2 (TSL2), which is used to control the clock generation for and serialization of data
	// to be sent to the DAC devices.  Slot 0 is a NOP that is used to trap TSL execution; this permits other slots
	// to be safely modified without first turning off the TSL sequencer (which is apparently impossible to do).  Also,
	// SD3 (which is driven by a pull-up resistor) is shifted in and stored to the MSB of FB_BUFFER2 to be used as
	// evidence that the slot sequence has not yet finished executing.
	SETVECT( 0, XSD2 | RSD3 | SIB_A2 | EOS );				// Slot 0: Trap TSL execution, shift 0xFF into FB_BUFFER2.

	// Initialize slot 1, which is constant.  Slot 1 causes a DWORD to be transferred from audio channel 2's output
	// FIFO to the FIFO's output buffer so that it can be serialized and sent to the DAC during subsequent slots.  All
	// remaining slots are dynamically populated as required by the target DAC device.
	SETVECT( 1, LF_A2 );	// Slot 1: Fetch DWORD from Audio2's output FIFO.

	// Start DAC's audio interface (TSL2) running.
	WR7146( P_ACON1, ACON1_DACSTART );
}

////////////////////////////////////////////////////////////////////
// Read all TrimDac setpoints from EEPROM and write to TrimDacs.
// Logical TrimDac channels are ordered as follows:
// ADCZero, ADCGain10V, ADCGain5V, DACZero[4], DACGain[4].

void Cls626::LoadTrimDACs()
{
	// Copy TrimDac setpoint values from EEPROM to TrimDacs.
	for ( register unsigned short i = 0; i < sizeof(trimchan); i++ )
		WriteTrimDAC( (unsigned char)i, I2Cread( trimadrs[i] ) );
}

//////////////////////////////////////////////////////////////////////////////
// Write/return setpoint to/from a TrimDAC channel.

void Cls626::WriteTrimDAC( unsigned char LogicalChan, unsigned char DacData )
{
	// Abort if illegal TrimDac channel number specified.
	ValidateParm( LogicalChan, MAX_TRIMDAC );

	// Start critical segment.
	CriticalBegin();

	// Save the new setpoint in case the application needs to read it back later.
	MyTrimSetpoint[LogicalChan] = (unsigned char)DacData;

	// Map logical channel number to physical channel number.
	unsigned long chan = (unsigned long)trimchan[LogicalChan];

	// Set up TSL2 records for TrimDac write operation.  All slots shift 0xFF in from pulled-up SD3 so that the end
	// of the slot sequence can be detected.
	SETVECT( 2, XSD2 | XFIFO_1 | WS3 );			// Slot 2: Send high unsigned char to target TrimDac.
	SETVECT( 3, XSD2 | XFIFO_0 | WS3 );			// Slot 3: Send low unsigned char to target TrimDac.
	SETVECT( 4, XSD2 | XFIFO_3 | WS1 );			// Slot 4: Send NOP high unsigned char to DAC0 to keep clock running.
	SETVECT( 5, XSD2 | XFIFO_2 | WS1 | EOS );	// Slot 5: Send NOP low unsigned char to DAC0.

	// Construct and transmit target DAC's serial packet: ( 0000 AAAA ),( DDDD DDDD ),( 0x00 ),( 0x00 ) where A<3:0> is the
	// DAC channel's address, and D<7:0> is the DAC setpoint.  Append a WORD value (that writes a channel 0 NOP command to a
	// non-existent main DAC channel) that serves to keep the clock running after the packet has been sent to the target DAC.
	SendDAC( ( (unsigned long)chan << 8 )		// Address the DAC channel within the trimdac device.
		| (unsigned long)DacData );				// Include DAC setpoint data.

	// End critical segment.
	CriticalEnd();
}

unsigned char Cls626::ReadTrimDAC( unsigned char LogicalChan )
{
	// Abort if illegal TrimDac channel number specified.
	ValidateParmRtnval( LogicalChan, MAX_TRIMDAC );

	// Return the setpoint image.
	return MyTrimSetpoint[LogicalChan];
}

//////////////////////////////////////////////////////////////////////////////////////////////////
// Write setpoint to a DAC channel.  If polarity of new setpoint differs from that of the
// current setpoint, two DAC writes are performed: the first write switches the DAC output to
// 0V and the second write switches the output to the desired setpoint.  This prevents a glitch
// from occuring on the DAC output, which could be caused by the polarity change occuring before
// the new DAC setpoint value is programmed.

void Cls626::WriteDAC( unsigned short chan, short dacdata )
{
	// Abort if illegal DAC channel number specified.
	ValidateParm( chan, MAX_DAC );

	// Start critical segment.
	CriticalBegin();

	// If new setpoint has polarity that differs from current setpoint, write an intermediate
	// output setpoint near 0V to prevent the DAC output from glitching.
	if ( ( ( dacdata ^ MyDacSetpoint[chan] ) & 0x8000 ) != 0 )
		SetDAC( chan, ( dacdata < 0 ) ? 0 : -1 );

	// Write new DAC setpoint.
	SetDAC( chan, dacdata );

	// End critical segment.
	CriticalEnd();
}

/////////////////////////////////////////////////////////////////////////////////
// Return DAC channel's cached setpoint value.

short Cls626::ReadDAC( unsigned short chan )
{
	return MyDacSetpoint[chan];
}

/////////////////////////////////////////////////////////////////////////////////
// Private helper function: Write setpoint to an application DAC channel.

void Cls626::SetDAC( unsigned short chan, short dacdata )
{
	// Cache new setpoint.
	MyDacSetpoint[chan] = dacdata;

	// Adjust DAC data polarity and set up Polarity Control Register image.
	register unsigned short signmask = 1 << chan;
	if ( dacdata < 0 )
	{
		dacdata = -dacdata;
		MyDacpol |= signmask;
	}
	else
		MyDacpol &= ~signmask;

	// Limit DAC setpoint value to valid range.
	if ( (unsigned short)dacdata > 0x1FFF )
		dacdata = 0x1FFF;

	// Set up TSL2 records (aka "vectors") for DAC update.  Vectors V2 and V3 transmit the setpoint to the target DAC.
	// V4 and V5 send data to a non-existent TrimDac channel just to keep the clock running after sending data to the target
	// DAC.  This is necessary to eliminate the clock glitch that would otherwise occur at the end of the target DAC's serial
	// data stream.  When the sequence restarts at V0 (after executing V5), the gate array automatically disables gating for
	// the DAC clock and all DAC chip selects.
	register unsigned long WSImage = ( chan & 2 ) ? WS1 : WS2;	// Choose DAC chip select to be asserted.
	SETVECT( 2, XSD2 | XFIFO_1 | WSImage );						// Slot 2: Transmit high data byte to target DAC.
	SETVECT( 3, XSD2 | XFIFO_0 | WSImage );						// Slot 3: Transmit low data byte to target DAC.
	SETVECT( 4, XSD2 | XFIFO_3 | WS3 );							// Slot 4: Transmit to non-existent TrimDac channel to keep clock
	SETVECT( 5, XSD2 | XFIFO_2 | WS3 | EOS );					// Slot 5:   running after writing target DAC's low data byte.

	// Construct and transmit target DAC's serial packet: ( A10D DDDD ),( DDDD DDDD ),( 0x0F ),( 0x00 ) where A is chan<0>, and D<12:0>
	// is the DAC setpoint.  Append a WORD value (that writes to a non-existent TrimDac channel) that serves to keep the clock running
	// after the packet has been sent to the target DAC.
	SendDAC( 0x0F000000									// Continue clock after target DAC data (write to non-existent trimdac).
		| 0x00004000									// Address the two main dual-DAC devices (TSL's chip select enables target device).
		| ( (unsigned long)( chan & 1 ) << 15 )			// Address the DAC channel within the device.
		| (unsigned long)dacdata );						// Include DAC setpoint data.
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// Private helper function: Transmit serial data to DAC via Audio channel 2.
// Assumes: (1) TSL2 slot records initialized, and (2) MyDacpol contains valid target image.

void Cls626::SendDAC( unsigned long val )
{
	// START THE SERIAL CLOCK RUNNING -------------

	// Assert DAC polarity control and enable gating of DAC serial clock and audio bit stream signals.  At
	// this point in time we must be assured of being in time slot 0.  If we are not in slot 0, the serial
	// clock and audio stream signals will be disabled; this is because the following DEBIwrite statement
	// (which enables signals to be passed through the gate array) would execute before the trailing edge of
	// WS1/WS3 (which turns off the signals), thus causing the signals to be inactive during the DAC write.
	DEBIwrite( LP_DACPOL, MyDacpol );

	// TRANSFER OUTPUT DWORD VALUE INTO A2'S OUTPUT FIFO ----------------

	// Copy DAC setpoint value to DAC's output DMA buffer.
	//*pMyDacWBuf = val;
  WriteDmaMem( MyANABuf, (PUCHAR)&val, sizeof(val), 0, sizeof(unsigned long) * DAC_WDMABUF_OS );

	// Enable the output DMA transfer.  This will cause the DMAC to copy the DAC's data value to A2's output
	// FIFO.  The DMA transfer will then immediately terminate because the protection address is reached
	// upon transfer of the first DWORD value.
	MC_ENABLE( P_MC1, MC1_A2OUT );

	// While the DMA transfer is executing ...

	// Reset Audio2 output FIFO's underflow flag (along with any other FIFO underflow/overflow flags).  When set,
	// this flag will indicate that we have emerged from slot 0.
	WR7146( P_ISR, ISR_AFOU );

	// Wait for the DMA transfer to finish so that there will be data available in the FIFO when time slot 1
	// tries to transfer a DWORD from the FIFO to the output buffer register.  We test for DMA Done by polling the
	// DMAC enable flag; this flag is automatically cleared when the transfer has finished.
	while ( ( RR7146( P_MC1 ) & MC1_A2OUT ) != 0 );

	// START THE OUTPUT STREAM TO THE TARGET DAC --------------------

	// FIFO data is now available, so we enable execution of time slots 1 and higher by clearing the
	// EOS flag in slot 0.  Note that SD3 will be shifted in and stored in FB_BUFFER2 for end-of-slot-list detection.
	SETVECT( 0, XSD2 | RSD3 | SIB_A2 );

	// Wait for slot 1 to execute to ensure that the Packet will be transmitted.  This is detected
	// by polling the Audio2 output FIFO underflow flag, which will be set when slot 1 execution has
	// finished transferring the DAC's data DWORD from the output FIFO to the output buffer register.
	while ( ( RR7146( P_SSR ) & SSR_AF2_OUT ) == 0 );

	// Set up to trap execution at slot 0 when the TSL sequencer cycles back to slot 0 after executing the
	// EOS in slot 5.  Also, simultaneously shift out and in the 0x00 that is ALWAYS the value stored in the
	// last byte to be shifted out of the FIFO's DWORD buffer register.
	SETVECT( 0, XSD2 | XFIFO_2 | RSD2 | SIB_A2 | EOS );

	// WAIT FOR THE TRANSACTION TO FINISH -----------------------

	// Wait for the TSL to finish executing all time slots before exiting this function.  We must do this so that
	// the next DAC write doesn't start, thereby enabling clock/chip select signals:
	//   1.	Before the TSL sequence cycles back to slot 0, which disables the clock/cs signal gating and traps slot
	//		list execution.  If we have not yet finished slot 5 then the clock/cs signals are still gated and we have
	//		not finished transmitting the stream.
	//   2.	While slots 2-5 are executing due to a late slot 0 trap.  In this case, the slot sequence is currently
	//		repeating, but with clock/cs signals disabled.  We must wait for slot 0 to trap execution before setting
	//		up the next DAC setpoint DMA transfer and enabling the clock/cs signals.
	// To detect the end of slot 5, we test for the FB_BUFFER2 MSB contents to be equal to 0xFF.  If the TSL has
	// not yet finished executing slot 5 ...
	if ( ( RR7146( P_FB_BUFFER2 ) & 0xFF000000 ) != 0 )
	{
		// The trap was set on time and we are still executing somewhere in slots 2-5, so we now wait for slot 0 to execute
		// and trap TSL execution.  This is detected when FB_BUFFER2 MSB changes from 0xFF to 0x00, which slot 0 causes to
		// happen by shifting out/in on SD2 the 0x00 that is always referenced by slot 5.
		while ( ( RR7146( P_FB_BUFFER2 ) & 0xFF000000 ) != 0 );
	}

	// Either (1) we were too late setting the slot 0 trap; the TSL sequencer restarted slot 0 before we could set the EOS
	// trap flag, or (2) we were not late and execution is now trapped at slot 0.  In either case, we must now change slot 0
	// so that it will store value 0xFF (instead of 0x00) to FB_BUFFER2 next time it executes.  In order to do this, we
	// reprogram slot 0 so that it will shift in SD3, which is driven only by a pull-up resistor.
	SETVECT( 0, RSD3 | SIB_A2 | EOS );

	// Wait for slot 0 to execute, at which time the TSL is setup for the next DAC write.  This is detected when FB_BUFFER2
	// MSB changes from 0x00 to 0xFF.
	while ( ( RR7146( P_FB_BUFFER2 ) & 0xFF000000 ) == 0 );
}

////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////  PRIVATE HIGH-LEVEL FUNCTION HELPERS  ////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////
// Define new image for MISC2 register and write it to MISC2 reg.

void Cls626::WriteMISC2( unsigned short NewImage )
{
	DEBIwrite( LP_MISC1, MISC1_WENABLE );		// Enable writes to MISC2 register.
	DEBIwrite( LP_WRMISC2, NewImage );			// Write new image to MISC2.
	DEBIwrite( LP_MISC1, MISC1_WDISABLE );		// Disable writes to MISC2.
}

/////////////////////////////////////////////////////////////////////////////////////////////
// Return image of what was last written to the MISC2 register.

#define BIT14	0x4000
#define BIT13	0x2000

#define RBMASK	( MISC2_BATT_ENABLE | MISC2_WDENABLE | MISC2_WDPERIOD_MASK )

unsigned short Cls626::WRMISC2image()
{
	// Start critical segment.
	CriticalBegin();

	// Fetch DIO 0-5 mode flags.  These must be read from the counter B control registers as
	// they are write-only bits in the MISC2 register and therefore have no direct readback from MISC2.
	register unsigned short c3 = DEBIread( LP_CR2B );		// DIO 0-1 modes.
	register unsigned short c2 = DEBIread( LP_CR1B );		// DIO 2-3 modes.
	register unsigned short c1 = DEBIread( LP_CR0B );		// DIO 4-5 modes.

	// Fetch BackupEnable, WatchdogEnable and WatchdogInterval flags, and combine with
	// ChargeEnable (image only, no readback register is available).
	unsigned short retval = ( DEBIread( LP_RDMISC2 ) & RBMASK ) | MyChargeEnabled;

	// End critical segment.
	CriticalEnd();

	// Assemble all the bits to produce an image of the current state of the WRMISC2 register.
	return retval
		| ( ( c3 & BIT13 ) >>  4 )		// DIO 5 mode.
		| ( ( c3 & BIT14 ) >>  6 )		// DIO 4 mode.
		| ( ( c2 & BIT13 ) >>  6 )		// DIO 3 mode.
		| ( ( c2 & BIT14 ) >>  8 )		// DIO 2 mode.
		| ( ( c1 & BIT13 ) >>  8 )		// DIO 1 mode.
		| ( ( c1 & BIT14 ) >> 10 );		// DIO 0 mode.
}

////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////  BATTERY CHARGING AND BACKUP FUNCTIONS  //////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////
// Return non-zero if battery backup is enabled.

unsigned short Cls626::BackupEnable()
{
	return DEBIread( LP_RDMISC2 ) & MISC2_BATT_ENABLE;
}

////////////////////////////////////////////////////////////////////
// Enable/disable battery backup.  Enabled if Enable is non-zero.

void Cls626::BackupEnable( unsigned short Enable )
{
	// Start critical segment.
	CriticalBegin();

	// Construct new image of MISC2 register and write it to MISC2.
	WriteMISC2( Enable ? ( WRMISC2image() | MISC2_BATT_ENABLE ) : ( WRMISC2image() & ~MISC2_BATT_ENABLE ) );

	// End critical segment.
	CriticalEnd();
}

/////////////////////////////////////////////////////
// Return non-zero if battery charging is enabled.

unsigned short Cls626::ChargeEnable()
{
	return MyChargeEnabled;
}

//////////////////////////////////////////////////////////////////////
// Enable/disable battery charging.  Enabled if Enable is non-zero.

void Cls626::ChargeEnable( unsigned short Enable )
{
	// Init image of WRMISC2 Battery Charge Enabled control bit.  This image is used when the
	// state of the control bit is queried as it has no direct hardware readback mechanism.
	MyChargeEnabled = Enable ? WRMISC2_CHARGE_ENABLE : 0;

	// Start critical segment.
	CriticalBegin();

	// Program the enable output for the battery charger.
	WriteMISC2( ( WRMISC2image() & ~WRMISC2_CHARGE_ENABLE ) | MyChargeEnabled );

	// End critical segment.
	CriticalEnd();
}

////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  WATCHDOG FUNCTIONS  /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////
// Return non-zero if the watchdog has timed out.

unsigned short Cls626::WatchdogTimeout()
{
	return DEBIread( LP_MISC1 ) & RDMISC1_WDTIMEOUT;
}

//////////////////////////////////////////////////////
// Return non-zero if the watchdog is enabled.

unsigned short Cls626::WatchdogEnable()
{
	return DEBIread( LP_RDMISC2 ) & MISC2_WDENABLE;
}

/////////////////////////////////////////////////////////////////////////
// Enable/disable the watchdog timer.  Enabled if Enable is non-zero.

void Cls626::WatchdogEnable( unsigned short Enable )
{
	// Start critical segment.
	CriticalBegin();

	// Get image of value last written to MISC2.
	register unsigned short image = WRMISC2image();

	if ( Enable )
	{
		// Construct new image to be written to MISC2.
		image = MISC2_WDENABLE					// Enable the WD timer.
			| ( image & ~MISC2_WDPERIOD_MASK )	// Get image of value last written to MISC2.
			| MyWDInterval;						// Combine with image of WD interval control.

		// The MISC2 register must be written twice because the WD interval control bits are forced
		// to zero any time the WD enable is inactive.  The first write sets WD enable active and
		// releases the forced reset on the WD interval control bits.  The second write programs the
		// WD interval bits since they are no longer forced to zero.
		DEBIwrite( LP_MISC1, MISC1_WENABLE );	// Enable write access to WD enable/period control in MISC2.
		DEBIwrite( LP_WRMISC2, image );			// Enable WD timer and enable WD period control programming.
		DEBIwrite( LP_WRMISC2, image );			// Program WD period.
		DEBIwrite( LP_MISC1, MISC1_WDISABLE );	// Disable write access to WD enable/period control.
	}
	else
	{
		// Disable the WD timer.  This also causes the gate array to automatically reset the programmed
		// WD interval control bits to zero (and to hold them there until the WD timer is again enabled).
		WriteMISC2( image & ~MISC2_WDENABLE );
	}

	// End critical segment.
	CriticalEnd();
}

/////////////////////////////////////////////////////////////////////
// Return/set the programmed watchdog timer interval.
// Legal timer interval values: 0, 1, 2 or 3.

unsigned short Cls626::WatchdogPeriod()	// Return WD interval.
{
	// Fetch WD enable and interval control bits.
	register unsigned short m2 = DEBIread( LP_RDMISC2 );

	// If WD is enabled, simply return the programmed interval control bits.  If disabled, we
	// instead return an image of the desired interval control bits since the physical control bits
	// are forced to zero by the gate array whenever the WD timer is disabled.
	return ( m2 & MISC2_WDENABLE ) ? ( m2 & MISC2_WDPERIOD_MASK ) : MyWDInterval;
}

void Cls626::WatchdogPeriod( unsigned short Value )	// Set WD interval.
{
	if ( Value > MISC2_WDPERIOD_MASK )
		DeclareError( ERR_ILLEGAL_PARM );
	else
	{
		// Start critical segment.
		CriticalBegin();

		// Cache the new WD interval.
		MyWDInterval = Value;

		// Get the value last written to the MISC2 register.
		register unsigned short image = WRMISC2image();

		// If the WD timer is enabled, program the new interval value.  We skip this if the WD timer
		// is disabled because the gate array will be forcing the WD interval control bits to zero;
		// this doesn't matter, because we have a copy of the interval control bits in MyWDInterval.
		if ( image & MISC2_WDENABLE )
			WriteMISC2( ( image & ~MISC2_WDPERIOD_MASK ) | Value );		// Program new interval.
	}

	// End critical segment.
	CriticalEnd();
}

///////////////////////////////////////////////////////////////////
// Reset the watchdog timer.

void Cls626::WatchdogReset()
{
	DEBIwrite( LP_WRMISC2, WRMISC2_WDCLEAR );		// Reset watchdog timer.
}

/////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  DIGITAL I/O FUNCTIONS  /////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
// All DIO functions address a group of DIO channels by means of "Group" argument.
// Group may be 0, 1 or 2, which correspond to DIO ports A, B and C, respectively.
/////////////////////////////////////////////////////////////////////////////////////////

#define MAX_DIOBANK			2		// Maximum valid DIO group number.
#define MAX_DIOCHAN			47		// Maximum valid DIO channel number.
#define MAX_DIOEXTCHAN		39		// Maximum valid extended-capability DIO channel number.

// These arrays map a DIO group number into the various port addresses needed to access DIO registers.
static const unsigned short RDDIn[]		= { LP_RDDINA,    LP_RDDINB,    LP_RDDINC    };	// Read  physical input states.
static const unsigned short WRDOut[]	= { LP_WRDOUTA,   LP_WRDOUTB,   LP_WRDOUTC   };	// Write physical output states.
static const unsigned short RDEdgSel[]	= { LP_RDEDGSELA, LP_RDEDGSELB, LP_RDEDGSELC };	// Read  edge polarity.
static const unsigned short WREdgSel[]	= { LP_WREDGSELA, LP_WREDGSELB, LP_WREDGSELC };	// Write edge polarity.
static const unsigned short RDCapSel[]	= { LP_RDCAPSELA, LP_RDCAPSELB, LP_RDCAPSELC };	// Read  event capture enables.
static const unsigned short WRCapSel[]	= { LP_WRCAPSELA, LP_WRCAPSELB, LP_WRCAPSELC };	// Write event capture enables.
static const unsigned short RDCapFlg[]	= { LP_RDCAPFLGA, LP_RDCAPFLGB, LP_RDCAPFLGC };	// Read  event capture flags.
static const unsigned short RDIntSel[]	= { LP_RDINTSELA, LP_RDINTSELB, LP_RDINTSELC };	// Read  interrupt enables.
static const unsigned short WRIntSel[]	= { LP_WRINTSELA, LP_WRINTSELB, LP_WRINTSELC };	// Write interrupt enables.

/////////////////////////////////////////////////////////////////////////////////////////
// Initialize all DIO hardware to be consistent with states following a PCI hard reset.

void Cls626::DIOInit()
{
	// Prepare to treat writes to WRCapSel as capture disables.
	DEBIwrite( LP_MISC1, MISC1_NOEDCAP );

	// For each group of sixteen channels ...
	for ( register unsigned short Group = 0; Group < 3; Group++ )
	{
		DEBIwrite( WRIntSel[Group], 0 );						// Disable all interrupts.
		DEBIwrite( WRCapSel[Group], 0xFFFF );				// Disable all event captures.
		DEBIwrite( WREdgSel[Group], 0 );						// Init all DIOs to default edge polarity.
		DEBIwrite( WRDOut[Group], MyDOutImage[Group] = 0 );	// Program all outputs to inactive state.
	}
}

/////////////////////////////////////////////////////////////////////////////////////////
// Helper function: insert a bit value into a word.

unsigned short Cls626::DIOBitInsert( unsigned short WordVal, unsigned short Chan, unsigned short BitVal )
{
	// Cache the channel's bit mask.
	register unsigned short Mask = 1 << ( Chan & 15 );

	// Return modified word value.
	return BitVal ? ( WordVal | Mask ) : ( WordVal & ~Mask );
}

////////////////////////////////////////////////////////////////////////////
// Return the input states of a group of DIO channels.

unsigned short Cls626::DIOGroupRead( unsigned short Group )
{
	// Abort if illegal DIO group number specified.
	ValidateParmRtnval( Group, MAX_DIOBANK );

	// Return DIO input states.
	return DEBIread( RDDIn[Group] );
}

////////////////////////////////////////////////////////////////////////////
// Return the input state of a DIO channel.

unsigned short Cls626::DIOChanRead( unsigned short Chan )
{
	// Abort if illegal DIO chan number specified.
	ValidateParmRtnval( Chan, MAX_DIOCHAN );

	// Return DIO input state.
	return ( DEBIread( RDDIn[Chan >> 4] ) >> ( Chan & 15 ) ) & 1;
}

////////////////////////////////////////////////////////////////////////////
// Return/set physical output register states of a DIO channel group.

unsigned short Cls626::DIOGroupWrite( unsigned short Group )
{
	// Abort if illegal DIO group number specified.
	ValidateParmRtnval( Group, MAX_DIOBANK );

	// Return DIO physical output states.
	return MyDOutImage[Group];
}

void Cls626::DIOGroupWrite( unsigned short Group, unsigned short Value )
{
	// Abort if illegal DIO group number specified.
	ValidateParm( Group, MAX_DIOBANK );

	// Start critical segment.
	CriticalBegin();

	// Modify DIO group output image and write to physical outputs.
	DEBIwrite( WRDOut[Group], MyDOutImage[Group] = Value );

	// End critical segment.
	CriticalEnd();
}

////////////////////////////////////////////////////////////////////////////
// Return/set physical output register state of a DIO channel.

unsigned short Cls626::DIOChanWrite( unsigned short Chan )
{
	// Abort if illegal DIO chan number specified.
	ValidateParmRtnval( Chan, MAX_DIOCHAN );

	// Return DIO physical output state.
	return ( MyDOutImage[Chan >> 4] >> ( Chan & 15 ) ) & 1;
}

void Cls626::DIOChanWrite( unsigned short Chan, unsigned short Value )
{
	// Abort if illegal DIO chan number specified.
	ValidateParm( Chan, MAX_DIOCHAN );

	// Cache the channel's group number.
	register unsigned short Group = Chan >> 4;

	// Start critical segment.
	CriticalBegin();

	// Construct new DIO group image.
	MyDOutImage[Group] = DIOBitInsert( MyDOutImage[Group], Chan, Value );

	// Modify DIO group output image and write to physical outputs.
	DEBIwrite( WRDOut[Group], MyDOutImage[Group] );

	// End critical segment.
	CriticalEnd();
}

/////////////////////////////////////////////////////////////////////
// Return/set the DIO edge polarity to use for event captures.

unsigned short Cls626::DIOGroupEdge( unsigned short Group )
{
	// Abort if illegal DIO group number specified.
	ValidateParmRtnval( Group, MAX_DIOBANK );

	// Return edge polarities for target DIO group.
	return DEBIread( RDEdgSel[Group] );
}

void Cls626::DIOGroupEdge( unsigned short Group, unsigned short Value )
{
	// Abort if illegal DIO group number specified.
	ValidateParm( Group, MAX_DIOBANK );

	// Set edge polarities for target DIO group.
	DEBIwrite( WREdgSel[Group], Value );
}

unsigned short Cls626::DIOChanEdge( unsigned short Chan )
{
	// Abort if illegal DIO chan number specified.
	ValidateParmRtnval( Chan, MAX_DIOEXTCHAN );

	// Return edge polarity.
	return ( DEBIread( RDEdgSel[Chan >> 4] ) >> ( Chan & 15 ) ) & 1 ;
}

void Cls626::DIOChanEdge( unsigned short Chan, unsigned short Value )
{
	// Abort if illegal DIO chan number specified.
	ValidateParm( Chan, MAX_DIOEXTCHAN );

	// Cache the channel's group number.
	unsigned short Group = Chan >> 4;

	// Start critical segment.
	CriticalBegin();

	// Set edge polarity.
	DEBIwrite( WREdgSel[Group], DIOBitInsert( DEBIread(RDEdgSel[Group]), Chan, Value ) );

	// End critical segment.
	CriticalEnd();
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Enable/disable DIO edge event capture for a group of DIO channels.
//   Value  : Logic "1" in a bit position specifies that the corresponding channel's event
//            capture enable is to be modified.
//   Enable : Specifies whether modified channels are to be enabled (non-zero) or disabled (zero).

void Cls626::DIOGroupCapEnable( unsigned short Group, unsigned short Value, unsigned short Enable )
{
	// Abort if illegal DIO group number specified.
	ValidateParm( Group, MAX_DIOBANK );

	// Start critical segment.
	CriticalBegin();

	// Prepare to treat the following data as capture enable/disable.
	DEBIwrite( LP_MISC1, Enable ? MISC1_EDCAP : MISC1_NOEDCAP );

	// Write the capture enables/disables.
	DEBIwrite( WRCapSel[Group], Value );

	// End critical segment.
	CriticalEnd();
}

//////////////////////////////////////////////////////////////////////////////////
// Return DIO capture enables for a group of DIO channels.  A "1" in a bit
// position indicates edge capture is enabled for the corresponding channel.

unsigned short Cls626::DIOGroupCapEnable( unsigned short Group )
{
	// Abort if illegal DIO group number specified.
	ValidateParmRtnval( Group, MAX_DIOBANK );

	// Return capture enables for target DIO group.
	return DEBIread( RDCapSel[Group] );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Enable/disable DIO edge event capture for DIO channel.

void Cls626::DIOChanCapEnable( unsigned short Chan, unsigned short Enable )
{
	// Abort if illegal DIO group number specified.
	ValidateParm( Chan, MAX_DIOEXTCHAN );

	// Start critical segment.
	CriticalBegin();

	// Prepare to treat the following data as capture enable/disable.
	DEBIwrite( LP_MISC1, Enable ? MISC1_EDCAP : MISC1_NOEDCAP );

	// Write the capture enables/disables.
	DEBIwrite( WRCapSel[Chan >> 4], 1 << ( Chan & 15 ) );

	// End critical segment.
	CriticalEnd();
}

//////////////////////////////////////////////////////////////////////////////////
// Return capture enable for a DIO channel.

unsigned short Cls626::DIOChanCapEnable( unsigned short Chan )
{
	// Abort if illegal DIO chan number specified.
	ValidateParmRtnval( Chan, MAX_DIOEXTCHAN);

	// Return capture enable.
	return DEBIread( RDCapSel[Chan >> 4] );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Clear event captures for a DIO group.  A logic "1" in a DIO's bit position in Value specifies
// that the DIO's event capture is to be reset by first disabling, then re-enabling event captures.

void Cls626::DIOGroupCapReset( unsigned short Group, unsigned short Value )
{
	// Abort if illegal DIO group number specified.
	ValidateParm( Group, MAX_DIOBANK );

	// Start critical segment.
	CriticalBegin();

	// Disable the specified captures.
	DEBIwrite( LP_MISC1, MISC1_NOEDCAP );
	DEBIwrite( WRCapSel[Group], Value );

	// Re-enable the specified captures.
	DEBIwrite( LP_MISC1, MISC1_EDCAP );
	DEBIwrite( WRCapSel[Group], Value );

	// End critical segment.
	CriticalEnd();
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Clear event captures for a DIO channel.

void Cls626::DIOChanCapReset( unsigned short Chan )
{
	// Abort if illegal DIO chan number specified.
	ValidateParm( Chan, MAX_DIOEXTCHAN );

	// Cache the group number and channel mask.
	register unsigned short Group = Chan >> 4;
	register unsigned short Mask = 1 << ( Chan & 15 );

	// Start critical segment.
	CriticalBegin();

	// Disable the capture.
	DEBIwrite( LP_MISC1, MISC1_NOEDCAP );
	DEBIwrite( WRCapSel[Group], Mask );

	// Re-enable the capture.
	DEBIwrite( LP_MISC1, MISC1_EDCAP );
	DEBIwrite( WRCapSel[Group], Mask );

	// End critical segment.
	CriticalEnd();
}

////////////////////////////////////////////////////////////////////////////////////////
// Return DIO capture flags for a group of DIO channels.  A "1" in a bit position
// indicates that a capture has occured on the corresponding channel.

unsigned short Cls626::DIOGroupCapStatus( unsigned short Group )
{
	// Abort if illegal DIO group number specified.
	ValidateParmRtnval( Group, MAX_DIOBANK );

	// Return event capture flags for target DIO group.
	return DEBIread( RDCapFlg[Group] );
}

////////////////////////////////////////////////////////////////////////////////////////
// Return DIO capture flag for a DIO channel.

unsigned short Cls626::DIOChanCapStatus( unsigned short Chan )
{
	// Abort if illegal DIO chan number specified.
	ValidateParmRtnval( Chan, MAX_DIOEXTCHAN );

	// Return event capture flag.
	return ( DEBIread( RDCapFlg[Chan >> 4] ) >> ( Chan & 15 ) ) & 1;
}

/////////////////////////////////////////////////////////////////////////////////////
// Enable/disable interrupts for a group of DIO channels.  A "1" in a bit position
// enables interrupts for the corresponding channel.

unsigned short Cls626::DIOGroupIntEnable( unsigned short Group)
{
	// Abort if illegal DIO group number specified.
	ValidateParmRtnval( Group, MAX_DIOBANK );

	// Return interrupt enables for target DIO group.
	return DEBIread( RDIntSel[Group] );
}

void Cls626::DIOGroupIntEnable( unsigned short Group, unsigned short Value)
{
	// Abort if illegal DIO group number specified.
	ValidateParm( Group, MAX_DIOBANK );

	// Set interrupt enables for target DIO group.
	DEBIwrite( WRIntSel[Group], Value );
}

/////////////////////////////////////////////////////////////////////////////////////
// Enable/disable interrupt for a DIO channel.

unsigned short Cls626::DIOChanIntEnable( unsigned short Chan )
{
	// Abort if illegal DIO chan number specified.
	ValidateParmRtnval( Chan, MAX_DIOEXTCHAN );

	// Return interrupt enable.
	return ( DEBIread( RDIntSel[Chan >> 4] ) >> ( Chan & 15 ) ) & 1;
}

void Cls626::DIOChanIntEnable( unsigned short Chan, unsigned short Value )
{
	// Abort if illegal DIO chan number specified.
	ValidateParm( Chan, MAX_DIOEXTCHAN );

	// Cache the group number.
	register unsigned short Group = Chan >> 4;

	// Start critical segment.
	CriticalBegin();

	// Set interrupt enable.
	DEBIwrite( WRIntSel[Group], DIOBitInsert( DEBIread(RDIntSel[Group]), Chan, Value ) );

	// End critical segment.
	CriticalEnd();
}

//////////////////////////////////////////////////////////////////////////////////////////
// Return/set operating mode of specified DIO, which can function either as a DIO or as
// counter number DIOnum's overflow output.  DIOnum must be in the range 0 to 5.

unsigned short Cls626::DIOChanMode( unsigned short DIOnum )
{
	// Abort if illegal DIO channel number specified.
	ValidateParmRtnval( DIOnum, 5 );

	// Return operating mode of target DIO.
	return WRMISC2image() & ( 1 << ( DIOnum + 4 ) );
}

void Cls626::DIOChanMode( unsigned short DIOnum, unsigned short Value )
{
	// Abort if illegal DIO channel number specified.
	ValidateParm( DIOnum, 5 );

	// Construct bit mask that will be used to generate new image of MISC2 register.
	unsigned short bitmask = 1 << ( DIOnum + 4 );

	// Start critical segment.
	CriticalBegin();

	// Construct new image of MISC2 register and write it to MISC2.
	WriteMISC2( Value ? ( WRMISC2image() | bitmask ) : ( WRMISC2image() & ~bitmask ) );

	// End critical segment.
	CriticalEnd();
}

/////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////  COUNTER FUNCTIONS  /////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
// All counter functions address a specific counter by means of the "Counter" argument,
// which is a logical counter number.  The Counter argument may have any of the following
// legal values:  0=0A, 1=1A, 2=2A, 3=0B, 4=1B, 5=2B.
/////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////
// Counter Channel class declaration.

class ClsCounter {
public:
	ClsCounter( Cls626 *pCls626, unsigned short Counter );
	// Functions that are the same for both A and B counters:
	unsigned short			GetClkMult			();
	unsigned short			GetClkPol			();
	unsigned short			GetClkSrc			();
	unsigned short			GetIndexPol			();
	unsigned short			GetIndexSrc			();
	unsigned short			GetLatchSource		();
	void					SetLatchSource		( unsigned short Value );
	unsigned long			ReadLatch			();
	void					Preload				( unsigned long Value );
	void					SetClkMult			( unsigned short Value );
	void					SetClkPol			( unsigned short Value );
	void					SetClkSrc			( unsigned short Value );
	void					SetIndexPol			( unsigned short Value );
	void					SetIndexSrc			( unsigned short Value );
	// Functions that differ for A and B counters.
	unsigned short virtual	GetEnable			()														= 0;
	unsigned short virtual	GetIntSrc			()														= 0;
	unsigned short virtual	GetLoadTrig			()														= 0;
	unsigned short virtual	GetMode				()														= 0;
	void virtual			PulseIndex			()														= 0;
	void virtual			SetEnable			( unsigned short Enable )								= 0;
	void virtual			SetIntSrc			( unsigned short IntSource )							= 0;
	void virtual			SetLoadTrig			( unsigned short Trig )									= 0;
	void virtual			SetMode				( unsigned short Setup, unsigned short DisableIntSrc )	= 0;
	void virtual			ResetCapFlags		()														= 0;
protected:
	// Storage:
	Cls626					*MyBoard;			// Pointer to parent Cls626 object.
	unsigned short			MyCRA;				// Address of CRA register.
	unsigned short			MyCRB;				// Address of CRB register.
	unsigned short			MyLatchLsw;			// Address of Latch least-significant-unsigned short register.
	unsigned short			MyEventBits[4];		// Bit translations for IntSrc --> RDMISC2.
};

/////////////////////////////////////////////////////////////////////////////////////
// CounterA and CounterB class declarations.

class ClsCounterA : public ClsCounter {
public:
	ClsCounterA( Cls626 *pCls626, unsigned short Counter );
	unsigned short	GetEnable		();
	unsigned short	GetIntSrc		();
	unsigned short	GetLoadTrig		();
	unsigned short	GetMode			();
	void			PulseIndex		();
	void			SetEnable		( unsigned short Enable );
	void			SetIntSrc		( unsigned short IntSource );
	void			SetLoadTrig		( unsigned short Trig );
	void			SetMode			( unsigned short Setup, unsigned short DisableIntSrc );
	void			ResetCapFlags	();
};

class ClsCounterB : public ClsCounter {
public:
	ClsCounterB( Cls626 *pCls626, unsigned short Counter );
	unsigned short	GetEnable		();
	unsigned short	GetIntSrc		();
	unsigned short	GetLoadTrig		();
	unsigned short	GetMode			();
	void			PulseIndex		();
	void			SetEnable		( unsigned short Enable );
	void			SetIntSrc		( unsigned short IntSource );
	void			SetLoadTrig		( unsigned short Trig );
	void			SetMode			( unsigned short Setup, unsigned short DisableIntSrc );
	void			ResetCapFlags	();
};

///////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  CLS626 COUNTER FUNCTIONS  //////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////

#define MAX_COUNTER		5				// Maximum valid counter logical channel number.

void Cls626::CountersInit()
{
	// Disable all counter interrupts and clear any captured counter events.
	for ( unsigned short c = 0; c < 6; c++ )
		Counters[c]->SetIntSrc( 0 );
}

///////////////////////////////////////////////////////////////////////////////
// Counter object constructor/destructors.

// Create counter objects.
void Cls626::CreateCounters()
{
	// Create A counters.
	for ( unsigned short c = 0; c < 3; c++ )
		Counters[c] = new ClsCounterA( this, c );

	// Create B counters.
	for ( c = 3; c < 6; c++ )
		Counters[c] = new ClsCounterB( this, c );
}

// Destroy all counter objects.
void Cls626::DestroyCounters()
{
	for ( unsigned short i = 0; i < 6; i++ )
		delete Counters[i];
}

//////////////////////////////////////////////////////////////////////////
// Return overflow and index capture flags for all counter channels.

unsigned short Cls626::CounterCapStatus()
{
	return DEBIread( LP_RDMISC2 ) & 0xFFF0;
}

//////////////////////////////////////////////////////////////////////////
// Return overflow capture status for a specific counter channel.

unsigned short Cls626::CounterOverCapGet( unsigned short Counter ) 
{
	register unsigned short mask = 1 << ( (Counter << 1) + ( (Counter > 2) ? 5 : 10 ) );
	return ( ( ( DEBIread(LP_RDMISC2) & 0xFFF0 ) & mask ) != 0 );
}

//////////////////////////////////////////////////////////////////////////
// Return index capture status for a specific counter channel.

unsigned short Cls626::CounterIndexCapGet( unsigned short Counter ) 
{
   	register unsigned short mask = 1 << ( 4 + (Counter << 1) - ( (Counter > 2) ? 5 : 0 ) );
	return ( ( ( DEBIread(LP_RDMISC2) & 0xFFF0 ) & mask ) != 0 );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Return counter setup in a format (COUNTER_SETUP) that is consistent for both A and B counters.

unsigned short Cls626::CounterModeGet( unsigned short Counter )
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Return standardized counter setup info.
	return Counters[Counter]->GetMode();
}

/////////////////////////////////////////////////////////////////////////////////////////////
// Set the operating mode for the specified counter.  Setup is treated as a COUNTER_SETUP.

void Cls626::CounterModeSet( unsigned short Counter, unsigned short Setup )
{
	// Abort if illegal counter number specified.
	ValidateParm( Counter, MAX_COUNTER );

	// Program counter operating mode.
	Counters[Counter]->SetMode( Setup, true );
}

//////////////////////////////////////////////////////////////////////////////////////////////
// Return/set a counter's enable.  Enable: 0=always enabled, NOT_0=enabled by index.

void Cls626::CounterEnableSet( unsigned short Counter, unsigned short Enable )
{
	// Abort if illegal Counter value specified.
	ValidateParm( Counter, MAX_COUNTER );

	// Convert Enable to C/C++ boolean and enable/disable the counter channel.
	Counters[Counter]->SetEnable( Enable != 0 );
}

unsigned short Cls626::CounterEnableGet( unsigned short Counter )
{
	// Abort if illegal Counter value specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Convert Enable to C/C++ boolean and enable/disable the counter channel.
	return Counters[Counter]->GetEnable();
}

/////////////////////////////////////////////////////////////////
// Latch and read the counts from the specified counter.

unsigned long Cls626::CounterReadLatch( unsigned short Counter )
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Return latched counts.
	return Counters[Counter]->ReadLatch();
}

//////////////////////////////////////////////////////////////////////////////
// Reset the specified counter's index and overflow event capture flags.

void Cls626::CounterResetCapFlags( unsigned short Counter )
{
	// Abort if illegal counter channel specified.
	ValidateParm( Counter, MAX_COUNTER );

	// Reset counter's index and overflow capture flags.
	Counters[Counter]->ResetCapFlags();
}

///////////////////////////////////////////////////////////////////////////////
// Return/set the interrupt source for the specified counter channel.
// IntSource:  0=Disabled, 1=OverflowOnly, 2=IndexOnly, 3=IndexAndOverflow.

void Cls626::CounterIntSourceSet( unsigned short Counter, unsigned short IntSource )
{
	// Abort if illegal counter channel or interrupt source specified.
	ValidateParm( Counter,   MAX_COUNTER );
	ValidateParm( IntSource, 3           );

	// Program interrupt source.
	Counters[Counter]->SetIntSrc( IntSource );
}

unsigned short Cls626::CounterIntSourceGet( unsigned short Counter )
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter,   MAX_COUNTER );

	// Return interrupt source.
	return Counters[Counter]->GetIntSrc();
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the latch trigger source of the specified counter.
// 0: On_read_access, 1: A_index_latches_A, 2: B_index_latches_B, 3: A_overflow_latches_B.

void Cls626::CounterLatchSrcSet( unsigned short Counter, unsigned short Value )
{
	// Abort if illegal counter channel or latch trigger source specified.
	ValidateParm( Counter, MAX_COUNTER );
	ValidateParm( Value,   3           );

	// Program the counter's latch source.
	Counters[Counter]->SetLatchSource( Value );
}

unsigned short Cls626::CounterLatchSrcGet( unsigned short Counter )
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Return the counter's latch source.
	return Counters[Counter]->GetLatchSource();
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the clock multiplier.

void Cls626::CounterClkMultSet( unsigned short Counter, unsigned short Value ) 
{
	// Abort if illegal counter channel or multiplier specified.
	ValidateParm( Counter, MAX_COUNTER );
	ValidateParm( Value,   3           );

	// Program the counter's clock multiplier.
	Counters[Counter]->SetClkMult( Value );
}

unsigned short Cls626::CounterClkMultGet( unsigned short Counter ) 
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Return the counter's clock multiplier.
	return Counters[Counter]->GetClkMult();
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the clock source.

void Cls626::CounterClkSrcSet( unsigned short Counter, unsigned short Value ) 
{
	// Abort if illegal counter channel or clock source specified.
	ValidateParm( Counter, MAX_COUNTER );
	ValidateParm( Value,   3           );

	// Program the counter's clock source.
	Counters[Counter]->SetClkSrc( Value );
}

unsigned short Cls626::CounterClkSrcGet( unsigned short Counter ) 
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Return the counter's clock source.
	return Counters[Counter]->GetClkSrc();
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the clock polarity.

void Cls626::CounterClkPolSet( unsigned short Counter, unsigned short Value ) 
{
	// Abort if illegal counter channel or clock polarity specified.
	ValidateParm( Counter, MAX_COUNTER );
	ValidateParm( Value,   1           );

	// Program the counter's clock polarity.
	Counters[Counter]->SetClkPol( Value );
}

unsigned short Cls626::CounterClkPolGet( unsigned short Counter ) 
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Return the counter's clock polarity.
	return Counters[Counter]->GetClkPol();
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the index polarity.

void Cls626::CounterIndexPolSet( unsigned short Counter, unsigned short Value ) 
{
	// Abort if illegal counter channel or index polarity specified.
	ValidateParm( Counter, MAX_COUNTER );
	ValidateParm( Value,   1           );

	// Program the counter's index polarity.
	Counters[Counter]->SetIndexPol( Value );
}

unsigned short Cls626::CounterIndexPolGet( unsigned short Counter ) 
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Return the counter's index polarity.
	return Counters[Counter]->GetIndexPol();
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the index source.

void Cls626::CounterIndexSrcSet( unsigned short Counter, unsigned short Value ) 
{
	// Abort if illegal counter channel or index source specified.
	ValidateParm( Counter, MAX_COUNTER );
	ValidateParm( Value,   1           );

	// Program the counter's index source.
	Counters[Counter]->SetIndexSrc( Value );
}

unsigned short Cls626::CounterIndexSrcGet( unsigned short Counter ) 
{
	// Abort if illegal counter channel specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Return the counter's index source.
	return Counters[Counter]->GetIndexSrc();
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the source that will trigger transfer of the preload register into the counter.
// 0=ThisCntr_Index, 1=ThisCntr_Overflow, 2=OverflowA (B counters only), 3=disabled.

void Cls626::CounterLoadTrigSet( unsigned short Counter, unsigned short Trig )
{
	// Abort if illegal parameter specified.
	ValidateParm( Counter, MAX_COUNTER );
	ValidateParm( Trig, 3 );

	// Program new trigger source for counter.
	Counters[Counter]->SetLoadTrig( Trig );
}

unsigned short Cls626::CounterLoadTrigGet( unsigned short Counter )
{
	// Abort if illegal parameter specified.
	ValidateParmRtnval( Counter, MAX_COUNTER );

	// Program new trigger source for counter.
	return Counters[Counter]->GetLoadTrig();
}

///////////////////////////////////////////////////////////////////
// Generate an index pulse.

void Cls626::CounterSoftIndex( unsigned short Counter )
{
	// Abort if illegal counter channel specified.
	ValidateParm( Counter, MAX_COUNTER );

	// Toggle the appropriate index to produce an index pulse.
	Counters[Counter]->PulseIndex();
}

/////////////////////////////////////////////////////////////////////
// Write Value into specified counter's preload register.

void Cls626::CounterPreload( unsigned short Counter, unsigned long Value )
{
	// Abort if illegal counter channel specified.
	ValidateParm( Counter, MAX_COUNTER );

	// Write new value to counter preload register.
	Counters[Counter]->Preload( Value );
}

///////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  CLSCOUNTER FUNCTIONS  //////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

// Enumerated counter operating modes specified by ClkSrc bit field in a COUNTER_SETUP.

#define CLKSRC_COUNTER			0		// Counter:  ENC_C clock, ENC_D direction.
#define CLKSRC_TIMER			2		// Timer:    SYS_C clock, direction specified by ClkPol.
#define CLKSRC_EXTENDER			3		// Extender: OVR_A clock, ENC_D direction.

// Enumerated counter clock multipliers.

#define MULT_X0				0x0003		// Supports no multipliers; fixed physical multiplier = 3.
#define MULT_X1				0x0002		// Supports multiplier x1; fixed physical multiplier = 2.
#define MULT_X2				0x0001		// Supports multipliers x1, x2; physical multipliers = 1 or 2.
#define MULT_X4				0x0000		// Supports multipliers x1, x2, x4; physical multipliers = 0, 1 or 2.

// Bit field positions in CRA and CRB counter control registers.

										// Bit field positions in CRA:
#define CRABIT_INDXSRC_B		14		//   B index source.
#define CRABIT_CLKSRC_B			12		//   B clock source.
#define CRABIT_INDXPOL_A		11		//   A index polarity.
#define CRABIT_LOADSRC_A		 9		//   A preload trigger.
#define CRABIT_CLKMULT_A		 7		//   A clock multiplier.
#define CRABIT_INTSRC_A			 5		//   A interrupt source.
#define CRABIT_CLKPOL_A			 4		//   A clock polarity.
#define CRABIT_INDXSRC_A		 2		//   A index source.
#define CRABIT_CLKSRC_A			 0		//   A clock source.

										// Bit field positions in CRB:
#define CRBBIT_INTRESETCMD		15		//   Interrupt reset command.
#define CRBBIT_INTRESET_B		14		//   B interrupt reset enable.
#define CRBBIT_INTRESET_A		13		//   A interrupt reset enable.
#define CRBBIT_CLKENAB_A		12		//   A clock enable.
#define CRBBIT_INTSRC_B			10		//   B interrupt source.
#define CRBBIT_LATCHSRC			 8		//   A/B latch source.
#define CRBBIT_LOADSRC_B		 6		//   B preload trigger.
#define CRBBIT_CLKMULT_B		 3		//   B clock multiplier.
#define CRBBIT_CLKENAB_B		 2		//   B clock enable.
#define CRBBIT_INDXPOL_B		 1		//   B index polarity.
#define CRBBIT_CLKPOL_B			 0		//   B clock polarity.

// Bit field masks for CRA and CRB.

#define CRAMSK_INDXSRC_B		( (unsigned short)( 3 << CRABIT_INDXSRC_B	) )
#define CRAMSK_CLKSRC_B			( (unsigned short)( 3 << CRABIT_CLKSRC_B	) )
#define CRAMSK_INDXPOL_A		( (unsigned short)( 1 << CRABIT_INDXPOL_A	) )
#define CRAMSK_LOADSRC_A		( (unsigned short)( 3 << CRABIT_LOADSRC_A	) )
#define CRAMSK_CLKMULT_A		( (unsigned short)( 3 << CRABIT_CLKMULT_A	) )
#define CRAMSK_INTSRC_A			( (unsigned short)( 3 << CRABIT_INTSRC_A	) )
#define CRAMSK_CLKPOL_A			( (unsigned short)( 3 << CRABIT_CLKPOL_A	) )
#define CRAMSK_INDXSRC_A		( (unsigned short)( 3 << CRABIT_INDXSRC_A	) )
#define CRAMSK_CLKSRC_A			( (unsigned short)( 3 << CRABIT_CLKSRC_A	) )

#define CRBMSK_INTRESETCMD		( (unsigned short)( 1 << CRBBIT_INTRESETCMD	) )
#define CRBMSK_INTRESET_B		( (unsigned short)( 1 << CRBBIT_INTRESET_B	) )
#define CRBMSK_INTRESET_A		( (unsigned short)( 1 << CRBBIT_INTRESET_A	) )
#define CRBMSK_CLKENAB_A		( (unsigned short)( 1 << CRBBIT_CLKENAB_A	) )
#define CRBMSK_INTSRC_B			( (unsigned short)( 3 << CRBBIT_INTSRC_B	) )
#define CRBMSK_LATCHSRC			( (unsigned short)( 3 << CRBBIT_LATCHSRC	) )
#define CRBMSK_LOADSRC_B		( (unsigned short)( 3 << CRBBIT_LOADSRC_B	) )
#define CRBMSK_CLKMULT_B		( (unsigned short)( 3 << CRBBIT_CLKMULT_B	) )
#define CRBMSK_CLKENAB_B		( (unsigned short)( 1 << CRBBIT_CLKENAB_B	) )
#define CRBMSK_INDXPOL_B		( (unsigned short)( 1 << CRBBIT_INDXPOL_B	) )
#define CRBMSK_CLKPOL_B			( (unsigned short)( 1 << CRBBIT_CLKPOL_B	) )

#define CRBMSK_INTCTRL			( CRBMSK_INTRESETCMD | CRBMSK_INTRESET_A | CRBMSK_INTRESET_B )	// Interrupt reset control bits.

// Bit field positions for standardized SETUP structure.

#define STDBIT_INTSRC			13
#define STDBIT_LATCHSRC			11
#define STDBIT_LOADSRC			 9
#define STDBIT_INDXSRC			 7
#define STDBIT_INDXPOL			 6
#define STDBIT_CLKSRC			 4
#define STDBIT_CLKPOL			 3
#define STDBIT_CLKMULT			 1
#define STDBIT_CLKENAB			 0

// Bit field masks for standardized SETUP structure.

#define STDMSK_INTSRC			( (unsigned short)( 3 << STDBIT_INTSRC   ) )
#define STDMSK_LATCHSRC			( (unsigned short)( 3 << STDBIT_LATCHSRC ) )
#define STDMSK_LOADSRC			( (unsigned short)( 3 << STDBIT_LOADSRC  ) )
#define STDMSK_INDXSRC			( (unsigned short)( 1 << STDBIT_INDXSRC  ) )
#define STDMSK_INDXPOL			( (unsigned short)( 1 << STDBIT_INDXPOL  ) )
#define STDMSK_CLKSRC			( (unsigned short)( 3 << STDBIT_CLKSRC   ) )
#define STDMSK_CLKPOL			( (unsigned short)( 1 << STDBIT_CLKPOL   ) )
#define STDMSK_CLKMULT			( (unsigned short)( 3 << STDBIT_CLKMULT  ) )
#define STDMSK_CLKENAB			( (unsigned short)( 1 << STDBIT_CLKENAB  ) )

/////////////////////////////////////////////////////////////////
// Read a counter's output latch.

unsigned long ClsCounter::ReadLatch()
{
	// Start critical segment.
	MyBoard->CriticalBegin();

	// Latch counts and fetch LSW of latched counts value.
	register unsigned long value = (unsigned long)MyBoard->DEBIread( MyLatchLsw );

	// Fetch MSW of latched counts and combine with LSW.
	value |= ( (unsigned long)MyBoard->DEBIread( MyLatchLsw + 2 ) << 16 );

	// End critical segment.
	MyBoard->CriticalEnd();
	
	// Return latched counts.
	return value;
}

///////////////////////////////////////////////////////////////////
// Reset a counter's index and overflow event capture flags.

void ClsCounterA::ResetCapFlags()
{
	MyBoard->DEBIreplace( MyCRB, ~CRBMSK_INTCTRL, CRBMSK_INTRESETCMD | CRBMSK_INTRESET_A );
}

void ClsCounterB::ResetCapFlags()
{
	MyBoard->DEBIreplace( MyCRB, ~CRBMSK_INTCTRL, CRBMSK_INTRESETCMD | CRBMSK_INTRESET_B );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Return counter setup in a format (COUNTER_SETUP) that is consistent for both A and B counters.

unsigned short ClsCounterA::GetMode()
{
	// Fetch CRA and CRB register images.
	MyBoard->CriticalBegin();
	register unsigned short cra = MyBoard->DEBIread( MyCRA );
	register unsigned short crb = MyBoard->DEBIread( MyCRB );
	MyBoard->CriticalEnd();

	// Populate the standardized counter setup bit fields.  Note: IndexSrc is restricted to ENC_X or IndxPol.
	register unsigned short setup = ( ( cra & STDMSK_LOADSRC )								// LoadSrc  = LoadSrcA.
		| ( ( crb << ( STDBIT_LATCHSRC  - CRBBIT_LATCHSRC        ) ) & STDMSK_LATCHSRC )	// LatchSrc = LatchSrcA.
		| ( ( cra << ( STDBIT_INTSRC    - CRABIT_INTSRC_A        ) ) & STDMSK_INTSRC   )	// IntSrc   = IntSrcA.
		| ( ( cra << ( STDBIT_INDXSRC   - (CRABIT_INDXSRC_A + 1) ) ) & STDMSK_INDXSRC  )	// IndxSrc  = IndxSrcA<1>.
		| ( ( cra >> ( CRABIT_INDXPOL_A - STDBIT_INDXPOL         ) ) & STDMSK_INDXPOL  )	// IndxPol  = IndxPolA.
		| ( ( crb >> ( CRBBIT_CLKENAB_A - STDBIT_CLKENAB         ) ) & STDMSK_CLKENAB  ) );	// ClkEnab  = ClkEnabA.

	// Adjust mode-dependent parameters.
	if ( cra & ( 2 << CRABIT_CLKSRC_A ) )													// If Timer mode (ClkSrcA<1> == 1):
		setup |= ( ( CLKSRC_TIMER << STDBIT_CLKSRC )										//   Indicate Timer mode.
			| ( ( cra << ( STDBIT_CLKPOL - CRABIT_CLKSRC_A ) ) & STDMSK_CLKPOL )			//   Set ClkPol to indicate count direction (ClkSrcA<0>).
			| ( MULT_X1 << STDBIT_CLKMULT ) );												//   ClkMult must be 1x in Timer mode.

	else 																					// If Counter mode (ClkSrcA<1> == 0):
		setup |= ( ( CLKSRC_COUNTER << STDBIT_CLKSRC )										//   Indicate Counter mode.
			| ( ( cra >> ( CRABIT_CLKPOL_A - STDBIT_CLKPOL ) ) & STDMSK_CLKPOL )			//   Pass through ClkPol.
			| ( ( ( cra & CRAMSK_CLKMULT_A ) == ( MULT_X0 << CRABIT_CLKMULT_A ) ) ?			//   Force ClkMult to 1x if not legal, else pass through.
				( MULT_X1 << STDBIT_CLKMULT ) :
				( ( cra >> ( CRABIT_CLKMULT_A - STDBIT_CLKMULT ) ) & STDMSK_CLKMULT ) ) );

	// Return adjusted counter setup.
	return setup;
}

unsigned short ClsCounterB::GetMode()
{
	// Fetch CRA and CRB register images.
	MyBoard->CriticalBegin();
	register unsigned short cra = MyBoard->DEBIread( MyCRA );
	register unsigned short crb = MyBoard->DEBIread( MyCRB );
	MyBoard->CriticalEnd();

	// Populate the standardized counter setup bit fields.  Note: IndexSrc is restricted to ENC_X or IndxPol.
	register unsigned short setup = 
		( ( ( crb << ( STDBIT_INTSRC          - CRBBIT_INTSRC_B  ) ) & STDMSK_INTSRC   )	// IntSrc   = IntSrcB.
		| ( ( crb << ( STDBIT_LATCHSRC        - CRBBIT_LATCHSRC  ) ) & STDMSK_LATCHSRC )	// LatchSrc = LatchSrcB.
		| ( ( crb << ( STDBIT_LOADSRC         - CRBBIT_LOADSRC_B ) ) & STDMSK_LOADSRC  )	// LoadSrc  = LoadSrcB.
		| ( ( crb << ( STDBIT_INDXPOL         - CRBBIT_INDXPOL_B ) ) & STDMSK_INDXPOL  )	// IndxPol  = IndxPolB.
		| ( ( crb >> ( CRBBIT_CLKENAB_B       - STDBIT_CLKENAB   ) ) & STDMSK_CLKENAB  )	// ClkEnab  = ClkEnabB.
		| ( ( cra >> ( (CRABIT_INDXSRC_B + 1) - STDBIT_INDXSRC   ) ) & STDMSK_INDXSRC  ) );	// IndxSrc  = IndxSrcB<1>.

	// Adjust mode-dependent parameters.
	if ( ( crb & CRBMSK_CLKMULT_B ) == ( MULT_X0 << CRBBIT_CLKMULT_B ) )					// If Extender mode (ClkMultB == MULT_X0):
		setup |= ( ( CLKSRC_EXTENDER << STDBIT_CLKSRC )										//   Indicate Extender mode.
			| ( MULT_X1 << STDBIT_CLKMULT )													//   Indicate multiplier is 1x.
			| ( ( cra >> ( CRABIT_CLKSRC_B - STDBIT_CLKPOL ) ) & STDMSK_CLKPOL ) );			//   Set ClkPol equal to Timer count direction (ClkSrcB<0>).

	else if ( cra & ( 2 << CRABIT_CLKSRC_B ) )												// If Timer mode (ClkSrcB<1> == 1):
		setup |= ( ( CLKSRC_TIMER << STDBIT_CLKSRC )										//   Indicate Timer mode.
			| ( MULT_X1 << STDBIT_CLKMULT )													//   Indicate multiplier is 1x.
			| ( ( cra >> ( CRABIT_CLKSRC_B - STDBIT_CLKPOL ) ) & STDMSK_CLKPOL ) );			//   Set ClkPol equal to Timer count direction (ClkSrcB<0>).

	else																					// If Counter mode (ClkSrcB<1> == 0):
		setup |= ( ( CLKSRC_COUNTER << STDBIT_CLKSRC )										//   Indicate Timer mode.
			| ( ( crb >> ( CRBBIT_CLKMULT_B - STDBIT_CLKMULT ) ) & STDMSK_CLKMULT )			//   Clock multiplier is passed through.	
			| ( ( crb << ( STDBIT_CLKPOL - CRBBIT_CLKPOL_B ) ) & STDMSK_CLKPOL ) );			//   Clock polarity is passed through.	

	// Return adjusted counter setup.
	return setup;
}

/////////////////////////////////////////////////////////////////////////////////////////////
// Set the operating mode for the specified counter.  The setup parameter is treated as
// a COUNTER_SETUP data type.  The following parameters are programmable (all other parms
// are ignored):  ClkMult, ClkPol, ClkEnab, IndexSrc, IndexPol, LoadSrc.

void ClsCounterA::SetMode( unsigned short Setup, unsigned short DisableIntSrc )
{
	// Cache the Standard Setup.
	register unsigned short setup = Setup;

	// Initialize CRA and CRB images.
	register unsigned short cra = (	( setup & CRAMSK_LOADSRC_A )							// Preload trigger is passed through.
		| ( ( setup & STDMSK_INDXSRC ) >> ( STDBIT_INDXSRC - (CRABIT_INDXSRC_A + 1) ) ) );	// IndexSrc is restricted to ENC_X or IndxPol.
	
	register unsigned short crb = ( CRBMSK_INTRESETCMD | CRBMSK_INTRESET_A					// Reset any pending CounterA event captures.
		| ( ( setup & STDMSK_CLKENAB ) << ( CRBBIT_CLKENAB_A - STDBIT_CLKENAB ) ) );		// Clock enable is passed through.
	
	// Force IntSrc to Disabled if DisableIntSrc is asserted.
	if ( !DisableIntSrc )
		cra |= ( ( setup & STDMSK_INTSRC ) >> ( STDBIT_INTSRC - CRABIT_INTSRC_A ) );

	// Populate all mode-dependent attributes of CRA & CRB images.
	switch ( ( setup & STDMSK_CLKSRC ) >> STDBIT_CLKSRC )
	{
	case CLKSRC_EXTENDER:																	// Extender Mode: Force to Timer mode (Extender valid only for B counters).

	case CLKSRC_TIMER:																		// Timer Mode:
		cra |= ( ( 2 << CRABIT_CLKSRC_A )													//   ClkSrcA<1> selects system clock
			| ( ( setup & STDMSK_CLKPOL ) >> ( STDBIT_CLKPOL - CRABIT_CLKSRC_A ) )			//     with count direction (ClkSrcA<0>) obtained from ClkPol.
			| ( 1 << CRABIT_CLKPOL_A )														//   ClkPolA behaves as always-on clock enable.
			| ( MULT_X1 << CRABIT_CLKMULT_A ) );											//   ClkMult must be 1x.
		break;

	default:																				// Counter Mode:
		cra |= ( CLKSRC_COUNTER																//   Select ENC_C and ENC_D as clock/direction inputs.
			| ( ( setup & STDMSK_CLKPOL ) << ( CRABIT_CLKPOL_A - STDBIT_CLKPOL ) )			//   Clock polarity is passed through.
			| ( ( ( setup & STDMSK_CLKMULT ) == ( MULT_X0 << STDBIT_CLKMULT ) ) ?			//   Force multiplier to x1 if not legal, otherwise pass through.
				( MULT_X1 << CRABIT_CLKMULT_A ) :
				( ( setup & STDMSK_CLKMULT ) << ( CRABIT_CLKMULT_A - STDBIT_CLKMULT ) ) ) );
	}

	// Force positive index polarity if IndxSrc is software-driven only, otherwise pass it through.
	if ( ~setup & STDMSK_INDXSRC )
		cra |= ( ( setup & STDMSK_INDXPOL ) << ( CRABIT_INDXPOL_A - STDBIT_INDXPOL ) );

	// Start critical segment.
	MyBoard->CriticalBegin();

	// If IntSrc has been forced to Disabled, update the MISC2 interrupt enable mask to indicate the counter interrupt is disabled.
	if ( DisableIntSrc )
		MyBoard->MyCounterIntEnabs &= ~MyEventBits[3];

	// While retaining CounterB and LatchSrc configurations, program the new counter operating mode.
	MyBoard->DEBIreplace( MyCRA, CRAMSK_INDXSRC_B | CRAMSK_CLKSRC_B, cra );
	MyBoard->DEBIreplace( MyCRB, ~( CRBMSK_INTCTRL | CRBMSK_CLKENAB_A ), crb );

	// End critical segment.
	MyBoard->CriticalEnd();
}

void ClsCounterB::SetMode( unsigned short Setup, unsigned short DisableIntSrc )
{
	// Cache the Standard Setup.
	register unsigned short setup = Setup;

	// Initialize CRA and CRB images.
	register unsigned short cra = (
		( setup & STDMSK_INDXSRC ) << ( (CRABIT_INDXSRC_B + 1) - STDBIT_INDXSRC ) );	// IndexSrc field is restricted to ENC_X or IndxPol.

	register unsigned short crb = ( CRBMSK_INTRESETCMD | CRBMSK_INTRESET_B				// Reset event captures and disable interrupts.
		| ( ( setup & STDMSK_CLKENAB ) << ( CRBBIT_CLKENAB_B - STDBIT_CLKENAB ) )		// Clock enable is passed through.
		| ( ( setup & STDMSK_LOADSRC ) >> ( STDBIT_LOADSRC - CRBBIT_LOADSRC_B ) ) );	// Preload trigger source is passed through.

	// Force IntSrc to Disabled if DisableIntSrc is asserted.
	if ( !DisableIntSrc )
		crb |= ( ( setup & STDMSK_INTSRC ) >> ( STDBIT_INTSRC - CRBBIT_INTSRC_B ) );

	// Populate all mode-dependent attributes of CRA & CRB images.
	switch ( ( setup & STDMSK_CLKSRC ) >> STDBIT_CLKSRC )
	{
	case CLKSRC_TIMER:																	// Timer Mode:
		cra |= ( ( 2 << CRABIT_CLKSRC_B )												//   ClkSrcB<1> selects system clock
			| ( ( setup & STDMSK_CLKPOL ) << ( CRABIT_CLKSRC_B - STDBIT_CLKPOL ) ) );	//     with direction (ClkSrcB<0>) obtained from ClkPol.
		crb	|= ( ( 1 << CRBBIT_CLKPOL_B )												//   ClkPolB behaves as always-on clock enable.
			| ( MULT_X1 << CRBBIT_CLKMULT_B ) );										//   ClkMultB must be 1x.
		break;

	case CLKSRC_EXTENDER:																// Extender Mode:
		cra |= ( ( 2 << CRABIT_CLKSRC_B )												//   ClkSrcB source is OverflowA (same as "timer")
			| ( ( setup & STDMSK_CLKPOL ) << ( CRABIT_CLKSRC_B - STDBIT_CLKPOL ) ) );	//     with direction obtained from ClkPol.
		crb |= ( ( 1 << CRBBIT_CLKPOL_B )												//   ClkPolB controls IndexB -- always set to active.
			| ( MULT_X0 << CRBBIT_CLKMULT_B ) );										//   ClkMultB selects OverflowA as the clock source.
		break;

	default:																			// Counter Mode:
		cra |= ( CLKSRC_COUNTER << CRABIT_CLKSRC_B );									//   Select ENC_C and ENC_D as clock/direction inputs.
		crb |= ( ( ( setup & STDMSK_CLKPOL ) >> ( STDBIT_CLKPOL - CRBBIT_CLKPOL_B ) )	//   ClkPol is passed through.
			| ( ( ( setup & STDMSK_CLKMULT ) == ( MULT_X0 << STDBIT_CLKMULT ) ) ?		//   Force ClkMult to x1 if not legal, otherwise pass through.
			( MULT_X1 << CRBBIT_CLKMULT_B ) :
			( ( setup & STDMSK_CLKMULT ) << ( CRBBIT_CLKMULT_B - STDBIT_CLKMULT ) ) ) );
	}

	// Force positive index polarity if IndxSrc is software-driven only, otherwise pass it through.
	if ( ~setup & STDMSK_INDXSRC )
		crb |= ( ( setup & STDMSK_INDXPOL ) >> ( STDBIT_INDXPOL - CRBBIT_INDXPOL_B ) );

	// Start critical segment.
	MyBoard->CriticalBegin();

	// If IntSrc has been forced to Disabled, update the MISC2 interrupt enable mask to indicate the counter interrupt is disabled.
	if ( DisableIntSrc )
		MyBoard->MyCounterIntEnabs &= ~MyEventBits[3];

	// While retaining CounterA and LatchSrc configurations, program the new counter operating mode.
	MyBoard->DEBIreplace( MyCRA, ~( CRAMSK_INDXSRC_B | CRAMSK_CLKSRC_B ), cra );
	MyBoard->DEBIreplace( MyCRB, CRBMSK_CLKENAB_A | CRBMSK_LATCHSRC, crb );

	// End critical segment.
	MyBoard->CriticalEnd();
}

//////////////////////////////////////////////////////////////////////////////////
// Return/set a counter's enable.  Enable: 0=always enabled, 1=enabled by index.

void ClsCounterA::SetEnable( unsigned short Enable )
{
	MyBoard->DEBIreplace( MyCRB, ~( CRBMSK_INTCTRL | CRBMSK_CLKENAB_A ), Enable << CRBBIT_CLKENAB_A );
}

void ClsCounterB::SetEnable( unsigned short Enable )
{
	MyBoard->DEBIreplace( MyCRB, ~( CRBMSK_INTCTRL | CRBMSK_CLKENAB_B ), Enable << CRBBIT_CLKENAB_B );
}

unsigned short ClsCounterA::GetEnable()
{
	return ( MyBoard->DEBIread(MyCRB) >> CRBBIT_CLKENAB_A ) & 1;
}

unsigned short ClsCounterB::GetEnable()
{
	return ( MyBoard->DEBIread(MyCRB) >> CRBBIT_CLKENAB_B ) & 1;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set a counter pair's latch trigger source.
// 0: On read access, 1: A index latches A, 2: B index latches B, 3: A overflow latches B.

void ClsCounter::SetLatchSource( unsigned short Value )
{
	MyBoard->DEBIreplace( MyCRB, ~( CRBMSK_INTCTRL | CRBMSK_LATCHSRC ), Value << CRBBIT_LATCHSRC );
}

unsigned short ClsCounter::GetLatchSource()
{
	return ( MyBoard->DEBIread(MyCRB) >> CRBBIT_LATCHSRC ) & 3;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the event that will trigger transfer of the preload register into the counter.
// 0=ThisCntr_Index, 1=ThisCntr_Overflow, 2=OverflowA (B counters only), 3=disabled.

void ClsCounterA::SetLoadTrig( unsigned short Trig )
{
	MyBoard->DEBIreplace( MyCRA, ~CRAMSK_LOADSRC_A, Trig << CRABIT_LOADSRC_A );
}

void ClsCounterB::SetLoadTrig( unsigned short Trig )
{
	MyBoard->DEBIreplace( MyCRB, ~( CRBMSK_LOADSRC_B | CRBMSK_INTCTRL ), Trig << CRBBIT_LOADSRC_B );
}

unsigned short ClsCounterA::GetLoadTrig()
{
	return ( MyBoard->DEBIread(MyCRA) >> CRABIT_LOADSRC_A ) & 3;
}

unsigned short ClsCounterB::GetLoadTrig()
{
	return ( MyBoard->DEBIread(MyCRB) >> CRBBIT_LOADSRC_B ) & 3;
}

//////////////////////////////////////////////////////////////////////////////////
// Return/set counter interrupt source and clear any captured index/overflow events.
// IntSource: 0=Disabled, 1=OverflowOnly, 2=IndexOnly, 3=IndexAndOverflow.

void ClsCounterA::SetIntSrc( unsigned short IntSource )
{
	MyBoard->CriticalBegin();		// Start critical segment.
	
	// Reset any pending counter overflow or index captures.
	MyBoard->DEBIreplace( MyCRB, ~CRBMSK_INTCTRL, CRBMSK_INTRESETCMD | CRBMSK_INTRESET_A );

	// Program counter interrupt source.
	MyBoard->DEBIreplace( MyCRA, ~CRAMSK_INTSRC_A, IntSource << CRABIT_INTSRC_A );

	// Update MISC2 interrupt enable mask.
	MyBoard->MyCounterIntEnabs = ( MyBoard->MyCounterIntEnabs & ~MyEventBits[3] ) | MyEventBits[IntSource];

	MyBoard->CriticalEnd();			// End critical segment.
}

void ClsCounterB::SetIntSrc( unsigned short IntSource )
{
	MyBoard->CriticalBegin();		// Start critical segment.
	
	// Cache writeable CRB register image.
	unsigned short crb = MyBoard->DEBIread( MyCRB ) & ~CRBMSK_INTCTRL;
	
	// Reset any pending counter overflow or index captures.
	MyBoard->DEBIwrite( MyCRB, crb | CRBMSK_INTRESETCMD | CRBMSK_INTRESET_B );

	// Program counter interrupt source.
	MyBoard->DEBIwrite( MyCRB, ( crb & ~CRBMSK_INTSRC_B ) | ( IntSource << CRBBIT_INTSRC_B ) );

	// Update MISC2 interrupt enable mask.
	MyBoard->MyCounterIntEnabs = ( MyBoard->MyCounterIntEnabs & ~MyEventBits[3] ) | MyEventBits[IntSource];

	MyBoard->CriticalEnd();			// End critical segment.
}

unsigned short ClsCounterA::GetIntSrc()
{
	return ( MyBoard->DEBIread(MyCRA) >> CRABIT_INTSRC_A ) & 3;
}

unsigned short ClsCounterB::GetIntSrc()
{
	return ( MyBoard->DEBIread(MyCRB) >> CRBBIT_INTSRC_B ) & 3;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the clock multiplier.

void ClsCounter::SetClkMult( unsigned short Value ) 
{
	SetMode( ( GetMode() & ~STDMSK_CLKMULT ) | ( Value << STDBIT_CLKMULT ), false );
}

unsigned short ClsCounter::GetClkMult() 
{
	return ( GetMode() >> STDBIT_CLKMULT ) & 3;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the clock polarity.

void ClsCounter::SetClkPol( unsigned short Value ) 
{
	SetMode( ( GetMode() & ~STDMSK_CLKPOL ) | ( Value << STDBIT_CLKPOL ), false );
}

unsigned short ClsCounter::GetClkPol() 
{
	return ( GetMode() >> STDBIT_CLKPOL ) & 1;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the clock source.

void ClsCounter::SetClkSrc( unsigned short Value ) 
{
	SetMode( ( GetMode() & ~STDMSK_CLKSRC ) | ( Value << STDBIT_CLKSRC ), false );
}

unsigned short ClsCounter::GetClkSrc() 
{
	return ( GetMode() >> STDBIT_CLKSRC ) & 3;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the index polarity.

void ClsCounter::SetIndexPol( unsigned short Value ) 
{
	SetMode( ( GetMode() & ~STDMSK_INDXPOL ) | ( (Value != 0) << STDBIT_INDXPOL ), false );
}

unsigned short ClsCounter::GetIndexPol() 
{
	return ( GetMode() >> STDBIT_INDXPOL ) & 1;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Return/set the index source.

void ClsCounter::SetIndexSrc( unsigned short Value ) 
{
	SetMode( ( GetMode() & ~STDMSK_INDXSRC ) | ( (Value != 0) << STDBIT_INDXSRC ), false );
}

unsigned short ClsCounter::GetIndexSrc() 
{
	return ( GetMode() >> STDBIT_INDXSRC ) & 1;
}

///////////////////////////////////////////////////////////////////
// Generate an index pulse.

void ClsCounterA::PulseIndex()
{
	MyBoard->CriticalBegin();													// Start critical segment.
	register unsigned short cra = MyBoard->DEBIread( MyCRA );					// Pulse index.
	MyBoard->DEBIwrite( MyCRA, cra ^ CRAMSK_INDXPOL_A );
	MyBoard->DEBIwrite( MyCRA, cra );
	MyBoard->CriticalEnd();														// End critical segment.
}

void ClsCounterB::PulseIndex()
{
	MyBoard->CriticalBegin();													// Start critical segment.
	register unsigned short crb = MyBoard->DEBIread( MyCRB ) & ~CRBMSK_INTCTRL;	// Pulse index.
	MyBoard->DEBIwrite( MyCRB, crb ^ CRBMSK_INDXPOL_B );
	MyBoard->DEBIwrite( MyCRB, crb);
	MyBoard->CriticalEnd();														// End critical segment.
}

/////////////////////////////////////////////////////////
// Write Value into counter preload register.

void ClsCounter::Preload( unsigned long Value )
{
	MyBoard->CriticalBegin();													// Start critical segment.
	MyBoard->DEBIwrite( MyLatchLsw,     (unsigned short)  Value         );		// Write value to preload register.
	MyBoard->DEBIwrite( MyLatchLsw + 2, (unsigned short)( Value >> 16 ) );
	MyBoard->CriticalEnd();														// End critical segment.
}

///////////////////////////////////////////////////////////////////////////////
// Counter class constructors.

// Register address to use for PreLoad and Latched counts LSW.
static const unsigned short AdrsCntrLSW[] = { LP_CNTR0ALSW, LP_CNTR1ALSW, LP_CNTR2ALSW, LP_CNTR0BLSW, LP_CNTR1BLSW, LP_CNTR2BLSW };

// CRA register address to use.
static const unsigned short AdrsCRA[] = { LP_CR0A, LP_CR1A, LP_CR2A, LP_CR0A, LP_CR1A, LP_CR2A };

// CRB register address to use.
static const unsigned short AdrsCRB[] = { LP_CR0B, LP_CR1B, LP_CR2B, LP_CR0B, LP_CR1B, LP_CR2B };

// Counter overflow/index event flag masks for RDMISC2.
#define INDXMASK(C)		( 1 << ( ( (C) > 2 ) ? ( (C) * 2 - 1 ) : ( (C) * 2 +  4 ) ) )
#define OVERMASK(C)		( 1 << ( ( (C) > 2 ) ? ( (C) * 2 + 5 ) : ( (C) * 2 + 10 ) ) )
#define EVBITS(C)		{ 0, OVERMASK(C), INDXMASK(C), OVERMASK(C) | INDXMASK(C) }

// Translation table to map IntSrc into equivalent RDMISC2 event flag bits.
static const unsigned short EventBits[][4] = { EVBITS(0), EVBITS(1), EVBITS(2), EVBITS(3), EVBITS(4), EVBITS(5) };

// Constructor for counter base class.
ClsCounter::ClsCounter( Cls626 *pCls626, unsigned short Counter )
{
	MyBoard			= pCls626;						// Pointer to parent Cls626 object.
	MyCRA			= AdrsCRA[Counter];				// Address of CRA register.
	MyCRB			= AdrsCRB[Counter];				// Address of CRB register.
	MyLatchLsw		= AdrsCntrLSW[Counter];			// Address of Preload & Latch LSW registers.
	for ( unsigned short i = 0; i < 4; i++ )
		MyEventBits[i] = EventBits[Counter][i];		// Event flag bit overlays for RDMISC2.
}

// Constructors for CounterA and CounterB classes.
ClsCounterA::ClsCounterA( Cls626 *pCls626, unsigned short Counter ) : ClsCounter( pCls626, Counter ) {}
ClsCounterB::ClsCounterB( Cls626 *pCls626, unsigned short Counter ) : ClsCounter( pCls626, Counter ) {}
