//////////////////////////////////////////////////////////////
// Module    : Cls626.h
// Function  : Header file for Sensoray 626 board class.
// Usage     : Included by OS interface class.
// Author    : Jim Lamberson
// Copyright : (C) 2000 Sensoray
//////////////////////////////////////////////////////////////

#ifndef DEF_CLS626
#define DEF_CLS626

#include "App626.h"

// Type definitions ------------------------------------

//typedef void	(* FPTR_ERR)(unsigned long ErrFlags);	// Pointer to application's error callback function.

// DMA buffer description used by Board Class.
typedef struct {
	unsigned long PhysicalBase;		// Physical base address.
	void*			    LogicalBase;		// Logical base address.
	unsigned long	DMAHandle;			// Buffer ID/reference used by interface class.
  unsigned long bufByteSize;    // allocated size in bytes of buffer
} DMABUF;

// Used by Board Class to specify a register access operation to the Interface Class.
typedef struct {
	unsigned long	IsWrite;			// Boolean: TRUE = dword write, FALSE = dword read.
	unsigned long	AdrsOS;				// Register address offset.
	unsigned long	DataVal;			// Data value (applicable only for write operations).
} ACCESS_SPEC;

typedef struct {
	unsigned long	NumActions;			// Number of actions in list.
	ACCESS_SPEC		Actions[20];		// List of access actions.
} ACTION_LIST;

typedef struct {
	unsigned long	StatusAndMask;		// AND bit mask for interrupt status flag.
	unsigned long	StatusXorMask;		// XOR bit mask for interrupt status flag.
	ACTION_LIST		StatusActions;		// Actions that will retrieve the board's IRQ status.
	ACTION_LIST		ProcessActions;		// Actions that will negate board's IRQ.
} INT_ACTIONS;

// PCI device and vendor IDs.
#define VENDORID626			0x1131		// PCI VendorID for Philips.
#define DEVICEID626			0x7146		// PCI DeviceID for SAA7146 PCI bridge.
#define SUBVENDORID626		0x6000		// PCI SubVendorID for Sensoray.
#define SUBDEVICEID626		626			// PCI SubDeviceID for Model 626.

// Array of interrupt actions to be performed by kernel-mode interrupt handler.
extern INT_ACTIONS IntActions;

/////////////////////////////////////////////////////////////////////////////////////
// Model 626 board class header.
/////////////////////////////////////////////////////////////////////////////////////

class ClsCounter;		// Forward reference to counter channel class.

class Cls626 {
public:
	Cls626()									;
	// Destructor is virtual so that derived class destructors will execute.
	virtual					~Cls626()			;
	// Primitive functions that may be substituted in a derived dummy OS interface class:
	virtual void			DEBIinit			();
	virtual unsigned short	DEBIread			( unsigned short addr );
	virtual void			DEBIwrite			( unsigned short addr, unsigned short wdata );
	virtual void			DEBIreplace			( unsigned short addr, unsigned short mask, unsigned short wdata );
	virtual unsigned char	I2Cread				( unsigned char addr );
	virtual void			I2Cwrite			( unsigned char addr, unsigned char wdata );
	virtual void			InterruptEnable		( unsigned short enable );		// InterruptEnableSet.
	virtual unsigned short	InterruptEnableGet	();
	virtual void			InterruptMask		();
	virtual unsigned short	InterruptPending	();
	virtual void			InterruptStatus		( unsigned short *IntFlags );
	// High-level functions that may be substituted in a derived dummy OS interface class.
	//   Analog input functions:
	virtual void			ReadADC				( unsigned short *pdata );
	virtual void			ResetADC			( unsigned char *ppl );
	virtual unsigned short	SingleADC			( unsigned short AdcSpec );
	void					StartADC			();
	void					WaitDoneADC			( unsigned short *pdata );
	//   Analog output functions:
	virtual void			WriteDAC			( unsigned short chan, short dacdata );
	short					ReadDAC				( unsigned short chan );
	virtual void			WriteTrimDAC		( unsigned char LogicalChan, unsigned char DacData );
	unsigned char			ReadTrimDAC			( unsigned char LogicalChan );
	//   Battery functions:
	virtual unsigned short	BackupEnable		();
	virtual void			BackupEnable		( unsigned short Enable );
	virtual unsigned short	ChargeEnable		();
	virtual void			ChargeEnable		( unsigned short Enable );
	//   Watchdog functions:
	virtual unsigned short	WatchdogTimeout		();
	virtual unsigned short	WatchdogEnable		();
	virtual void			WatchdogEnable		( unsigned short Enable );
	virtual unsigned short	WatchdogPeriod		();
	virtual void			WatchdogPeriod		( unsigned short Value );
	virtual void			WatchdogReset		();
	//   Digital I/O functions:
	virtual void			DIOChanCapEnable	( unsigned short Chan, unsigned short Enable );
	virtual unsigned short	DIOChanCapEnable	( unsigned short Chan );
	virtual void			DIOChanCapReset		( unsigned short Chan );
	virtual unsigned short	DIOChanCapStatus	( unsigned short Chan );
	virtual unsigned short	DIOChanEdge			( unsigned short Chan );
	virtual void			DIOChanEdge			( unsigned short Chan, unsigned short Value );
	virtual unsigned short	DIOChanIntEnable	( unsigned short Chan );
	virtual void			DIOChanIntEnable	( unsigned short Chan, unsigned short Value );
	virtual unsigned short	DIOChanMode			( unsigned short Chan );
	virtual void			DIOChanMode			( unsigned short Chan, unsigned short Value );
	virtual unsigned short	DIOChanRead			( unsigned short Chan );
	virtual unsigned short	DIOChanWrite		( unsigned short Chan );
	virtual void			DIOChanWrite		( unsigned short Chan, unsigned short Value );
	virtual void			DIOGroupCapEnable	( unsigned short Group, unsigned short Value, unsigned short Enable );
	virtual unsigned short	DIOGroupCapEnable	( unsigned short Group );
	virtual void			DIOGroupCapReset	( unsigned short Group, unsigned short Value );
	virtual unsigned short	DIOGroupCapStatus	( unsigned short Group );
	virtual unsigned short	DIOGroupEdge		( unsigned short Group );
	virtual void			DIOGroupEdge		( unsigned short Group, unsigned short Value );
	virtual unsigned short	DIOGroupIntEnable	( unsigned short Group );
	virtual void			DIOGroupIntEnable	( unsigned short Group, unsigned short Value );
	virtual unsigned short	DIOGroupRead		( unsigned short Group );
	virtual unsigned short	DIOGroupWrite		( unsigned short Group );
	virtual void			DIOGroupWrite		( unsigned short Group, unsigned short Value );
	//   Counter functions:
	virtual unsigned short	CounterCapStatus	();
	virtual void			CounterClkMultSet	( unsigned short Counter, unsigned short Value );
	virtual unsigned short	CounterClkMultGet	( unsigned short Counter );
	virtual void			CounterClkSrcSet	( unsigned short Counter, unsigned short Value );
	virtual unsigned short	CounterClkSrcGet	( unsigned short Counter );
	virtual void			CounterClkPolSet	( unsigned short Counter, unsigned short Value );
	virtual unsigned short	CounterClkPolGet	( unsigned short Counter );
	virtual unsigned short	CounterEnableGet	( unsigned short Counter );
	virtual void			CounterEnableSet	( unsigned short Counter, unsigned short Enable );
	virtual unsigned short	CounterIndexCapGet	( unsigned short Counter );
	virtual void			CounterIndexPolSet	( unsigned short Counter, unsigned short Value );
	virtual unsigned short	CounterIndexPolGet	( unsigned short Counter );
	virtual void			CounterIndexSrcSet	( unsigned short Counter, unsigned short Value );
	virtual unsigned short	CounterIndexSrcGet	( unsigned short Counter );
	virtual unsigned short	CounterIntSourceGet	( unsigned short Counter );
	virtual void			CounterIntSourceSet	( unsigned short Counter, unsigned short IntSource );
	virtual unsigned short	CounterLatchSrcGet	( unsigned short Counter );
	virtual void			CounterLatchSrcSet	( unsigned short Counter, unsigned short Value );
	virtual unsigned short	CounterLoadTrigGet	( unsigned short Counter );
	virtual void			CounterLoadTrigSet	( unsigned short Counter, unsigned short Trig );
	virtual unsigned short	CounterModeGet		( unsigned short Counter );
	virtual void			CounterModeSet		( unsigned short Counter, unsigned short Options );
	virtual unsigned short	CounterOverCapGet	( unsigned short Counter );
	virtual void			CounterPreload		( unsigned short Counter, unsigned long Value );
	virtual unsigned long	CounterReadLatch	( unsigned short Counter );
	virtual void			CounterResetCapFlags( unsigned short Counter );
	virtual void			CounterSoftIndex	( unsigned short Counter );
	// Functions that must be implemented in the derived OS interface class:
	virtual void			CriticalBegin		()												= 0;	// Overall critical section begin.
	virtual void			CriticalEnd			()												= 0;	// Overall critical section end.
	virtual void			CriticalBeginAdc	()												= 0;	// ADC critical section begin.
	virtual void			CriticalEndAdc		()												= 0;	// ADC critical section end.
	virtual	unsigned long	ReadMem				( unsigned long address )						= 0;
	virtual	void			    WriteMem			( unsigned long address, unsigned long data )	= 0;
  virtual void ReadDmaMem( const DMABUF&, PUCHAR trgBuf, ULONG cntXferBytes,
       ULONG trgStartIdx = 0, ULONG dmaStartIdx = 0) = 0;
  virtual void WriteDmaMem( const DMABUF&, PUCHAR srcBuf, ULONG cntXferBytes,
       ULONG srcStartIdx = 0, ULONG dmaStartIdx = 0) = 0;
	// Functions that are always provided by the board class:
	unsigned long			GetAddress			();
	unsigned long			GetErrors			();
	void					    SetErrCallback		( FPTR_ERR Callback );
	// Storage:
	unsigned long			MyErrFlags;			// Error flags.
	unsigned short			MyCounterIntEnabs;	// Counter interrupt enable mask for MISC2 register.
protected:
	// Functions that must be implemented in the derived OS interface class:
	virtual void			AllocDMAB			( unsigned long nBytes, DMABUF *pDmaBuf )	= 0;	// Allocate DMA buffer.
	virtual void			CloseDMAB			( DMABUF *pDmaBuf )							= 0;	// Deallocate DMA buffer.
	// Functions provided by the board class:
	void					BoardClose			();
	void					BoardOpen			( unsigned char IsBoardRevA );
	void					DeclareError		( unsigned long ErrMask );
	// Storage that must be initialized by the derived OS interface class:
	unsigned long			MyAddress;			// PCI bus & slot number of card.
	unsigned long			MyBase;				// Board base memory address.
private:
	// Primitive functions:
	virtual	unsigned short	DIOBitInsert		( unsigned short WordVal, unsigned short Chan, unsigned short BitVal );
	void					DIOInit				();
	void					CountersInit		();
	unsigned long			I2Chandshake		( unsigned long val );
	void					I2CInit				();
	void					InitADC				();
	void					InitDAC				();
	void					LoadTrimDACs		();
	void					SendDAC				( unsigned long val );
	void					SetDAC				( unsigned short chan, short dacdata );
	// Helper functions:
	void					CreateCounters		();
	void					DEBItransfer		();
	void					DestroyCounters		();
	void					WriteMISC2			( unsigned short NewImage );
	unsigned short			WRMISC2image		();
	// Storage:
	ClsCounter				*Counters[6];		// Pointers to Counter Channel objects.
	FPTR_ERR				MyErrCallback;		// Pointer to API's error handler callback function.
	unsigned char			MyAdcItems;			// Number of items in ADC poll list.
	DMABUF					MyRPSBuf;			// DMA buffer used to hold ADC (RPS1) program.
	DMABUF					MyANABuf;			// DMA buffer used to receive ADC data and hold DAC data.
	//unsigned long			*pMyDacWBuf;		// Pointer to logical adrs of DMA buffer used to hold DAC data.
	unsigned short			MyDacpol;			// Image of DAC polarity register.
	short					MyDacSetpoint[4];	// Images of DAC setpoints.
	unsigned char			MyTrimSetpoint[12];	// Images of TrimDAC setpoints.
	unsigned short			MyDOutImage[2];		// Images of DIO output registers.
	unsigned short			MyChargeEnabled;	// Image of MISC2 Battery Charge Enabled (0 or WRMISC2_CHARGE_ENABLE).
	unsigned short			MyWDInterval;		// Image of MISC2 watchdog interval control bits.
	unsigned long			MyI2CAdrs;			// I2C device address for onboard EEPROM (board rev dependent).
};

#endif	// #ifdif DEF_CLS626
