//$Id$

/* ****************************************************************************
 *
 *  Cls626mm.h
 *  Created by David Beck, on Fri Feb 09 2007, 19:05:18 Pacific Standard Time
 *
 *  Copyright (c) 2006, indigocode - www.indigocode.com
 *
 *  Permission is granted to use this code without restriction,
 *  as long as this copyright notice appears in all source files.
 *
 *
 *  File Contents: Interface and documentation of the Cls626mm component.
 *
 *  %version: 1 %
 *  %date_modified: Fri Feb 09 2007, 19:05:18 Pacific Standard Time %
 *  %created_by: David Beck %
 *  Version: %I%
 *  Date modified: %G%
 *
 *************************************************************************** */
 
#ifndef IC__CLS626MM_H
#define IC__CLS626MM_H

#include <boost/array.hpp>
#include <boost/utility.hpp>

#include "Cls626.h"
#include "DeviceList.h"

#include <ace/Refcounted_Auto_Ptr.h>
#include <ace/Recursive_Thread_Mutex.h>
#include <ace/Task.h>

// forward declaration
class S626IntrFifoBufSet;

//------------------------------------------------------------------------------
//  IOCTL codes to access driver.
// -----------------------------------------------------------------------------
#define IOCTL_CLS626_GET_VERSION      0x222000  // Retrieve version information (CLS626_GET_VERSION structure)
#define IOCTL_CLS626_GET_STATUS       0x222004  // Retrieve status information (CLS626_GET_STATUS structure) 
#define IOCTL_CLS626_SAA7146_READ     0x222008  // Read  a SAA7146 register 
#define IOCTL_CLS626_SAA7146_WRITE    0x22200C  // Write a SAA7146 register 
#define IOCTL_CLS626_WAIT_NOTIFY      0x222010  // Wait for a notification 
#define IOCTL_CLS626_GENERATE_EVENT   0x222014  // Manually trigger the notification 
#define IOCTL_CLS626_DMA_ALLOCATE     0x222018  // Allocate  DMA buffer 
#define IOCTL_CLS626_DMA_RELEASE      0x22201C  // Release   DMA buffer
#define IOCTL_CLS626_DMA_READ         0x222020  // Read from DMA buffer
#define IOCTL_CLS626_DMA_WRITE        0x222024  // Write to  DMA buffer
#define IOCTL_FIFO_ALLOCATE           0x222028  // Allocate interrupt evt data capture circular DMA buffers
#define IOCTL_FIFO_RELEASE            0x22202C  // Release interrupt evt data capture circular DMA buffers
#define IOCTL_FIFO_READ               0x222030  // Remove data from interrupt evt data capture circular DMA buffer
#define IOCTL_IRQ_WRITE               0x222034  // Write interrupt action/handling mechanism
#define IOCTL_IRQ_READ                0x222038  // Read interrupt action/handling mechanism
#define IOCTL_CLS626_BUS_LOC          0x22203C  // Read PCI bus location (CLS626_GET_ADDRESS)
#define IOCTL_IRQ_FIFOS_READ          0x222040  // Remove all data from interrupt evt data capture (all buffers for specified IRQ)

#ifndef NTSTATUS
typedef LONG NTSTATUS;
#endif
#define STATUS_SUCCESS                   ((NTSTATUS)0x00000000L) // ntsubauth
#define STATUS_MORE_ENTRIES              ((NTSTATUS)0x00000105L)
#define STATUS_INTERNAL_ERROR            ((NTSTATUS)0xC00000E5L)

typedef ACE_Guard<ACE_Recursive_Thread_Mutex> AceGdRecThrMtx;

/** Purpose: Describe the purpose of the class.
 *  (Everything that you write here will show up in the
 *  doc++ generated documentation.)
 */
class Cls626mm : public Cls626, public ACE_Task_Base, boost::noncopyable
{  
public:

  // Structure used for returning status information (native size: 32 or 64-bit)
  typedef struct _GETSTATUS
  {
    ULONG            interrupts;                           // Interrupts occured
    ULONG            isr;                                  // Interrupt status register of last interrupt
    PVOID            vaBuffer;                             // Virtual address DMA buffer
    PHYSICAL_ADDRESS paBuffer;                             // Physical address DMA buffer
    ULONG            buffersize;                           // Size of DMA buffer (in bytes)
    ULONG            fifoOverflows;                        // Global FIFO overflows
  } GETSTATUS, *PGETSTATUS;

  // Structure used to pass transfer buffer information (native size: 32 or 64-bit)
  typedef struct _TRANSFERBUFFER
  {
    LONG             bufferId;                             // Identification of buffer
    PUCHAR           bufferTransfer;                       // Transfer buffer (source/target)
    ULONG            bufferSourceIndex;                    // Source transfer index in buffer
    ULONG            bufferTargetIndex;                    // Target transfer index in buffer
    ULONG            bufferTransferLength;                 // Transfer length
  } TRANSFERBUFFER, *PTRANSFERBUFFER;

  // Structure used to pass FIFO buffer information (native size: 32 or 64-bit)
  typedef struct _FIFOTRANSFERBUFFER
  {
    LONG             bufferId;                // Identification of FIFO buffer (<0 means not in use)
    LONG             buffers;                 // Number of buffers
    PVOID            bufferTransfer;          // Transfer buffer (source/target)
    ULONG            bufferTransferLength;    // Transfer length (buffer sizes)
    BOOLEAN          bufferWritten;           // Copy of <writeTag> indicating buffer has been written to
    ULONG            bufferOrder;             // Copy of <irqsActiveReceived> counter when FIFO was written = order number
    ULONG            bufferOverflows;         // Counter incremented when written to buffer with <writeTag> TRUE
    ULONG            bufferIrqs;              // Counter incremented for each interrupt associated with the FIFO
    ULONG            bufferIrqOverflows;      // Counter incremented when written to buffer with <writeTag> TRUE for ALL FIFO buffers of the associated IRQ
  } FIFOTRANSFERBUFFER, *PFIFOTRANSFERBUFFER;

  typedef struct _IRQFIFOSREADBUFFER
  {
    LONG                IrqId;
    ULONG               cntFifosInBuf;
    FIFOTRANSFERBUFFER  fifos[1];
  } IRQFIFOSREADBUFFER, *PIRQFIFOSREADBUFFER;

  typedef struct _GET_VERSION
  {
    unsigned short  MajorVersion;  // Major version
    unsigned short  MinorVersion;  // Minor version
    unsigned long   Build;         // Build (date): $YYYMMDD
  } GET_VERSION;

  typedef struct _DMABUFFER // (used by DMAAllocate)
  {
    LONG              BufferId;       // ID of buffer
    PVOID             VaBuffer;       // Virtual address of DMA buffer
    PHYSICAL_ADDRESS  PaBuffer;       // Physical address of DMA buffer
    ULONG             BufferSize;     // Size of DMA buffer (bytes)
  } DMABUFFER;                       

  // -- IRQ notification information
  // Each interrupt can signal something, so a RPS program
  // can be informed about a specific interrupt being generated.
  // An interrupt can also trigger a buffering action. It uses the FIFO buffers for this.
  // A buffering action always copies a complete buffer (<bufferSize> of FIFOBUFFER).
  typedef struct _IRQBUFFER
  {
    ULONG            irqsReceived;                         // Counter incremented for each interrupt of this type
    ULONG            irqsActiveReceived;                   // Counter incremented for each interrupt of this type when active
    BOOLEAN          active;                               // Global activation flag of the interrupt
    BOOLEAN          eventActive;                          // Indicates notify event to be triggered
    BOOLEAN          signalActive;                         // Indicates SIGx to be triggered
    BOOLEAN          fifoActive;                           // Indicates FIFO to be used
    BOOLEAN          irqAutoDisable;                       // Indicates if interrupt is only triggered once (interrupt enable register is cleared)
                                                           // which is necessary for 'persistent' interrupts (like RPS timeout)

    ULONG            signalRegister;                       // SAA7146 register to read and write back for banking info
    ULONG            signalAnd;                            // AND operation to perform on register contents to extract bit
    ULONG            signalOr;                             // OR  peration to perform on ANDED register contents to extract bit
    ULONG            signalXor;                            // XOR (invert) operation to perform on ANDED and ORED register contents to write bit

    UCHAR            fifoBufferPrevious;                   // Previous FIFO used (index FIFO buffer array)
    UCHAR            fifoBufferFirst;                      // First FIFO to use (index FIFO buffer array)
    UCHAR            fifoBufferLast;                       // Last  FIFO to use (index FIFO buffer array)
    ULONG            fifoBufferCirculated;                 // Counter incremented each time <fifoBufferNext> == <fifoBufferStart>
    ULONG            fifoOverflows;                        // Counter incremented each time a FIFO <writeTag> was still set when written to it
  } IRQBUFFER;

  typedef struct _IRQTRANSFERBUFFER // IRQ notificiation information
  {
    LONG             IrqId;           // Identification of SAA7146A IRQ (0..31)
    IRQBUFFER        IrqInformation;  // Information
  } IRQTRANSFERBUFFER;

  // S626 Interrupt Status for all channels (DIO, Counter Index, Counter Overflow)
  typedef struct _S626IRQSTATUS
  {
    LARGE_INTEGER timeStamp;
    USHORT        intrStatusWord[4]; // format as described in Sensoray S626 Manual
  } S626IRQSTATUS;
  
  // S626 IRQ numbers map using this offset
  //static const unsigned long S626IRQNUMOFFSET = 32;
  
  enum _S626IRQPARAMS { S626IRQNUMOFFSET = 32, S626DIOIRQSTARTNUM = 0, S626DIOIRQENDNUM = 39 };

public:

  /// Constructor.
  Cls626mm( HANDLE hnd, const CDeviceListEntry& devInfo, FPTR_ISR IntFunc, 
    unsigned long Priority );
  virtual ~Cls626mm();  

private:
  // Prevent use of default c-tor:
  Cls626mm();

public:
	// Functions that must be implemented in the derived OS interface class:

  // Overall critical section begin.
	virtual void			    CriticalBegin();	
  // Overall critical section end.
	virtual void			    CriticalEnd();	
  // ADC critical section begin.
  virtual void			    CriticalBeginAdc();	
  // ADC critical section end.
	virtual void			    CriticalEndAdc();	

	virtual	unsigned long	ReadMem( unsigned long address );
	virtual	void			    WriteMem( unsigned long address, unsigned long data );

  // Allocate DMA buffer.
	virtual void			    AllocDMAB( unsigned long nBytes, DMABUF *pDmaBuf );	
  // Deallocate DMA buffer.
	virtual void			    CloseDMAB( DMABUF *pDmaBuf );

  virtual void ReadDmaMem( const DMABUF&, PUCHAR trgBuf, ULONG cntXferBytes,
       ULONG trgStartIdx = 0, ULONG dmaStartIdx = 0);  
  virtual void WriteDmaMem( const DMABUF&, PUCHAR srcBuf, ULONG cntXferBytes, 
    ULONG srcStartIdx = 0, ULONG dmaStartIdx = 0);

  //Functions that optionally are implemented in the derived OS interface class:

  virtual unsigned long GetAddress();
  virtual void InterruptStatus( unsigned short* IntFlags );

  inline LARGE_INTEGER InterruptTimestamp() const
  {
    return irqStatus_.timeStamp;
  };

  inline HANDLE getDrvHnd() { return drvHnd_; }

  // FIFO buffer operations (a SAA7146A driver feature)
  long OpenIrqFifoBuffers( unsigned long irqId, unsigned long* regAddrs, 
                             long numAddrs, long numBufsReq );
  void CloseIrqFifoBuffers( unsigned long irqId ); 
  unsigned long FetchIrqFifoBuffers( unsigned long irqId, unsigned char* buffer, 
        unsigned long bufferByteCnt ); // returns count of bytes fetched
  long AllocFifoBuffers( long bufferCnt, void* dmaSrcAddr, 
        unsigned long fifoBufByteSize ); // returns startFifoBufId of range allocated
  long CloseFifoBuffers( long startFifoBufId, long bufCnt, 
        long& bufCntClosed ); // returns startFifoBufId of range closed

  // IRQ event (handling) operations
  inline  ACE_Recursive_Thread_Mutex&  CriticalS626IrqConfig() { return irqBufConfigLock_; }
  void    WriteS626IrqConfig( long irqId, IRQBUFFER& irqInfo );
  void    ReadS626IrqConfig( long irqId, IRQBUFFER& irqInfo );

  // Log driver errors
  static void logMsg( const ACE_TCHAR* msg, unsigned long errCode = 0ul );

protected:
	// Functions that must be implemented in the derived OS interface class:
  
  // Interrupt notification thread methods
  void StopIntrNotifyThrd();
  //  ACE_Task_Base method run by a daemon thread
  virtual int svc (void);

private:
  static void traceAllSAA7146AIntrs( HANDLE hnd, IRQTRANSFERBUFFER xferBufs[] );

private:

  // SAA7146A interrupt number representing S626 master interrupt
  static const int SAA7146A_DEBI_BUS_INTR_BIT = 6;
  // Number of S626 IRQ buffers 
  static const int SAA7146A_S626_IRQBUFCNT = 40+12;  
  // S626 IRQ numbers map into SAA7146 interrupt numbers using this offset
  static const int SAA7146A_S626_IRQNUMOFFSET = 32;  

  HANDLE                      drvHnd_;
  const CDeviceListEntry      devInfo_;
  ACE_Recursive_Thread_Mutex  genLock_;
  ACE_Recursive_Thread_Mutex  adcLock_;
  ACE_Recursive_Thread_Mutex  irqBufConfigLock_;

  FPTR_ISR                    intrHdlrCallBack_;
  ACE_hthread_t               threadHandles_[1];
  bool                        shutdownIntrHdlrThrd_;
  S626IRQSTATUS               irqStatus_;

  typedef boost::array< S626IntrFifoBufSet*, SAA7146A_S626_IRQBUFCNT > S626INTRFIFOBUFMAP;
  S626INTRFIFOBUFMAP          s626IntrFifoBufSets_;
};

typedef ACE_Refcounted_Auto_Ptr< Cls626mm, ACE_Recursive_Thread_Mutex > Cls626mm_Ptr;

#include "Cls626mm.inl"
#endif // #ifndef IC__CLS626MM_H
 
