//$Id$

/* ****************************************************************************
 *
 *  S626IntrFifoBufSet.cpp
 *  Created by David Beck, on Sun Mar 18 2007, 16:25:59 Pacific Daylight Time
 *
 *  Copyright (c) 2006, indigocode - www.indigocode.com
 *
 *  Permission is granted to use this code without restriction,
 *  as long as this copyright notice appears in all source files.
 *
 *
 *  File Contents: Implementation of the S626IntrFifoBufSet component.
 *  Please see S626IntrFifoBufSet.h for full documentation of this system.
 *
 *  %version: 1 %
 *  %date_modified: Sun Mar 18 2007, 16:25:59 Pacific Daylight Time %
 *  %created_by: David Beck %
 *  Version: %I%
 *  Date modified: %G%
 *
 *************************************************************************** */

#include "stdafx.h"

#include <ace/OS.h>
#include "Trace.h"

#undef ACE_NTRACE
#define ACE_NTRACE 0

#include "Cls626mm.h"
#include "S626IntrFifoBufSet.h"

#include <boost/format.hpp>
#include <boost/scoped_array.hpp>

S626IntrFifoBufSet::S626IntrFifoBufSet( Cls626mm* board, 
                                       const unsigned long s626IntrId, 
                                       const RegAddrList& regAddrList, 
                                       const long fifoBufCnt ) 
     : board_( board ), s626IntrId_( s626IntrId ), startingFifoId_( -1 ),
     fifoBufCnt_( fifoBufCnt )
{
  // allocate DMA buffer and install register address list in the buffer
  dmaBufByteCnt_ = 
       sizeof(S626IRQREGLIST) + (sizeof(UINT32)*(unsigned long)(regAddrList.size() - 1));
  
  ::memset( &dmaBuf_, 0, sizeof(dmaBuf_) );
  ACE_ASSERT( board_ );
  board_->AllocDMAB( dmaBufByteCnt_, &dmaBuf_ );
  
  if( dmaBufByteCnt_ == dmaBuf_.bufByteSize ) 
  {
    // need a buffer to write initialization data into the DMA
    boost::scoped_array< unsigned char > writeDmaBuf( new unsigned char [ dmaBufByteCnt_ ] );
    ::memset( writeDmaBuf.get(), 0, dmaBufByteCnt_ );
        
    PS626IRQREGLIST irqRegList = (PS626IRQREGLIST)writeDmaBuf.get();
    irqRegList->regCount = (LONG)regAddrList.size();
    
    // copy register addresses into buffer
    std::copy( regAddrList.begin(), regAddrList.end(), irqRegList->s626RegAddr );

    // write into the allocated DMA buffer
    board_->WriteDmaMem( dmaBuf_, writeDmaBuf.get(), dmaBufByteCnt_ );

    // allocate FIFO buffers
    startingFifoId_ = board->AllocFifoBuffers( fifoBufCnt, dmaBuf_.LogicalBase, 
      dmaBuf_.bufByteSize );  
  }
  else
  {    
    Cls626mm::logMsg( 
      boost::str( boost::wformat( 
      ACE_TEXT("S626IntrFifoBufSet(): DMA alloc size err: req=%d, actual=%d") )
      % dmaBufByteCnt_ % dmaBuf_.bufByteSize ).c_str(), ::GetLastError() );
  }

  if( -1 < startingFifoId_ )
  {
    // config IRQ with FIFO buffers
    Cls626mm::IRQBUFFER irqConfig = { 0 };

    AceGdRecThrMtx guard( board_->CriticalS626IrqConfig() );
    board_->ReadS626IrqConfig( s626IntrId_, irqConfig );

    irqConfig.fifoActive = TRUE;
    irqConfig.fifoBufferFirst = (UCHAR)startingFifoId_;
    irqConfig.fifoBufferLast = (UCHAR)(irqConfig.fifoBufferFirst + fifoBufCnt - 1);
    irqConfig.fifoBufferPrevious = irqConfig.fifoBufferLast;

    irqConfig.active = 
      irqConfig.eventActive || irqConfig.fifoActive || irqConfig.signalActive;

    board_->WriteS626IrqConfig( s626IntrId_, irqConfig );

    // ensure S626 interrupt source is enabled (DIO or Counter)
    if( Cls626mm::S626DIOIRQENDNUM < s626IntrId_ )
    {
      // Counter interrupt

      const unsigned short counterNum = 
                  ((unsigned short)s626IntrId_ - Cls626mm::S626DIOIRQENDNUM - 1) % 6;

      // read existing counter
      unsigned short cntrIntrs = board_->CounterIntSourceGet( counterNum );

      //Cls626mm::logMsg( 
      //  boost::str( boost::wformat( 
      //  ACE_TEXT("S626IntrFifoBufSet(): S626Irq=%d, cntr num=%d, existFlags=0x%04X") )
      //  % s626IntrId_ % counterNum % cntrIntrs ).c_str(), ::GetLastError() );

      // INTSRC_OVER = 1, INTSRC_INDEX = 2 (IRQ 40-45=INDX, 46-51=OVER)
      cntrIntrs |= ( 2 >> ((s626IntrId - Cls626mm::S626DIOIRQENDNUM - 1) / 6) );

      //Cls626mm::logMsg( 
      //  boost::str( boost::wformat( 
      //  ACE_TEXT("S626IntrFifoBufSet(): S626Irq=%d, cntr num=%d, updateFlags=0x%04X") )
      //  % s626IntrId_ % counterNum % cntrIntrs ).c_str(), ::GetLastError() );
      
      board_->CounterIntSourceSet( counterNum, cntrIntrs );
    }
    else
    {
      // DIO interrupt
      board_->DIOChanIntEnable( (unsigned short)s626IntrId, 1 );
    }
  }
}

S626IntrFifoBufSet::~S626IntrFifoBufSet()
{
  // ensure S626 interrupt source is enabled (DIO or Counter)
  if( Cls626mm::S626DIOIRQENDNUM < s626IntrId_ )
  {
    // Counter interrupt

    const unsigned short counterNum = 
                ((unsigned short)s626IntrId_ - Cls626mm::S626DIOIRQENDNUM - 1) % 6;

    // read existing counter
    unsigned short cntrIntrs = board_->CounterIntSourceGet( counterNum );

    //Cls626mm::logMsg( 
    //    boost::str( boost::wformat( 
    //    ACE_TEXT("~S626IntrFifoBufSet(): S626Irq=%d, cntr num=%d, existFlags=0x%04X") )
    //    % s626IntrId_ % counterNum % cntrIntrs ).c_str(), ::GetLastError() );

    // INTSRC_OVER = 1, INTSRC_INDEX = 2 (IRQ 40-45=INDX, 46-51=OVER)
    cntrIntrs &= ~( 2 >> ((s626IntrId_ - Cls626mm::S626DIOIRQENDNUM - 1) / 6) );

    //Cls626mm::logMsg( 
    //  boost::str( boost::wformat( 
    //  ACE_TEXT("~S626IntrFifoBufSet(): S626Irq=%d, cntr num=%d, updateFlags=0x%04X") )
    //  % s626IntrId_ % counterNum % cntrIntrs ).c_str(), ::GetLastError() );

    board_->CounterIntSourceSet( counterNum, cntrIntrs );
  }
  else
  {
    // DIO interrupt
    board_->DIOChanIntEnable( (unsigned short)s626IntrId_, 0 );
  }

  {
    // deactivate IRQ
    Cls626mm::IRQBUFFER irqConfig = { 0 };

    AceGdRecThrMtx guard( board_->CriticalS626IrqConfig() );
    board_->ReadS626IrqConfig( s626IntrId_, irqConfig );

    irqConfig.fifoActive = FALSE;
    irqConfig.active = 
        irqConfig.eventActive || irqConfig.fifoActive || irqConfig.signalActive;

    board_->WriteS626IrqConfig( s626IntrId_, irqConfig );
  }

  // release FIFO buffers
  if( -1 < startingFifoId_ )
  {
    long bufCntClosed;
    board_->CloseFifoBuffers( startingFifoId_, fifoBufCnt_, bufCntClosed );
    startingFifoId_ = -1;
  }

  // release DMA buffer
  if( 0 < dmaBuf_.bufByteSize )
  {
    board_->CloseDMAB( &dmaBuf_ );
    dmaBuf_.bufByteSize = 0;
  }
}

unsigned long S626IntrFifoBufSet::fetchBuffers( unsigned char* buffer, 
                                               unsigned long bufferByteCnt )
{
  unsigned long byteCntFetched = 0;

  // check and round incoming buffer size to nearest integral 
  //  multiple of DMA buffer size used with these FIFO's
  //  (ie. the size of the FIFO chunks to be retrieved)
  const unsigned long numFifosToRead = bufferByteCnt / dmaBufByteCnt_;

  // determine the IOCTL buffer size required
  // it must hold 1) IRQFIFOSREADBUFFER, 2) additional FIFOTRANSFERBUFFER data
  //  for multiple FIFO's read (first one included in IRQFIFOSREADBUFFER),
  //  and then 3) the chunks of actual FIFO data
  const unsigned long fifoReadBufHdrsSize = sizeof(Cls626mm::IRQFIFOSREADBUFFER) +                               
    (numFifosToRead - 1) * sizeof(Cls626mm::FIFOTRANSFERBUFFER);

  const unsigned long fifoReadBufSize = fifoReadBufHdrsSize + 
                            numFifosToRead * dmaBufByteCnt_;

  boost::scoped_array< unsigned char > 
                          fifosReadBuf( new unsigned char [ fifoReadBufSize ] );
  ::memset( fifosReadBuf.get(), 0, fifoReadBufSize);

  Cls626mm::IRQFIFOSREADBUFFER* bufHdr = 
    (Cls626mm::IRQFIFOSREADBUFFER*)fifosReadBuf.get();
  bufHdr->IrqId = s626IntrId_ + Cls626mm::S626IRQNUMOFFSET;
  bufHdr->cntFifosInBuf = numFifosToRead;

  // set transferBuffer size in each FIFOTRANSFERBUFFER
  Cls626mm::PFIFOTRANSFERBUFFER fifoBuf = bufHdr->fifos;
  for( unsigned long i = 0; i < numFifosToRead; ++i, ++fifoBuf )
  {
    fifoBuf->bufferTransferLength = dmaBufByteCnt_;
  }

  DWORD cntRead = 0;

  if( !::DeviceIoControl( board_->getDrvHnd(), IOCTL_IRQ_FIFOS_READ, 
    (LPVOID) fifosReadBuf.get(), fifoReadBufSize, (LPVOID) fifosReadBuf.get(), 
                  fifoReadBufSize, &cntRead, NULL ) )
  {
    DWORD status = ::GetLastError();
    if( status == STATUS_MORE_ENTRIES )
    {
      Cls626mm::logMsg( boost::str( boost::wformat( 
        ACE_TEXT("Cls626mm::FetchIrqFifoBuffers: irqId=%d - STATUS_MORE_ENTRIES") 
        ) % bufHdr->IrqId ).c_str(), ::GetLastError() );
    }
    else
    {
      Cls626mm::logMsg( boost::str( boost::wformat( 
        ACE_TEXT("Cls626mm::FetchIrqFifoBuffers: ERROR, irqId=%d") ) % bufHdr->IrqId ).c_str(),
        status );

      // TODO: set MyErrFlags
    }
  } 
  else
  {
    if( cntRead )
    {
      // determine number of FIFO's actually fetched
      const unsigned long numFifosFetched = (cntRead - fifoReadBufHdrsSize) / dmaBufByteCnt_;

      // copy FIFO data back to caller's buffer and report bytes retrieved      
      unsigned char* fifoData = fifosReadBuf.get() + fifoReadBufHdrsSize;
      unsigned char* callersData = buffer;

      for( unsigned long i = 0; i < numFifosFetched; ++i )
      {
        ::memcpy( callersData, fifoData, dmaBufByteCnt_ );
        callersData += dmaBufByteCnt_;
        fifoData += dmaBufByteCnt_;
      }
      
      byteCntFetched = numFifosFetched * dmaBufByteCnt_;
    }
  }

  // returns count of bytes fetched
  return byteCntFetched;
}
